#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QTimer>
#include <QSplitter>
#include "registerviewwidget.h"
#include "abstractviewwidget.h"
#include "viewwidgetcontainer.h"
#include "disassemblerview.h"
#include "memorymapviewwidget.h"

namespace Ui {
class MainWindow;
}

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(QWidget *parent = nullptr);
    ~MainWindow();
public slots:
    void viewCloseRequested();
private slots:
    void parseRegisterDefinition();
    void verifyMemoryChange();
    void on_action_Add_Hub_Regs_triggered();
    void on_action_Add_Cog_Regs_triggered();
    void on_action_Add_Scratchpad_Asm_triggered();
    void on_action_Add_Dynamic_Cog_Asm_triggered();
    void on_action_Add_Memory_Map_triggered();
signals:
    void verifyMemorySignal();
private:
    void addView(QString id);
    QList<ViewWidgetContainer*> m_viewWidgets;
    QTimer *m_timer;
    Ui::MainWindow *ui;
    QSplitter *m_mainSplitter;
    MemoryMap *m_sharedMemoryMap;
    MemoryMap *m_copyPrevMemoryMap;
    quint8 *m_intermediateMemory;
    bool m_inMemoryUpdate;
};

#endif // MAINWINDOW_H
