#include "memorymapviewwidget.h"
#include <QtDebug>

MemoryMapViewWidget::MemoryMapViewWidget(MemoryMap *newMap):
    QTreeWidget(),
    m_newMap(newMap),
    m_lastMasterPtr(m_newMap->getMasterPointer())
{
    setFont(QFont("Monospace",6));
    setHeaderLabels(QStringList()<<"Name"<<"Rel. Address"<<"Abs Address");
    updateView();
}

void MemoryMapViewWidget::checkIfMemoryChanged() {
    auto mptr = m_newMap->getMasterPointer();
    if (m_lastMasterPtr == mptr)
        return;
    m_lastMasterPtr = mptr;
    updateView();
}

QTreeWidgetItem *MemoryMapViewWidget::createItem(QString name, quint8* ptr) {
    quint64 relPtr = quint64(ptr-m_newMap->fullMemoryMap());
    QString absAddr = QString::number(relPtr+m_lastMasterPtr,16).rightJustified(16,'_');
    return new QTreeWidgetItem(QStringList()
                               <<name
                               <<QString::number(relPtr,16).rightJustified(5,'0')
                               <<(absAddr.mid(0,4)+":"+absAddr.mid(4,4)+":"+absAddr.mid(8,4)+":"+absAddr.mid(12,4))
                               );
}

void MemoryMapViewWidget::updateView() {
    clear();
    addTopLevelItem(createItem("Hub Memory",m_newMap->hubMemory()));
    auto cogMemTr = createItem("Cog Memory",m_newMap->cogMemory(0));
    auto cogRegTr = createItem("Cog Registers",m_newMap->cogRegisters(0));
    auto cogDynAsmTr = createItem("Cog Dynamic Asm",m_newMap->cogDynamicAsm(0));
    for (int i=0; i<8; ++i) {
        cogMemTr->addChild(createItem(QString::number(i),m_newMap->cogMemory(i)));
        cogRegTr->addChild(createItem(QString::number(i),m_newMap->cogRegisters(i)));
        cogDynAsmTr->addChild(createItem(QString::number(i),m_newMap->cogDynamicAsm(i)));
    }
    addTopLevelItem(cogMemTr);
    addTopLevelItem(cogRegTr);
    addTopLevelItem(createItem("Hub Registers",m_newMap->hubRegisters()));
    addTopLevelItem(createItem("Hub Scratchpad",m_newMap->scratchPad()));
    addTopLevelItem(cogDynAsmTr);
            /*
            quint8* hubMemory() const { return &m_startOfMemoryMap[GlobalHubMemoryOffset]; }
            quint8* hubRegisters() const { return &m_startOfMemoryMap[GlobalHubRegisterOffset]; }
            quint8* hubDynamicAsm() const { return &m_startOfMemoryMap[GlobalDynamicHubAsmOffset]; }
            quint8* scratchPad() const { return &m_startOfMemoryMap[GlobalScratchpadOffset]; }
            quint8* staticAsm() const { return &m_startOfMemoryMap[GlobalStaticAsmOffset]; }
            quint8* cogMemory(int cog) const { return &m_startOfMemoryMap[GlobalCogMemoryOffset+CogMemoryMultiplicator*cog]; }
            quint8* cogRegisters(int cog) const { return &m_startOfMemoryMap[GlobalCogRegisterOffset+CogMemoryMultiplicator*cog]; }
            quint8* cogDynamicAsm(int cog) const { return &m_startOfMemoryMap[GlobalDynamicCogAsmOffset+CogMemoryMultiplicator*cog]; }*/
}
