#include "registerviewwidget.h"
#include <QLabel>

RegisterViewWidget::RegisterViewWidget(int cogIdOrHub, MemoryMap *newMap, MemoryMap *oldMap):
    QListWidget(nullptr),
    m_newMap(newMap),
    m_oldMap(oldMap),
    m_cogIdOrHub(cogIdOrHub)
{
    setFont(QFont("Monospace",6));
    setWindowTitle(cogIdOrHub<0 ? QString("Hub Regs") : QString("Cog %1 Regs").arg(cogIdOrHub));
}

void RegisterViewWidget::updateRegisterDefinitions(QList<RegisterInfo> registers) {
    m_registers = registers;
    clear();
    for (auto r: m_registers)
        addItem(getRegisterLine(r));
}

QString RegisterViewWidget::getRegisterLine(const RegisterInfo& r) const {
    const quint8* ptr = (m_cogIdOrHub<0 ? m_newMap->hubRegisters() : m_newMap->cogRegisters(m_cogIdOrHub))+r.relativeOffset;
    QString result = r.name.leftJustified(60,' ')+" ";
    switch(r.type) {
        case RegisterInfo::Boolean:
            result += (*ptr == 1) ? QString("TRUE") : ((*ptr == 0) ? QString("false") : QString("illg bool %1").arg(*ptr));
            break;
        case RegisterInfo::UInt32:
            result += "0x"+QString::number(*reinterpret_cast<const quint32*>(ptr),16).rightJustified(8,'_');
            break;
        case RegisterInfo::UInt64:
            result += "0x"+QString::number(*reinterpret_cast<const quint64*>(ptr),16).rightJustified(16,'_');
            break;
        case RegisterInfo::Pointer:
            result += "Px"+QString::number(*reinterpret_cast<const quint64*>(ptr),16).rightJustified(16,'_');
            break;
    }
    return result;
}

void RegisterViewWidget::checkIfMemoryChanged() {
    if (m_cogIdOrHub<0 && memcmp(m_newMap->hubRegisters(),m_oldMap->hubRegisters(),MemoryMap::HubRegisterSize)==0)
        return;
    if (m_cogIdOrHub>=0 && memcmp(m_newMap->cogRegisters(m_cogIdOrHub),m_oldMap->cogRegisters(m_cogIdOrHub),MemoryMap::CogRegisterSize)==0)
        return;
    qDebug()<<"changed";
    for (int i=0; i<m_registers.size(); ++i) {
        auto newStr = getRegisterLine(m_registers[i]);
        auto it = item(i);
        if (it->text() == newStr)
            it->setTextColor(QColor(255,255,255));
        else
            it->setTextColor(QColor(255,0,0));
        it->setText(newStr);
    }
}
