import cppgenerator.generateCppForParsedModule
import flatten.flattenModule
import ilanggenerator.generateILangForParsedModule
import ilanggenerator.moduleFunctionToILang
import ilanggenerator.pureFunctionToILang
import ilanggenerator.regSetToILang
import verilogparser.UnitParser
import verilogparser.fileLexer
import java.io.File
import java.lang.Exception
import java.lang.StringBuilder
import java.math.BigInteger

class BuiltInModules() {
    val builtIn = builtInFunctions()
    fun findSpecifier(lst: Pair<SpecifierReference.Type,List<Specifier>>, name: String): SpecifierReference {
        for ((i, l) in lst.second.withIndex())
            if (l.name == name)
                return SpecifierReference(-1, if (lst.first == SpecifierReference.Type.Register) -1 else 0, i, lst.first)
        throw Exception("no var $name found")
    }

    fun readSpecifier(lst: Pair<SpecifierReference.Type,List<Specifier>>, name: String): SpecifierExpression {
        for ((i, l) in lst.second.withIndex())
            if (l.name == name)
                return SpecifierExpression(SpecifierReference(-1, if (lst.first == SpecifierReference.Type.Register) -1 else 0, i, lst.first), l.dataType)
        throw Exception("no var $name found")
    }

    fun callFunc(funcName:String, arg1: ParsedExpression, arg2: ParsedExpression, arg3: ParsedExpression): ParsedExpression {
        return PureFunctionCall(builtIn[funcName]!!, arrayListOf(arg1,arg2,arg3))
    }

    fun callFunc(funcName:String, arg1: ParsedExpression, arg2: ParsedExpression): ParsedExpression {
        return PureFunctionCall(builtIn[funcName]!!, arrayListOf(arg1,arg2))
    }

    fun callFunc(funcName:String, arg1: ParsedExpression): ParsedExpression {
        return PureFunctionCall(builtIn[funcName]!!, arrayListOf(arg1))
    }

    fun i32(no:Int):ParsedExpression = ConstantValueExpression(BigInteger.valueOf(no.toLong()),DataType.TypeUInt32)
    fun i32(no:Long):ParsedExpression = ConstantValueExpression(BigInteger.valueOf(no),DataType.TypeUInt32)

    fun andB(arg1:ParsedExpression, arg2:ParsedExpression) = callFunc("andB", arg1, arg2)
    fun and32(arg1:ParsedExpression, arg2:ParsedExpression) = callFunc("and32", arg1, arg2)
    fun or32(arg1:ParsedExpression, arg2:ParsedExpression) = callFunc("or32", arg1, arg2)
    fun testBit32(arg1:ParsedExpression, arg2:Int) = callFunc("testBit32", arg1, i32(arg2))
    fun mux32(arg1:ParsedExpression, arg2:ParsedExpression, arg3:ParsedExpression) = callFunc("mux32", arg1, arg2, arg3)
    fun numberIfBitSet(lst: Pair<SpecifierReference.Type,List<Specifier>>, name: String, testBit:Int, no:Long) = mux32(
            testBit32(readSpecifier(lst, name),testBit),i32(no),i32(0)
    )

    fun assign(lst: Pair<SpecifierReference.Type,List<Specifier>>, name: String, value:ParsedExpression):ParsedStatement {
        return AssignmentStatement(findSpecifier(lst, name), value)
    }

    fun ifNoElse(condition:ParsedExpression, ifBody:ArrayList<ParsedStatement>):ParsedStatement {
        return IfStatement(condition, ifBody, ArrayList())
    }

    fun cogRam(): ParsedModule {
        val registers = Pair(SpecifierReference.Type.Register,arrayListOf(
                Specifier("q_reg", DataType.TypeUInt32, -1),
                Specifier("r", DataType.TypeUInt32, 512)
        ))
        val inputs = Pair(SpecifierReference.Type.Input,arrayListOf(
                Specifier("posedge_clk_cog", DataType.TypeBoolean, -1),
                Specifier("ena", DataType.TypeBoolean, -1),
                Specifier("w", DataType.TypeBoolean, -1),
                Specifier("d", DataType.TypeUInt32, -1),
                Specifier("a", DataType.TypeUInt32, -1)
        ))
        val outputs = Pair(SpecifierReference.Type.Output,arrayListOf(Specifier("q", DataType.TypeUInt32, -1)))
        val body = arrayListOf(
                assign(outputs, "q", readSpecifier(registers, "q_reg")),
                ifNoElse(andB(
                            readSpecifier(inputs, "posedge_clk_cog"),
                            readSpecifier(inputs, "ena")
                        ),
                        arrayListOf(
                                assign(registers, "q_reg", SpecifierIndexExpression(findSpecifier(registers, "r"), readSpecifier(inputs, "a"), DataType.TypeUInt32)),
                                ifNoElse(readSpecifier(inputs, "w"), arrayListOf(
                                        IndexAssignmentStatement(findSpecifier(registers, "r"), readSpecifier(inputs, "a"), readSpecifier(inputs, "d"))
                                ))
                ))
        )
        return ParsedModule("cloned_cog_ram", false, registers.second, arrayListOf(ParsedModuleFunction(
                inputs.second, outputs.second, ArrayList(), body)), ArrayList(), ArrayList())
    }

    fun hubRam(): ParsedModule {
        val registers = Pair(SpecifierReference.Type.Register,arrayListOf(
                Specifier("q_reg", DataType.TypeUInt32, -1),
                Specifier("memory", DataType.TypeUInt32, 16 * 1024)
        ))
        val inputs = Pair(SpecifierReference.Type.Input,arrayListOf(
                Specifier("posedge_clk_cog", DataType.TypeBoolean, -1),
                Specifier("ena_bus", DataType.TypeBoolean, -1),
                Specifier("w", DataType.TypeBoolean, -1),
                Specifier("wb", DataType.TypeUInt32, -1),
                Specifier("a", DataType.TypeUInt32, -1),
                Specifier("d", DataType.TypeUInt32, -1)
        ))
        val outputs = Pair(SpecifierReference.Type.Output,arrayListOf(Specifier("q", DataType.TypeUInt32, -1)))
        val locals = Pair(SpecifierReference.Type.Local,arrayListOf(
                Specifier("amasked", DataType.TypeUInt32, -1),
                Specifier("newmask", DataType.TypeUInt32, -1)
        ))
        val body = arrayListOf(
                assign(outputs, "q", readSpecifier(registers, "q_reg")),
                ifNoElse(andB(
                        readSpecifier(inputs, "posedge_clk_cog"),
                        readSpecifier(inputs, "ena_bus")
                ), arrayListOf(
                        assign(locals, "amasked", and32(
                                readSpecifier(inputs, "a"),
                                i32(0x3FFFF)
                        )),
                        ifNoElse(andB(
                                readSpecifier(inputs, "w"),
                                callFunc("notB",testBit32(readSpecifier(locals, "amasked"),13))
                        ), arrayListOf(
                                assign(locals,"newmask", or32(
                                        or32(numberIfBitSet(inputs,"wb",0,0xFF),numberIfBitSet(inputs,"wb",1,0xFF00)),
                                        or32(numberIfBitSet(inputs,"wb",2,0xFF0000),numberIfBitSet(inputs,"wb",3,0xFF000000L))
                                )),
                                IndexAssignmentStatement(findSpecifier(registers, "memory"), readSpecifier(locals, "amasked"),
                                        or32(
                                                and32(readSpecifier(inputs,"d"),readSpecifier(locals,"newmask")),
                                                and32(SpecifierIndexExpression(findSpecifier(registers, "memory"), readSpecifier(locals, "amasked"), DataType.TypeUInt32), callFunc("invert32", readSpecifier(locals,"newmask")))
                                        ))
                        )),
                        assign(registers, "q_reg", SpecifierIndexExpression(findSpecifier(registers, "memory"), readSpecifier(locals, "amasked"), DataType.TypeUInt32))
                ))
        )
        return ParsedModule("cloned_hub_mem", false, registers.second, arrayListOf(ParsedModuleFunction(
                inputs.second, outputs.second, locals.second, body)), ArrayList(), ArrayList())
    }
}

val baseClonedSrcPath = "../propeller-clone/"

fun runConvertToCpp() {
    val moduleNames = arrayListOf("cog_ctr","cog_vid","cog_alu","cog","hub")
    val modules = ArrayList<ParsedModule>()

    val builtInMods = BuiltInModules()
    modules.add(builtInMods.cogRam())
    modules.add(builtInMods.hubRam())

    for (modName in moduleNames)
        modules.add(UnitParser(fileLexer(modName, File("$baseClonedSrcPath/$modName/cloned_$modName.v").readText(Charsets.UTF_8)),modules).parsedResult)

    fun getModule(moduleName:String):ParsedModule {
        for (m in modules)
            if (m.name == moduleName)
                return m
        throw Exception("module '${moduleName}' not found")
    }

    val generated = HashSet<ParsedModule>()
    val destStr = StringBuilder()
    destStr.append("// Warning: Generated File! (by kotlin program: propeller-clone-translator)\n")
    destStr.append("#include \"propeller-gencpp/manualportedpropeller.h\"\n\n")
    generateCppForParsedModule(getModule("cloned_cog"),destStr,"",generated)
    generateCppForParsedModule(getModule("cloned_hub"),destStr,"", generated)

    File("/home/thilo/Projekte/Spielwiese/hive/propellersim/propeller-gencpp/generatedpropeller.gen.h").writeText(destStr.toString())
}

fun runConvertToILang() {
    val moduleNames = arrayListOf("cog_ctr","cog_vid","cog_alu","cog","hub")
    val modules = ArrayList<ParsedModule>()

    val builtInMods = BuiltInModules()
    modules.add(builtInMods.cogRam())
    modules.add(builtInMods.hubRam())
    for (modName in moduleNames)
        modules.add(UnitParser(fileLexer(modName, File("$baseClonedSrcPath/$modName/cloned_$modName.v").readText(Charsets.UTF_8)),modules).parsedResult)

    fun getModule(moduleName:String):ParsedModule {
        for (m in modules)
            if (m.name == moduleName)
                return m
        throw Exception("module '${moduleName}' not found")
    }

    val destStr = StringBuilder()

    val flattenCog = flattenModule(getModule("cloned_cog"))
    //generateILangForParsedModule(flattenCog,destStr,"")

    val flattenHub = flattenModule(getModule("cloned_hub"))
    //generateILangForParsedModule(flattenHub,destStr,"")

    destStr.append("//Diese Datei enthält die ungeänderte Ausgabe von dem propeller-clone-translator\n")
    regSetToILang(flattenHub,destStr,"","hub")
    regSetToILang(flattenCog,destStr,"","cog")
    moduleFunctionToILang("",destStr,flattenCog,flattenCog.moduleFunctions[0],0,"cogPreEval",listOf("cog"))
    moduleFunctionToILang("",destStr,flattenCog,flattenCog.moduleFunctions[1],1,"cogEvalMain",listOf("hub","cog"))
    for (f in flattenCog.pureFunctions)
        pureFunctionToILang("",destStr,f)
    for (f in flattenHub.pureFunctions)
        pureFunctionToILang("",destStr,f)
    moduleFunctionToILang("",destStr,flattenHub,flattenHub.moduleFunctions[0],0,"hubPreEval",listOf("hub"))
    moduleFunctionToILang("",destStr,flattenHub,flattenHub.moduleFunctions[1],1,"hubEvalMain",listOf("hub"))
    File("../propellersim/other-files/ilang.txt").writeText(destStr.toString())

}

fun main() {
    //runConvertToCpp()
    runConvertToILang()

    /*modules.add(ParsedModule("cloned_hub_mem",true,ArrayList(), arrayListOf(ParsedModuleFunction(
            arrayListOf(
                Specifier("posedge_clk_cog",DataType.TypeBoolean,-1),
                Specifier("ena_bus",DataType.TypeBoolean,-1),
                Specifier("w",DataType.TypeBoolean,-1),
                Specifier("wb",DataType.TypeUInt32,-1),
                Specifier("a",DataType.TypeUInt32,-1),
                Specifier("d",DataType.TypeUInt32,-1),
            ), arrayListOf(
                Specifier("q",DataType.TypeUInt32,-1)
            ),ArrayList(),ArrayList())),ArrayList(),ArrayList()))*/








    //generateCppForParsedModule(flattenCog,destStr,"",generated)
    //println(destStr)
    //simplify(getModule("cloned_cog_ctr"),1, hashMapOf("posedge_clk_cog" to BigInteger.ZERO,"posedge_clk_pll" to BigInteger.ONE))
    /*val flattenSimplifyCog = executePartialSimplify(flattenCog,1,
            IfPartialSimplify(hashMapOf("negedge_nres" to BigInteger.ZERO, "nres" to BigInteger.ONE, "negedge_ena" to BigInteger.ZERO, "ena" to BigInteger.ONE),
                    IfPartialSimplify(hashMapOf("posedge_vclk" to BigInteger.ZERO, "posedge_snc1" to BigInteger.ZERO, "negedge_ena" to BigInteger.ZERO, "posedge_clk_pll" to BigInteger.ONE),
                            IfPartialSimplify(hashMapOf("posedge_clk_cog" to BigInteger.ZERO),LeafPartialSimplify(),hashMapOf("posedge_clk_cog" to BigInteger.ONE),
                                    IfPartialSimplify(hashMapOf("state_wait" to BigInteger.ONE,"state_read_s" to BigInteger.ZERO,"state_read_d" to BigInteger.ZERO,"state_read_ni" to BigInteger.ZERO,"state_write_d" to BigInteger.ZERO), LeafPartialSimplify(), hashMapOf(),
                                            IfPartialSimplify(hashMapOf("state_wait" to BigInteger.ZERO,"state_read_s" to BigInteger.ONE,"state_read_d" to BigInteger.ZERO,"state_read_ni" to BigInteger.ZERO,"state_write_d" to BigInteger.ZERO), LeafPartialSimplify(), hashMapOf(),
                                                    IfPartialSimplify(hashMapOf("state_wait" to BigInteger.ZERO,"state_read_s" to BigInteger.ZERO,"state_read_d" to BigInteger.ONE,"state_read_ni" to BigInteger.ZERO,"state_write_d" to BigInteger.ZERO), LeafPartialSimplify(), hashMapOf(),
                                                            IfPartialSimplify(hashMapOf("state_wait" to BigInteger.ZERO,"state_read_s" to BigInteger.ZERO,"state_read_d" to BigInteger.ZERO,"state_read_ni" to BigInteger.ONE,"state_write_d" to BigInteger.ZERO), LeafPartialSimplify(), hashMapOf(),
                                                                    IfPartialSimplify(hashMapOf("state_wait" to BigInteger.ZERO,"state_read_s" to BigInteger.ZERO,"state_read_d" to BigInteger.ZERO,"state_read_ni" to BigInteger.ZERO,"state_write_d" to BigInteger.ONE), LeafPartialSimplify(), hashMapOf(),
                                                                            LeafPartialSimplify()
                                                                    )
                                                            )
                                                    )
                                            )
                                    )
                            ),
                            hashMapOf(),
                            LeafPartialSimplify()
                    ),
                    hashMapOf(),LeafPartialSimplify()
            )
    )*/

    //aktivierter cog, pll_cycle
    /*simplify(flattenCog,1, hashMapOf(
            "posedge_clk_cog" to BigInteger.ZERO,

            "negedge_nres" to BigInteger.ZERO,
            "nres" to BigInteger.ONE,
            "negedge_ena" to BigInteger.ZERO,
            "ena" to BigInteger.ONE,
            "posedge_vclk" to BigInteger.ZERO,
            "posedge_snc1" to BigInteger.ZERO,
            "negedge_ena" to BigInteger.ZERO,
            ))*/

    //cog cycle
    /*simplify(flattenCog,1, hashMapOf(
            "posedge_clk_cog" to BigInteger.ONE,
            "posedge_clk_pll" to BigInteger.ONE,
            "negedge_nres" to BigInteger.ZERO,
            "nres" to BigInteger.ONE,
            "negedge_ena" to BigInteger.ZERO,
            "ena" to BigInteger.ONE,
            "posedge_vclk" to BigInteger.ZERO,
            "posedge_snc1" to BigInteger.ZERO,
            "negedge_ena" to BigInteger.ZERO,
            "state_wait" to BigInteger.ONE,"state_read_s" to BigInteger.ZERO,"state_read_d" to BigInteger.ZERO,"state_read_ni" to BigInteger.ZERO,"state_write_d" to BigInteger.ZERO
            ))*/

    //nicht pll/cog
    /*simplify(flattenCog,1, hashMapOf(
            "posedge_clk_cog" to BigInteger.ZERO,
            "posedge_clk_pll" to BigInteger.ZERO,
    ))*/

    /*
    simplify(flattenCog,1, hashMapOf(
            "posedge_clk_cog" to BigInteger.ZERO,
            "posedge_clk_pll" to BigInteger.ZERO,
            "negedge_nres" to BigInteger.ONE,
            "nres" to BigInteger.ZERO,
    ))*/



}
