package simplify

import ConstantValueExpression
import IfStatement
import ParsedExpression
import ParsedModule
import ParsedModuleFunction
import ParsedStatement
import PureFunctionCall
import Specifier
import SpecifierExpression
import builtInFunctions
import java.math.BigInteger

//Zerlegt eine Methode in verschiedene Fälle
sealed class AbstractPartialSimplify
class IfPartialSimplify(val condition:HashMap<String,BigInteger>, val ifBranch:AbstractPartialSimplify, val elseImplicitValues:HashMap<String,BigInteger>, val elseBranch:AbstractPartialSimplify):AbstractPartialSimplify()
class LeafPartialSimplify:AbstractPartialSimplify()

fun executePartialSimplify(mod:ParsedModule, evaluationStep:Int, rootRules:AbstractPartialSimplify):ParsedModule {
    val funcs = builtInFunctions()
    fun generateConditionFromRule(ruleCond:HashMap<String,BigInteger>):ParsedExpression {
        var resultCondition:PureFunctionCall? = null
        for ((varName,varValue) in ruleCond) {
            val specRef = mod.getSpecifierReferenceByName(varName, evaluationStep)
                    ?: throw SimplifyException("var '$varName' not found")
            val specExpr = SpecifierExpression(specRef, mod.findSpecifier(specRef).dataType)
            val specCond = PureFunctionCall(funcs[when(specExpr.dataType) {
                DataType.TypeBoolean -> "eqB"
                DataType.TypeUInt32 -> "eq32"
                DataType.TypeUInt64 -> "eq64"
            }]!!, arrayListOf(specExpr,ConstantValueExpression(varValue,specExpr.dataType)))
            resultCondition = if (resultCondition==null) specCond else PureFunctionCall(funcs["andB"]!!, arrayListOf(resultCondition,specCond))
        }
        return resultCondition ?: ConstantValueExpression(BigInteger.ONE, DataType.TypeBoolean)
    }
    fun doSimplifyStep(rule:AbstractPartialSimplify, fixedVars:HashMap<String,BigInteger>):List<ParsedStatement> = when(rule) {
        is LeafPartialSimplify -> simplify(mod,evaluationStep, fixedVars)
        is IfPartialSimplify -> {
            val fixedVarsIfBranch = HashMap(fixedVars)
            fixedVarsIfBranch.putAll(rule.condition)
            val fixedVarsElseBranch = HashMap(fixedVars)
            fixedVarsElseBranch.putAll(rule.elseImplicitValues)
            arrayListOf(IfStatement(generateConditionFromRule(rule.condition),doSimplifyStep(rule.ifBranch,fixedVarsIfBranch),doSimplifyStep(rule.elseBranch,fixedVarsElseBranch)))
        }
    }
    val genCodeBody = doSimplifyStep(rootRules, HashMap())
    return ParsedModule(mod.name,false,mod.registers,mod.moduleFunctions.mapIndexed{
        //Funktion ersetzen
        index, parsedModuleFunction ->
        if (index != evaluationStep)
            parsedModuleFunction
        else
            ParsedModuleFunction(parsedModuleFunction.input, parsedModuleFunction.output, parsedModuleFunction.locals, genCodeBody)
    },mod.pureFunctions,mod.subModules)
}
