package simplify

import AssignmentStatement
import CaseStatement
import IfStatement
import IndexAssignmentStatement
import ParsedExpression
import ParsedModule
import ParsedStatement
import SpecifierExpression
import SpecifierReference
import SubModuleCallStatement
import cppgenerator.expressionToCpp

fun removeUnusedAssignments(mod:ParsedModule, evaluationIndex:Int,rootStatementList:List<ParsedStatement>):List<ParsedStatement> {
    println("Run Remove Unused Step")
    val usedAssignments = HashSet<AssignmentStatement>()
    fun markAssignmentsAsUsed(assignments:HashSet<AssignmentStatement>) {
        usedAssignments.addAll(assignments)
    }
    fun iterateStatement(stmList:List<ParsedStatement>, lastAssignedValues:VarStateMap<HashSet<AssignmentStatement>>) {
        fun searchExpression(baseExpr:ParsedExpression) = baseExpr.iterate {
            expr ->
            if (expr is SpecifierExpression)
                markAssignmentsAsUsed(lastAssignedValues.find(expr.spec))
        }
        val joinAssignmentLst = fun(lst:List<HashSet<AssignmentStatement>>):HashSet<AssignmentStatement> {
            val result = HashSet<AssignmentStatement>()
            for (l in lst)
                result.addAll(l)
            return result
        }
        for (stm in stmList) {
            when (stm) {
                is AssignmentStatement -> {
                    searchExpression(stm.expr)
                    lastAssignedValues.set(stm.dest, hashSetOf(stm))
                }
                is IndexAssignmentStatement -> {
                    searchExpression(stm.index)
                    searchExpression(stm.value)
                }
                is IfStatement -> {
                    searchExpression(stm.condition)
                    val subLastAssignedValues1 = VarStateMap(lastAssignedValues)
                    iterateStatement(stm.ifBody, subLastAssignedValues1)
                    val subLastAssignedValues2 = VarStateMap(lastAssignedValues)
                    iterateStatement(stm.elseBody, subLastAssignedValues2)
                    lastAssignedValues.joinSets(arrayListOf(subLastAssignedValues1,subLastAssignedValues2),joinAssignmentLst)
                }
                is CaseStatement -> {
                    searchExpression(stm.expresssion)
                    val subAssVals = ArrayList<VarStateMap<HashSet<AssignmentStatement>>>()
                    for (item in stm.items) {
                        val subAss = VarStateMap(lastAssignedValues)
                        iterateStatement(item.body,subAss)
                        subAssVals.add(subAss)
                    }
                    lastAssignedValues.joinSets(subAssVals,joinAssignmentLst)
                }
                is SubModuleCallStatement -> throw SimplifyException("submodules not supported")
            }
        }
    }
    val mainVarStateMap = VarStateMap<HashSet<AssignmentStatement>>(mod,evaluationIndex) {_,_ -> HashSet()}
    iterateStatement(rootStatementList, mainVarStateMap)
    //alle register und output assignemnts die zuletzt aktiv sind als verwendet markieren
    for (i in mod.registers.indices)
        markAssignmentsAsUsed(mainVarStateMap.find(SpecifierReference(-1,-1,i,SpecifierReference.Type.Register)))
    for (i in mod.moduleFunctions[evaluationIndex].output.indices)
        markAssignmentsAsUsed(mainVarStateMap.find(SpecifierReference(-1,evaluationIndex,i,SpecifierReference.Type.Output)))

    //alle assignments enfernen, die nicht verwendet wurden
    fun removeUnusedAssignments(stmList:List<ParsedStatement>):List<ParsedStatement> {
        val result = ArrayList<ParsedStatement>()
        for (stm in stmList) {
            when(stm) {
                is AssignmentStatement -> {
                    if (usedAssignments.contains(stm))
                        result.add(stm)
                    else
                        println("REMOVE Assignment: ${mod.findSpecifier(stm.dest)} = ${expressionToCpp(stm.expr) {mod.findSpecifier(it).name}}")
                }
                is IndexAssignmentStatement -> result.add(stm)
                is IfStatement -> result.add(IfStatement(stm.condition, removeUnusedAssignments(stm.ifBody), removeUnusedAssignments(stm.elseBody)))
                is CaseStatement -> {
                    val newItems = ArrayList<CaseStatement.Item>()
                    for (item in stm.items)
                        newItems.add(CaseStatement.Item(item.values, removeUnusedAssignments(item.body)))
                    result.add(CaseStatement(stm.expresssion, newItems))
                }
                is SubModuleCallStatement -> throw SimplifyException("submodule not supported")
            }
        }
        return result
    }
    return removeUnusedAssignments(rootStatementList)
}