package simplify

import ParsedModule
import Specifier
import SpecifierReference
import java.lang.Exception

class SimplifyException(message:String) : Exception(message)

class VarStateMap<T> {
    private val locals:ArrayList<T>
    private val inputs:ArrayList<T>
    private val outputs:ArrayList<T>
    private val registers:ArrayList<T>
    constructor(mod:ParsedModule, evaluationIndex:Int, cbInit:(spec:Specifier,ref:SpecifierReference)->T) {
        locals = ArrayList()
        inputs = ArrayList()
        outputs = ArrayList()
        registers = ArrayList()
        val modFun = mod.moduleFunctions[evaluationIndex]
        for ((i,r) in mod.registers.withIndex())
            registers.add(cbInit(r,SpecifierReference(-1,-1,i,SpecifierReference.Type.Register)))
        for ((i,inp) in modFun.input.withIndex())
            inputs.add(cbInit(inp,SpecifierReference(-1,evaluationIndex,i,SpecifierReference.Type.Input)))
        for ((i,outp) in modFun.output.withIndex())
            outputs.add(cbInit(outp,SpecifierReference(-1,evaluationIndex,i,SpecifierReference.Type.Output)))
        for ((i,loc) in modFun.locals.withIndex())
            locals.add(cbInit(loc,SpecifierReference(-1,evaluationIndex,i,SpecifierReference.Type.Local)))
    }
    constructor(other:VarStateMap<T>) {
        locals = ArrayList(other.locals)
        inputs = ArrayList(other.inputs)
        outputs = ArrayList(other.outputs)
        registers = ArrayList(other.registers)
    }

    fun joinSets(sets:List<VarStateMap<T>>, joinCb:(items:List<T>)->T) {
        for (i in 0 until locals.size)
            locals[i] = joinCb(sets.map { it.locals[i] })
        for (i in 0 until inputs.size)
            inputs[i] = joinCb(sets.map { it.inputs[i] })
        for (i in 0 until outputs.size)
            outputs[i] = joinCb(sets.map { it.outputs[i] })
        for (i in 0 until registers.size)
            registers[i] = joinCb(sets.map { it.registers[i] })
    }

    fun find(spec:SpecifierReference):T =when(spec.type) {
        SpecifierReference.Type.Input -> inputs[spec.listIndex]
        SpecifierReference.Type.Output -> outputs[spec.listIndex]
        SpecifierReference.Type.Local -> locals[spec.listIndex]
        SpecifierReference.Type.Register -> registers[spec.listIndex]
        SpecifierReference.Type.FromSubModule -> throw SimplifyException("illegal FromSubmodule type, no submodules supported in simplify")
    }
    fun set(spec:SpecifierReference, value:T) {
        when(spec.type) {
            SpecifierReference.Type.Input -> inputs[spec.listIndex] = value
            SpecifierReference.Type.Output -> outputs[spec.listIndex] = value
            SpecifierReference.Type.Local -> locals[spec.listIndex] = value
            SpecifierReference.Type.Register -> registers[spec.listIndex] = value
            SpecifierReference.Type.FromSubModule -> throw SimplifyException("illegal FromSubmodule type, no submodules supported in simplify")
        }
    }
}
