package assembler

import ilang.*
import java.lang.Exception
import kotlin.math.exp

//Data Expressions liefern einen Wert zurück in Registern oder in einer Speicherstelle
//Flags werden außerhalb nicht abgefragt

object AsmDataExpressionTranslator {
    /*private fun dependsOn(rootExpr:Expression, dest:AsmOperand):Boolean {
        var depends = false
        rootExpr.iterate { expr->
            if (expr is VariableExpression) {
                if (when (dest) {
                    is AsmConstantOperand, is AsmLabelOperand -> false
                    is AsmLocalOperand -> dest.belongsToVar === expr.variable
                    is AsmArrayOperand -> false //hier eine variable
                    is AsmMemOperand -> dest.variable === expr.variable
                    is AsmRamHubBaseOperand -> false
                })
                    depends = true
            }
            if (expr is ArrayAccessExpression || expr is PartialArrayAccessExpression) {
                if (when (dest) {
                        is AsmConstantOperand, is AsmLabelOperand -> false
                        is AsmLocalOperand -> false
                        is AsmArrayOperand -> dest.arrayVariable === dest.arrayVariable
                        is AsmMemOperand -> false
                        is AsmRamHubBaseOperand -> false
                    })
                    depends = true
            }
        }
        return depends
    }*/

    private fun generateBinOpExpression(state:AsmTranslatorState, dest:AsmOperand, opCode:AsmBinOp.Operation, left:Expression, right:Expression, rightCanBeAVariable:Boolean) {
        val is64Bit = left.dataType == ILangDataType.Int64
        fun transformInplace(src: Expression) {
            if (src is ConstantExpression && src.fits32Bit()) {
                if (!opCode.isNoOpWithRValue(src.value))
                    state.addBinOp(opCode, dest, state.operandOf(src), is64Bit)
                return
            }
            if (src is VariableExpression && rightCanBeAVariable)
                return state.addBinOp(opCode, dest, state.createLocalCopy(state.operandReadOf(src), dest.isMemory()), is64Bit)

            val tmp = state.nextLocal()
            generate(state, tmp, src)
            return state.addBinOp(opCode, dest, tmp, is64Bit)
        }
        fun transformNonInplace(tmpDest:AsmLocalOperand) {
            if (right is ConstantExpression && right.fits32Bit()) { //tmp = a op const
                generate(state,tmpDest,left)
                if (!opCode.isNoOpWithRValue(right.value))
                    state.addBinOp(opCode, tmpDest, state.operandOf(right), is64Bit)
                return
            }
            if (left is ConstantExpression && left.fits32Bit() && opCode.isCommutative()) { //tmp = const op a zu tmp = a op const, aber nur wenn op kommutativ ist
                generate(state, tmpDest, right)
                if (!opCode.isNoOpWithRValue(left.value))
                    state.addBinOp(opCode, tmpDest, state.operandOf(left), is64Bit)
                return
            }
            if (right is VariableExpression && rightCanBeAVariable)  { //tmp = a op var
                generate(state,tmpDest,left)
                state.addBinOp(opCode, tmpDest, state.operandReadOf(right), is64Bit)
                return
            }
            if (left is VariableExpression && opCode.isCommutative() && rightCanBeAVariable) { //tmp = var op a, aber nur wenn op kommutativ ist
                generate(state,tmpDest, right)
                state.addBinOp(opCode, tmpDest, state.operandReadOf(left), is64Bit)
                return
            }
            //if (dependsOn(right, tmpDest))
                //throw Exception("Warning depends on $tmpDest: $right")

            val tmpR = state.nextLocal()
            generate(state, tmpR, right)
            generate(state, tmpDest, left)
            state.addBinOp(opCode, tmpDest, tmpR, is64Bit)
        }
        fun checkLoadLower32(varExpr:VariableExpression,constantExpr:ConstantExpression):Boolean {
            if (constantExpr.value != 0xFFFFFFFFUL)
                return false
            val varRd = state.operandReadOf(varExpr)
            if (varRd !is AsmMemOperand)
                return false
            state.calculateInsideLocalIfDestIsMemLocation(dest) {
                //es können einfach die unteren 32 bit geladen werden
                tmpDest -> state.addMov(tmpDest, varRd, false)
            }
            return true
        }
        //spezialfälle
        if (is64Bit && opCode == AsmBinOp.Operation.And && left is VariableExpression && right is ConstantExpression && checkLoadLower32(left,right))
            return
        if (is64Bit && opCode == AsmBinOp.Operation.And && right is VariableExpression && left is ConstantExpression && checkLoadLower32(right,left))
            return
        //dest = a op r -> op a,r
        if (left is VariableExpression && state.operandInplaceWritable(left) && state.operandWriteOf(left) === dest)
            return transformInplace(right)
        //dest = l op dest -> op dest,l, aber nur wenn op kommutativ ist
        if (right is VariableExpression && state.operandInplaceWritable(right) && state.operandWriteOf(right) === dest && opCode.isCommutative())
            return transformInplace(left)
        return state.calculateInsideLocalIfDestIsMemLocation(dest) { transformNonInplace(it) }
    }

    private fun generateMuxExpression(state:AsmTranslatorState, dest:AsmOperand, expr: MuxExpression) {
        //TODO hier gibt es sehr viele möglichkeiten das effizient zu machen

        if (expr.trueExpr.dataType != ILangDataType.Bool && expr.trueExpr is ConstantExpression && expr.falseExpr is ConstantExpression) {
            //cond ? num1 : num2
            val is64Bit = expr.trueExpr.dataType == ILangDataType.Int64
            state.calculateInsideLocalIfDestIsMemLocation(dest) {
                tmpDest ->
                val tmpTrueLocal = state.nextLocal()
                state.addMov(tmpDest,state.operandOf(expr.falseExpr),is64Bit)
                state.addMov(tmpTrueLocal,state.operandOf(expr.trueExpr),is64Bit)
                val cond = AsmConditionTranslator.generate(state,expr.condition)
                state.add(AsmCMovCC(tmpDest,tmpTrueLocal,cond,is64Bit))
            }
            return
        }

        //letzte variante: wie ein if
        state.finishSectionWithBranch(expr.condition, { generate(it, dest, expr.trueExpr) }, { generate(it, dest, expr.falseExpr) })
    }

    private fun generateSetCC(state:AsmTranslatorState, dest: AsmOperand, condition: AsmCondition) {
        state.add(AsmSetCC(dest, condition))
    }

    private fun generateTestBitExpression(state:AsmTranslatorState, origDest:AsmOperand, expr:TestBitExpression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
            tmpDest ->
            generateBinOpExpression(state, tmpDest, AsmBinOp.Operation.Shr, expr.expr, expr.bitNo, false)
            state.addBinOp(AsmBinOp.Operation.And, tmpDest, state.operandOf(1UL), expr.expr.dataType == ILangDataType.Int64)
        }
    }

    private fun generateInvertExpression(state:AsmTranslatorState, origDest:AsmOperand, subExpression: Expression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
            tmpDest ->
            generate(state,tmpDest,subExpression)
            if (subExpression.dataType == ILangDataType.Bool)
                state.addBinOp(AsmBinOp.Operation.Xor,tmpDest,state.operandOf(1UL),false)
            else
                state.add(AsmInvertOp(tmpDest, subExpression.dataType == ILangDataType.Int64))
        }
    }

    private fun generateParityExpression(state:AsmTranslatorState, origDest:AsmOperand, subExpression: Expression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
            tmpDest ->
            val src = state.nextLocal()
            generate(state,src,subExpression)
            val dummyReserveReg = AsmReserveTmpCpuReg(state.nextLocal())
            state.add(dummyReserveReg)
            state.add(AsmGetParity(tmpDest, src, dummyReserveReg))
        }
    }

    private fun generateArrayAccessExpression(state:AsmTranslatorState, origDest: AsmOperand, expr:ArrayAccessExpression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
            tmpDest ->
            val idx = state.nextLocal()
            generate(state, idx, expr.indexExpr)
            if (expr.arrayVariable.dataType != ILangDataType.Int32)
                throw Exception("only Int32 supported for arrays")
            state.addMov(tmpDest,AsmArrayOperand(expr.arrayVariable,idx,4,0),false)
        }
    }

    private fun generatePartialArrayAccessExpression(state:AsmTranslatorState, origDest: AsmOperand, expr:PartialArrayAccessExpression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
                tmpDest ->
            val idx = state.nextLocal()
            generate(state, idx, expr.indexExpr)
            if (expr.arrayVariable.dataType != ILangDataType.Int32)
                throw Exception("only Int32 supported for arrays")
            state.add(AsmMovZXOp(tmpDest,AsmArrayOperand(expr.arrayVariable,idx,4, expr.byteOffset),if (expr.isByte) CpuAccessSize.Byte else CpuAccessSize.Word))
        }
    }

    private fun generateLookupExpression(state:AsmTranslatorState, origDest: AsmOperand, expr:LookupExpression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
            tmpDest ->
            val idx = state.nextLocal()
            generate(state, idx, expr.index)
            val dummyReserveReg = AsmReserveTmpCpuReg(state.nextLocal())
            state.add(dummyReserveReg)
            state.add(AsmLookupOp(
                state.addOrGetNumericLookupTable(expr.values),
                tmpDest,
                idx,
                dummyReserveReg
            ))
        }
    }

    private fun generateJumpLookupExpression(state:AsmTranslatorState, origDest: AsmOperand, expr:JumpDestinationLookupExpression) {
        state.calculateInsideLocalIfDestIsMemLocation(origDest) {
                tmpDest ->
            val idx = state.nextLocal()
            generate(state, idx, expr.index)
            val dummyReserveReg = AsmReserveTmpCpuReg(state.nextLocal())
            state.add(dummyReserveReg)
            state.add(AsmLookupOp(
                state.addOrGetJumpLookupTable(expr.values),
                tmpDest,
                idx,
                dummyReserveReg
            ))
        }
    }

    private fun generateEqualtiyExpression(state:AsmTranslatorState, dest:AsmOperand, expr:EqualityExpression) {
        val condition = if (expr.left is AdditiveOrBitwiseExpression && !expr.left.withFlags && expr.left.operation == AdditiveOrBitwiseExpression.Operation.And && expr.right is ConstantExpression && expr.right.value == 0UL) {
            //and(a,b)==0 oder and(a,b)!=0
            AsmConditionTranslator.generateCompare(state, if (expr.isEq) AsmCondition.Operation.TestZ else AsmCondition.Operation.TestNz, expr.left.left, expr.left.right)
        }
        else
            AsmConditionTranslator.generate(state, expr)
        generateSetCC(state, dest, condition)
    }

    private fun generateGetCogNumberExpression(state:AsmTranslatorState, dest:AsmOperand) {
        state.calculateInsideLocalIfDestIsMemLocation(dest) { tmpDest ->
            state.add(AsmGetCogNumber(tmpDest))
        }
    }

    fun generate(state:AsmTranslatorState, dest:AsmOperand, expr: Expression) {
        val is64Bit = expr.dataType == ILangDataType.Int64
        when (expr) {
            is ConstantExpression -> state.addMov(dest,state.operandOf(expr),is64Bit)
            is ConstantJumpDestinationExpression -> state.addMov(dest,AsmLabelOperand("(${expr.label}-startOfCode)"),false)
            is VariableExpression -> {
                state.addMov(dest,state.createLocalCopy(state.operandReadOf(expr),dest.isMemory()),is64Bit)
            }
            is AdditiveOrBitwiseExpression -> {
                if (expr.withFlags)
                    throw Exception("expression with flags not supported here: $expr")
                generateBinOpExpression(
                    state, dest, when (expr.operation) {
                        AdditiveOrBitwiseExpression.Operation.And -> AsmBinOp.Operation.And
                        AdditiveOrBitwiseExpression.Operation.Or -> AsmBinOp.Operation.Or
                        AdditiveOrBitwiseExpression.Operation.Xor -> AsmBinOp.Operation.Xor
                        AdditiveOrBitwiseExpression.Operation.Add -> AsmBinOp.Operation.Add
                        AdditiveOrBitwiseExpression.Operation.Sub -> AsmBinOp.Operation.Sub
                    },
                    expr.left,expr.right,true
                )
            }
            is ShiftExpression -> {
                if (expr.withFlags)
                    throw Exception("expression with flags not supported here: $expr")
                generateBinOpExpression(
                    state, dest,
                    if (expr.shiftLeft) AsmBinOp.Operation.Shl else AsmBinOp.Operation.Shr,
                    expr.left,expr.right,false
                )
            }
            is EqualityExpression -> generateEqualtiyExpression(state, dest, expr)
            is CompareExpression -> generateSetCC(state, dest, AsmConditionTranslator.generate(state, expr))
            is MuxExpression -> generateMuxExpression(state, dest, expr)
            is TestBitExpression -> generateTestBitExpression(state, dest, expr)
            is Lower32Expression -> {
                if (expr.expr is VariableExpression) //nur 32bit laden
                    state.addMov(dest,state.createLocalCopy(state.operandReadOf(expr.expr),dest.isMemory()),false)
                else
                    generate(state, dest, expr.expr)
            }
            is Cast64Expression -> {
                if (expr.expr is Lower32Expression)
                    throw Exception("lower32 inside cast64")
                generate(state, dest, expr.expr)
            }
            is ParityExpression -> generateParityExpression(state, dest, expr.expr)
            is InvertExpression -> {
                generateInvertExpression(state, dest, expr.expr)
            }
            is ArrayAccessExpression -> generateArrayAccessExpression(state, dest, expr)
            is PartialArrayAccessExpression -> generatePartialArrayAccessExpression(state, dest, expr)
            is LookupExpression -> generateLookupExpression(state, dest, expr)
            is JumpDestinationLookupExpression -> generateJumpLookupExpression(state, dest, expr)
            is GetCogNumberExpression -> generateGetCogNumberExpression(state, dest)
            else -> throw Exception("unsupported asm expression $expr")
        }
    }
}
