package ilang

import java.lang.Exception

//Entfernt unbenutzte Let Statements

class LetRemoverVarState {
    class State(val letStatements:HashSet<LetStatement>) {
        constructor(stm:LetStatement):this(hashSetOf(stm))
        constructor():this(HashSet())
        fun mergeWith(other:State) {
            letStatements.addAll(other.letStatements)
        }
    }
    private val map: HashMap<ILangVariable, State>
    val used:HashSet<LetStatement>

    constructor(parent: LetRemoverVarState) {
        map = HashMap(parent.map)
        used = parent.used
    }

    constructor() {
        map = HashMap()
        used = HashSet()
    }

    fun setLetStatement(s:LetStatement) {
        map[s.destVar] = State(s)
    }
    fun setUsed(v:ILangVariable) {
        used.addAll((map[v] ?: return).letStatements)
    }
    fun setUsed(e:Expression) {
        for (v in e.getUsedVariables())
            setUsed(v)
    }
    fun joinFromChildren(children:List<LetRemoverVarState>) {
        if (children.isEmpty())
            return
        val affectedVarsInChildren = HashSet<ILangVariable>() //welche variablen wurden in den zweigen modifiziert?
        for (c in children)
            affectedVarsInChildren.addAll(c.map.keys)
        for (v in affectedVarsInChildren) {
            val outerState = map[v] ?: State()
            var state = State()
            for (c in children)
                state.mergeWith(c.map[v] ?: outerState)
            map[v] = state
        }
    }
}

fun removeUnusedLetStatements(function: Function): Function {
    fun countUsage(stmList:List<Statement>, varState:LetRemoverVarState) {
        for (stm in stmList) {
            when (stm) {
                is LetStatement -> {
                    varState.setUsed(stm.expr)
                    varState.setLetStatement(stm)
                }
                is LetArrayStatement -> {
                    varState.setUsed(stm.idx)
                    varState.setUsed(stm.expr)
                }
                is SwitchStatement -> {
                    varState.setUsed(stm.expr)
                    val childStates = ArrayList<LetRemoverVarState>()
                    for (c in stm.cases) {
                        val childState = LetRemoverVarState(varState)
                        childStates.add(childState)
                        countUsage(c.body, childState)
                    }
                    varState.joinFromChildren(childStates)
                }
                is MacroStatement -> throw Exception("Macro Statement not allowed here")
            }
        }
    }

    fun markRegistersAndOutputAsUsed(varState:LetRemoverVarState) {
        //Register
        for (set in function.registerVarSet().sets) {
            for (v in set.getVariableList())
                varState.setUsed(v)
        }
        //Outputs
        for (v in function.localVarSet().getVariableList())
            if (v.direction == ILangVariableDirection.Output || v.direction == ILangVariableDirection.InOut)
                varState.setUsed(v)
    }

    //1. benutzung der let statements zählen
    val usageVarState = LetRemoverVarState()
    countUsage(function.body, usageVarState)
    markRegistersAndOutputAsUsed(usageVarState)
    //2. let statements entfernen, die nie genutzt werden
    val usedStatements = usageVarState.used
    fun doRemoveStep(stmList:List<Statement>):List<Statement> {
        val result = ArrayList<Statement>()
        for (stm in stmList) {
            when (stm) {
                is LetStatement -> if (usedStatements.contains(stm)) result.add(stm)
                is LetArrayStatement -> result.add(stm)
                is SwitchStatement -> {
                    val newCases = stm.cases.map { it.modified(doRemoveStep(it.body)) }
                    result.add(stm.modified(stm.expr, newCases))
                }
                is MacroStatement -> throw Exception("Macro Statement not allowed here")
            }
        }
        return useNewStatements(stmList, result)
    }
    val resultFun = function.modified(doRemoveStep(function.body))
    return resultFun
}
