package ilang

import kotlin.math.abs

fun muxAggressiveReduction(expr:MuxExpression):Expression {
    if (expr.trueExpr.isSameExpression(expr.falseExpr))
        return expr.trueExpr
    //optimiert a ? (a ? x : y) : z
    return expr.modified(expr.condition,
        if (expr.trueExpr is MuxExpression && expr.condition.isSameExpression(expr.trueExpr.condition)) expr.trueExpr.trueExpr else expr.trueExpr,
        if (expr.falseExpr is MuxExpression && expr.condition.isSameExpression(expr.falseExpr.condition)) expr.falseExpr.falseExpr else expr.falseExpr
    )
}

fun reduceAndExpression(subExpr:Expression, value:ULong):Expression? {
    //and(shl(expr,constA),constB) -> shl(and(expr,constB>>constA),constA
    val dt = subExpr.dataType
    if (subExpr is ShiftExpression && subExpr.left !is VariableExpression && subExpr.right is ConstantExpression && !subExpr.withFlags) {
        val shiftBy = subExpr.right
        val subSubExpr = subExpr.left
        val newAndMask = if (subExpr.shiftLeft) value shr shiftBy.value.toInt() else value shl shiftBy.value.toInt() //maske genau entgegengesezt shiften
        return ShiftExpression(dt,subExpr.shiftLeft,false,
            AdditiveOrBitwiseExpression(dt,AdditiveOrBitwiseExpression.Operation.And,false, subSubExpr, ConstantExpression(newAndMask,dt)),
            shiftBy
        )
        //println("SIMPLIFY AND: ${value.toString()}, $subExpr")
    }
    if (subExpr is AdditiveOrBitwiseExpression && subExpr.operation == AdditiveOrBitwiseExpression.Operation.Or && !subExpr.withFlags &&
        (subExpr.left is ConstantExpression || subExpr.right is ConstantExpression)) {
        //println("SIMPLIFY AND: ${value.toString()}, $subExpr")
        return AdditiveOrBitwiseExpression(dt,AdditiveOrBitwiseExpression.Operation.Or,false,
            AdditiveOrBitwiseExpression(dt,AdditiveOrBitwiseExpression.Operation.And,false,subExpr.left,ConstantExpression(value, dt)),
            AdditiveOrBitwiseExpression(dt,AdditiveOrBitwiseExpression.Operation.And,false,subExpr.right,ConstantExpression(value, dt))
        )
    }
    return null
}

fun reduceNestedShiftExpression(subSubExpr:Expression, outerShiftLeft:Boolean, shiftAmountOuter:Int, innerShiftLeft:Boolean, shiftAmountInner:Int):Expression? {
    val dt = subSubExpr.dataType
    val dtMask = dt.mask()
    var mask = dtMask
    val remainingShift = (if (innerShiftLeft) shiftAmountInner else -shiftAmountInner)+(if (outerShiftLeft) shiftAmountOuter else -shiftAmountOuter)
    mask = (if (innerShiftLeft) (mask shl shiftAmountInner) else (mask shr shiftAmountInner)) and dtMask
    mask = (if (outerShiftLeft) (mask shl shiftAmountOuter) else (mask shr shiftAmountOuter)) and dtMask
    //println("optimize nested shift: $subSubExpr, $outerShiftLeft, $shiftAmountOuter, $innerShiftLeft, $shiftAmountInner, ${mask.toString(16)}, $remainingShift")
    return AdditiveOrBitwiseExpression(dt,AdditiveOrBitwiseExpression.Operation.And,false,ShiftExpression(
        dt,(remainingShift>=0),false,subSubExpr,ConstantExpression(abs(remainingShift).toULong(),ILangDataType.Int32)),
        ConstantExpression(mask,dt)
    )
}

fun reduceEqualtiyExpression(expr:EqualityExpression):Expression {
    fun eqWithConstBool(subExpr:Expression, constV:Boolean):Expression {
        return if (constV) //e==true -> e, e!=true -> !e
            if (expr.isEq) subExpr else opNotB(subExpr)
        else //e==false -> !e, e!=false -> e
            if (expr.isEq) opNotB(subExpr) else subExpr
    }
    fun transformToTestBit(andExpr:AdditiveOrBitwiseExpression, constValue:ConstantExpression):Expression {
        val bitNo = constValue.is1ShlN()
        if (bitNo<0)
            return expr
        fun handleAnd(andSubExpr:Expression, andConst:ConstantExpression):Expression {
            if (andConst.value != constValue.value)
                return expr
            return opTestBit32(andSubExpr, opConst32(bitNo.toULong()))
        }
        if (andExpr.withFlags || andExpr.operation != AdditiveOrBitwiseExpression.Operation.And)
            return expr
        if (andExpr.right is ConstantExpression)
            return handleAnd(andExpr.left,andExpr.right)
        if (andExpr.left is ConstantExpression)
            return handleAnd(andExpr.right,andExpr.left)
        return expr
    }
    if (expr.left.dataType == ILangDataType.Bool) {
        if (expr.left is ConstantExpression)
            return eqWithConstBool(expr.right, expr.left.value != 0UL)
        if (expr.right is ConstantExpression)
            return eqWithConstBool(expr.left, expr.right.value != 0UL)
    }
    else if (expr.left.dataType == ILangDataType.Int32) {
        if (expr.left is AdditiveOrBitwiseExpression && expr.right is ConstantExpression)
            return transformToTestBit(expr.left,expr.right)
        if (expr.right is AdditiveOrBitwiseExpression && expr.left is ConstantExpression)
            return transformToTestBit(expr.right,expr.left)
    }
    return expr
}

fun reduceMuxCommonBitwiseEpxression(muxExpression: MuxExpression):Expression {
    //löst (mux cond (and e1, const) (and e2, const)) nach (and (mux cond e1 e2)) auf
    if (muxExpression.trueExpr !is AdditiveOrBitwiseExpression || muxExpression.trueExpr.withFlags)
        return muxExpression
    if (muxExpression.falseExpr !is AdditiveOrBitwiseExpression || muxExpression.falseExpr.withFlags)
        return muxExpression
    if (muxExpression.trueExpr.operation != muxExpression.falseExpr.operation)
        return muxExpression
    if (!muxExpression.trueExpr.right.isSameExpression(muxExpression.falseExpr.right))
        return muxExpression
    val repl = AdditiveOrBitwiseExpression(muxExpression.trueExpr.dataType,muxExpression.trueExpr.operation,false,
        MuxExpression(muxExpression.trueExpr.dataType,muxExpression.condition,muxExpression.trueExpr.left,muxExpression.falseExpr.left),
        muxExpression.trueExpr.right
    )
    return repl
}

fun reduceMuxCommonShiftEpxression(muxExpression: MuxExpression):Expression {
    //löst (mux cond (and e1, const) (and e2, const)) nach (and (mux cond e1 e2)) auf
    if (muxExpression.trueExpr !is ShiftExpression || muxExpression.trueExpr.withFlags)
        return muxExpression
    if (muxExpression.falseExpr !is ShiftExpression || muxExpression.falseExpr.withFlags)
        return muxExpression
    if (muxExpression.trueExpr.shiftLeft != muxExpression.falseExpr.shiftLeft)
        return muxExpression
    if (!muxExpression.trueExpr.right.isSameExpression(muxExpression.falseExpr.right))
        return muxExpression
    val repl = ShiftExpression(muxExpression.trueExpr.dataType,muxExpression.trueExpr.shiftLeft,false,
        MuxExpression(muxExpression.trueExpr.dataType,muxExpression.condition,muxExpression.trueExpr.left,muxExpression.falseExpr.left),
        muxExpression.trueExpr.right
    )
    return repl
}

fun reduceMuxCommonEpxression(muxExpression: MuxExpression):Expression {
    val r = reduceMuxCommonBitwiseEpxression(muxExpression)
    if (r !is MuxExpression)
        return r
    return reduceMuxCommonShiftEpxression(r)
}

fun reduceExpression(rootExpr:Expression, aggressive:Boolean):Expression = rootExpr.map {
    expr -> when(expr) {
        is EqualityExpression -> {
            reduceEqualtiyExpression(expr)
        }
        is MuxExpression -> {
            val m= if (expr.condition is InvertExpression) { //mux (notb c) a b -> mux c b a
                MuxExpression(expr.dataType, expr.condition.expr, expr.falseExpr, expr.trueExpr)
            }
            else
                expr
            if (aggressive) muxAggressiveReduction(m) else reduceMuxCommonEpxression(m)
        }
        is AdditiveOrBitwiseExpression -> {
            if (expr.operation == AdditiveOrBitwiseExpression.Operation.And && !expr.withFlags && expr.left is ConstantExpression)
                reduceAndExpression(expr.right, expr.left.value)?:expr
            else if (expr.operation == AdditiveOrBitwiseExpression.Operation.And && !expr.withFlags && expr.right is ConstantExpression)
                reduceAndExpression(expr.left, expr.right.value)?:expr
            else
                expr
        }
        is ShiftExpression -> {
            if (!expr.withFlags && expr.left is ShiftExpression && expr.right is ConstantExpression && !expr.left.withFlags && expr.left.right is ConstantExpression)
                reduceNestedShiftExpression(expr.left.left,expr.shiftLeft,expr.right.value.toInt(),expr.left.shiftLeft,expr.left.right.value.toInt()) ?: expr
            else
                expr
        }
        is TestBitExpression -> {
            if (expr.expr is ShiftExpression && !expr.expr.withFlags && expr.expr.right is ConstantExpression && expr.bitNo is ConstantExpression) {
                val subExpression = expr.expr.left
                val testBitNo = expr.bitNo.value
                val shiftBy = expr.expr.right.value
                val newTestBitNo = if(expr.expr.shiftLeft) {
                    if (shiftBy>testBitNo)
                        throw Exception("shift out of range")
                    testBitNo-shiftBy
                }
                else
                    testBitNo+shiftBy
                TestBitExpression(expr.expr.dataType,subExpression, opConst32(newTestBitNo))
            }
            else
                expr
        }
        else -> expr
    }
}

fun reduceStatements(rootStm:List<Statement>, aggressive: Boolean):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in rootStm) {
        when (stm) {
            is LetStatement -> {
                val expr = reduceExpression(stm.expr, aggressive)
                if (expr !is VariableExpression || expr.variable !== stm.destVar) //let a a entfernen
                    result.add(stm.modified(expr))
            }
            is LetArrayStatement -> result.add(stm.modified(reduceExpression(stm.idx, aggressive), reduceExpression(stm.expr, aggressive)))
            is SwitchStatement -> {
                val newCases = ArrayList<SwitchStatement.Case>()
                var numberOfEmptyCases = 0
                for (c in stm.cases) {
                    val nc = c.modified(reduceStatements(c.body, aggressive))
                    newCases.add(nc)
                    if (nc.body.isEmpty())
                        numberOfEmptyCases++
                }
                if (numberOfEmptyCases != newCases.size) { //kein leeres switch
                    val newExpr = reduceExpression(stm.expr, aggressive)
                    //spezialfall if notB
                    result.add(if (stm.expr.dataType == ILangDataType.Bool && newExpr is InvertExpression && newCases.size == 2) {
                        stm.modified(newExpr.expr, listOf(
                            SwitchStatement.Case(newCases[1].values,newCases[0].body),
                            SwitchStatement.Case(newCases[0].values,newCases[1].body)
                        ))
                    }
                    else
                        stm.modified(newExpr,newCases))
                }
            }
            is MacroStatement -> result.add(stm)
        }
    }
    return useNewStatements(rootStm, result)
}

fun reduceFunction(function:Function, aggressive:Boolean):Function = function.modified(reduceStatements(function.body,aggressive))