package ilang

fun simplifyFunction(function:Function, initialVarState:PartialValueVarState, aggressive:Boolean=false):Function {
    var f = function
    while (true) {
        val f2=partialEvaluateFunction(f,PartialValueVarState(initialVarState))
        val f3=removeUnusedLetStatements(f2)
        val f4=forwardLetStatements(f3, initialVarState.getNeverForwardVars(), false)
        val f5=reduceFunction(f4, aggressive)

        val lastF = f5

        if (lastF === f)
            return f
        f=lastF
    }
}

fun simplifyExpression(origExpr:Expression):Expression {
    var result = origExpr
    while (true) {
        val newR = reduceExpression(partialEvaluateExpression(result,PartialValueVarState()).expr,false)
        if (result === newR)
            return result
        result = newR
    }
}

fun removeUnsuedLocals(function:Function):Function {
    val usedVars = HashSet<ILangVariable>()
    function.iterate {
        when(it) {
            is LetStatement -> {
                usedVars.addAll(it.expr.getUsedVariables())
                usedVars.add(it.destVar)
            }
            is LetArrayStatement -> {
                usedVars.addAll(it.idx.getUsedVariables())
                usedVars.addAll(it.expr.getUsedVariables())
            }
            is SwitchStatement -> usedVars.addAll(it.expr.getUsedVariables())
            is MacroStatement -> {
                if (!it.doFunctionCall)
                    throw Exception("Macro Statement not allowed here")
                usedVars.addAll(it.inputs)
            }
        }
    }
    val prevLocals = function.localVarSet()
    val newLocals = ILangVariableSet(prevLocals.name)
    var changed = false
    for (v in prevLocals.getVariableList()) {
        if (usedVars.contains(v))
            newLocals.addVariable(v)
        else
            changed = true
    }
    if (!changed)
        return function
    for (v in prevLocals.getArrayVariableList())
        newLocals.addArrayVariable(v)
    return function.modified(newLocals)
}
