package ilang



fun partialEvaluateStatements(rootStm:List<Statement>, varState:PartialValueVarState):List<Statement> {
    val result = ArrayList<Statement>()
    for (stm in rootStm) {
        when(stm) {
            is LetStatement -> {
                val prevValue = varState.getPartialValue(stm.destVar)
                val expr = partialEvaluateExpression(stm.expr, varState)
                if (expr.expr !is ConstantExpression || !prevValue.value.fullySpecified() || prevValue.value.value != expr.expr.value)
                    result.add(stm.modified(expr.expr)) //statement nur verwenden, wenn wert sich geändert hat
                varState.setPartialValue(stm.destVar, expr, false)
            }
            is LetArrayStatement -> {
                val idx = partialEvaluateExpression(stm.idx, varState)
                val expr = partialEvaluateExpression(stm.expr, varState)
                result.add(stm.modified(idx.expr, expr.expr))
            }
            is SwitchStatement -> {
                val expr = partialEvaluateExpression(stm.expr, varState)
                if (expr.value.fullySpecified()) //wert steht fest, passendes case suchen
                    result.addAll(partialEvaluateStatements(stm.matchingCase(expr.value.value).body, varState))
                else {
                    val cases = ArrayList<SwitchStatement.Case>()
                    val subVarStates = ArrayList<PartialValueVarState>()
                    for (c in stm.cases) {
                        val subVarState = PartialValueVarState(varState)
                        //wenn switch if eine einfache variable als bedingung hat und dieses case nur einen einzeigen wert, dann
                        //ist für den teilzweig der wert der variable bekannt
                        if (expr.expr is VariableExpression && c.values.size == 1) {
                            subVarState.setPartialValue(
                                expr.expr.variable,
                                constantPartialResult(expr.expr.variable.dataType, c.values.first()), true
                            )
                        }
                        subVarStates.add(subVarState)
                        cases.add(c.modified(partialEvaluateStatements(c.body, subVarState)))
                    }
                    varState.joinFromChildren(subVarStates)
                    result.add(stm.modified(expr.expr, cases))
                }
            }
            is MacroStatement -> throw Exception("Macro Statement not allowed here")
        }
    }
    return useNewStatements(rootStm, result)
}

fun partialEvaluateFunction(function:Function, varState:PartialValueVarState):Function = function.modified(
    partialEvaluateStatements(function.body, varState))