package ilang

import special.affectedVars
import special.dependendVars

fun transformIfToMux(ifStm:SwitchStatement):Statement {
    val trueCase = ifStm.matchingCase(1UL)
    val falseCase = ifStm.matchingCase(0UL)
    if (trueCase.body.size != 1 || falseCase.body.size != 1)
        return ifStm
    val trueStm = trueCase.body.first()
    val falseStm = falseCase.body.first()
    if (trueStm !is LetStatement || falseStm !is LetStatement || trueStm.destVar !== falseStm.destVar)
        return ifStm
    val destVar = trueStm.destVar

    return LetStatement(destVar,MuxExpression(destVar.dataType,ifStm.expr,trueStm.expr,falseStm.expr))
}

fun transformIfToMux(f:Function):Function = f.modified(mapStmList(f.body) {
    if (it is SwitchStatement && it.expr.dataType == ILangDataType.Bool)
        transformIfToMux(it)
    else
        it
})

fun forceTransformIfToMux(ifStm:SwitchStatement, destVar:ILangVariable):Statement {
    fun extractValue(case:SwitchStatement.Case):Expression? {
        if (case.body.isEmpty())
            return VariableExpression(destVar)
        if (case.body.size>1)
            return null
        val firstStm = case.body.first()
        if (firstStm !is LetStatement || firstStm.destVar !== destVar)
            return null
        return firstStm.expr
    }
    val trueExpr = extractValue(ifStm.matchingCase(1UL)) ?: return ifStm
    val falseExpr = extractValue(ifStm.matchingCase(0UL)) ?: return ifStm
    return LetStatement(destVar,MuxExpression(destVar.dataType,ifStm.expr,trueExpr,falseExpr))
}

fun forceTransformIfToMux(f: Function, destVar:ILangVariable): Function = f.modified(mapStmList(f.body) {
    if (it is SwitchStatement && it.expr.dataType == ILangDataType.Bool)
        forceTransformIfToMux(it, destVar)
    else
        it
})

fun extractCommonStmInIf(ifStm: SwitchStatement):List<Statement> {
    fun allStmsAreEqual(stms:List<Statement?>):Boolean {
        if (stms.isEmpty())
            return false
        val cmpStm = stms.first() ?: return false
        for (stm in stms.drop(1))
            if (!cmpStm.isSameStatement(stm ?: return false))
                return false
        return true
    }
    val depVarsOfCondition = ifStm.expr.getUsedVariablesAndArrays()
    val commonTopStms=ArrayList<Statement>()
    val commonBotStms=ArrayList<Statement>()
    var cases = ifStm.cases
    while(true) {
        val firstStmsAreEqual = allStmsAreEqual(cases.map { it.body.firstOrNull() })
        val lastStmsAreEqual = allStmsAreEqual(cases.map { it.body.lastOrNull() })
        var modified = false
        if (firstStmsAreEqual) {
            val firstStm = cases.first().body.first()
            if (!affectedVars(firstStm).intersects(depVarsOfCondition)) { //es dürfen keine stms aus dem if nach oben geschoben werden, wenn diese variablen beeinflussen, die im if condition verwendet werden
                commonTopStms.add(firstStm)
                cases = cases.map { it.modified(it.body.drop(1)) }
                modified = true
            }
        }
        if (lastStmsAreEqual) {
            commonBotStms.add(0,cases.first().body.last())
            cases = cases.map { it.modified(it.body.dropLast(1)) }
            modified = true
        }
        if (!modified)
            return commonTopStms+listOf(ifStm.modified(ifStm.expr,cases))+commonBotStms
    }
}

fun extractCommonStmInIf(f: Function): Function = f.modified(mapStmMultiList(f.body) {
    if (it is SwitchStatement)
        extractCommonStmInIf(it)
    else
        listOf(it)
})
