import assembler.*
import cpp.generateCppCode
import ilang.*
import ilang.Function
import special.*
import java.io.File
import java.lang.Exception
import java.lang.StringBuilder

class BaseModuleImport(module: ILangModule) {
    val hubRegs = module.varSets["hub"]!!
    val cogRegs = module.varSets["cog"]!!

    val cogEvalMainClk:Function
    val cogEvalMainNoClkNegEnaRst:Function
    val cogEvalMainNoClkNegEnaNoRst:Function
    val cogEvalMainNoClkNoNegEnaNoRst:Function
    val cogPreEval:Function
    val hubPreEval:Function
    val hubEvalMain:Function
    val cppHelperFunctions:List<Function> = module.functions.values.filter { !it.name.startsWith("hub") && !it.name.startsWith("cog") }

    init {
        val zFlag = ILangVariable("zflag",ILangDataType.Bool,ILangVariableDirection.Register)
        val cFlag = ILangVariable("cflag",ILangDataType.Bool,ILangVariableDirection.Register)
        val origFlags = cogRegs.getVariable("flags")!!
        cogRegs.addVariable(zFlag)
        cogRegs.addVariable(cFlag)
        cogRegs.addVariable(ILangVariable("compound_state",ILangDataType.Int32,ILangVariableDirection.Register))
        val cogEvalMainPre0 = removeUnsuedLocals(simplifyFunction(
            replaceFlagRegister(module.functions["cogEvalMain"]!!,origFlags,zFlag,cFlag),
            PartialValueVarState()))
        val pllVars = listOf("reg_cog_ctra_pll_saved","reg_cog_ctrb_pll_saved","plla_saved","pllb_saved")
        checkNoAffectedVars(cogEvalMainPre0,pllVars)

        val pllASaved = cogEvalMainPre0.getVariableByName("plla_saved")
        val pllBSaved = cogEvalMainPre0.getVariableByName("pllb_saved")
        val cogEvalMainPre1 = replaceVariables(cogEvalMainPre0, hashMapOf(
            pllASaved to cogEvalMainPre0.getVariableByName("reg_cog_ctra_pll_saved"),
            pllBSaved to cogEvalMainPre0.getVariableByName("reg_cog_ctrb_pll_saved"),
            cogEvalMainPre0.getVariableByName("bus_q") to cogEvalMainPre0.getVariableByName("bus_q_reg"),
            cogEvalMainPre0.getVariableByName("bus_c") to cogEvalMainPre0.getVariableByName("sys_c")
        ))

        val cogEvalMain = cogEvalMainPre1

        val cogEvalMainClkPre0 = simplifyFunction(cogEvalMain,PartialValueVarState(cogEvalMain,listOf(Pair("posedge_clk_cog", true),Pair("negedge_nres", false),
            Pair("nres", true)))).rename(cogEvalMain.name+"_clk")

        //das setzen des progctr stms prüfen
        fun isSetProgCtrStm(stm:Statement):Boolean {
            if (stm !is SwitchStatement)
                return false
            val c1 = stm.matchingCase(1UL)
            if (c1.body.size != 1)
                return false
            val subStm = c1.body.first()
            if (subStm !is LetStatement || subStm.destVar.name != "progctr")
                return false
            if (subStm.expr !is VariableExpression || subStm.expr.variable.name != "ptr_d" || stm.expr.toILang() != "(andB ena_bus ptr_w)" || stm.matchingCase(0UL).body.isNotEmpty())
                throw Exception("illegal progctr stm: ${stm.expr.toILang()} $stm")
            return true
        }
        val cogEvalMainClkPre1 = moveMatchingStatementsDown(cogEvalMainClkPre0) { isSetProgCtrStm(it) }
        if (!isSetProgCtrStm(cogEvalMainClkPre1.body.last()))
            throw Exception("last stm of cogEvalMainClk is not progctr stm")
        cogEvalMainClk = cogEvalMainClkPre1.modified(cogEvalMainClkPre1.body.dropLast(1))


        val preNoClk = simplifyFunction(cogEvalMain,PartialValueVarState(cogEvalMain,listOf(Pair("posedge_clk_cog", false))))
        //ena ist 0 wenn negedge_ena==1 ist, also ist ena := ena & !negedge_ena
        val cogEvalMainNoClk = simplifyFunction(replaceVariablesByExpression(preNoClk, hashMapOf(preNoClk.getVariableByName("ena") to
                opAndB(
                    VariableExpression(preNoClk.getVariableByName("ena")),
                    InvertExpression(ILangDataType.Bool, VariableExpression(preNoClk.getVariableByName("negedge_ena")))
                ))),PartialValueVarState())

        cogEvalMainNoClkNegEnaRst = simplifyFunction(cogEvalMainNoClk, PartialValueVarState(cogEvalMainNoClk,listOf(Pair("negedge_ena", true),Pair("negedge_nres",true),Pair("nres",false)))).rename(cogEvalMain.name+"_noclk_negena_rst")
        cogEvalMainNoClkNegEnaNoRst = simplifyFunction(cogEvalMainNoClk, PartialValueVarState(cogEvalMainNoClk,listOf(Pair("negedge_ena", true),Pair("negedge_nres",false),Pair("nres",true)))).rename(cogEvalMain.name+"_noclk_negena_norst")
        cogEvalMainNoClkNoNegEnaNoRst = simplifyFunction(cogEvalMainNoClk, PartialValueVarState(cogEvalMainNoClk,listOf(Pair("negedge_ena", false),Pair("negedge_nres",false),Pair("nres",true)))).rename(cogEvalMain.name+"_noclk_nonegena")

        fun isPllSavedStm(stm:Statement) = stm is LetStatement && stm.expr is VariableExpression &&
                ((stm.destVar === pllASaved && stm.expr.variable.name == "reg_cog_ctra_pll_saved") || (stm.destVar === pllBSaved && stm.expr.variable.name == "reg_cog_ctrb_pll_saved"))
        val cogPreEvalOrig0 = moveMatchingStatementsDown(removeUnsuedLocals(simplifyFunction(replaceFlagRegister(module.functions["cogPreEval"]!!,origFlags,zFlag,cFlag),PartialValueVarState()))) { isPllSavedStm(it) }
        val cogPreEvalOrig1Stm = ArrayList(cogPreEvalOrig0.body)
        while (cogPreEvalOrig1Stm.isNotEmpty() && isPllSavedStm(cogPreEvalOrig1Stm.last()))
            cogPreEvalOrig1Stm.removeLast()
        val cogPreEvalOrig1 = cogPreEvalOrig0.modified(cogPreEvalOrig1Stm)

        cogPreEval = removeUnsuedLocals(simplifyFunction(removeOutputs(cogPreEvalOrig1,listOf("pll_out")),PartialValueVarState()))

        fun getCogRegs(lst:List<String>):Set<ILangVariable> = lst.map{ cogRegs.getVariable(it) ?: throw Exception("no cog reg $it foudn") }.toSet()
        checkForbiddenVars(cogPreEval,getCogRegs(listOf(pllASaved.name,pllBSaved.name)))

        val busSelReg = ILangVariable("bus_sel_reg",ILangDataType.Int32,ILangVariableDirection.Register)
        val busAck = ILangVariable("bus_ack_result",ILangDataType.Int32,ILangVariableDirection.Register)
        val busSelSaved = hubRegs.getVariable("bus_sel_saved")!!
        hubRegs.addVariable(busSelReg)
        hubRegs.addVariable(busAck)
        hubRegs.addVariable(ILangVariable("ena_bus_reg",ILangDataType.Bool,ILangVariableDirection.Register))
        val hubPreEvalOrig = replaceFlagRegister(module.functions["hubPreEval"]!!,origFlags,zFlag,cFlag)
        hubPreEval = removeUnsuedLocals(simplifyFunction(
            removeOutputs(replaceVariables(hubPreEvalOrig,
                hashMapOf(
                    hubPreEvalOrig.getVariableByName("bus_ack") to busAck,
                    hubPreEvalOrig.getVariableByName("bus_sel") to busSelReg,
                    busSelSaved to busSelReg
                )
                //createInOutRegReplaceMap(hubRegs, hubPreEvalOrig, listOf("bus_ack"))
            ),listOf("bus_q","bus_c","ptr_d","ptr_w","cog_ena")
            ),PartialValueVarState()))


        hubEvalMain = replaceVariables(
            replaceFlagRegister(module.functions["hubEvalMain"]!!,origFlags,zFlag,cFlag),
            hashMapOf(busSelSaved to busSelReg)
        )
        hubRegs.removeVariable(busSelSaved)



        val defaultForbidden=getCogRegs(listOf("reg_cog_ram_q_reg","match",/*"state",*/"cancel"))
        val hubInternals = listOf("cfg_reg","cog_e","cog_ena_reg","dc","lock_e","lock_state","enc","num","num_dcd","sys","sys_q","ec","wc","sc","rc","ac","sd","ad","rd","ed","reg_hub_mem_q_reg","bus_sel_reg","cog_start","bus_ack_result","ena_bus_reg").map{ hubRegs.getVariable(it) ?: throw Exception("no hub reg $it foudn") }.toSet()
        checkForbiddenVars(cogEvalMainClk,hubInternals) //Hier ist alles erlaubt
        checkForbiddenVars(cogEvalMainNoClkNegEnaRst,getCogRegs(listOf("reg_cog_ram_q_reg","match",/*"state","cancel"*/))+hubInternals) //TODO wird glaube ich nur auf 0 gesetzt
        checkForbiddenVars(cogEvalMainNoClkNegEnaNoRst,getCogRegs(listOf("reg_cog_ram_q_reg","match",/*"state","cancel"*/))+hubInternals) //TODO wird glaube ich nur auf 0 gesetzt
        checkForbiddenVars(cogEvalMainNoClkNoNegEnaNoRst,getCogRegs(listOf("reg_cog_ram_q_reg","match",/*"state",*/"cancel"))+hubInternals) //TODO hier wird noch state verwendet in bus funktionen
        checkForbiddenVars(cogPreEval,defaultForbidden+hubInternals)
        checkForbiddenVars(hubPreEval,defaultForbidden)
        checkForbiddenVars(hubEvalMain,defaultForbidden)


        val progctrVars = listOf("progctr")
        //checkNoAffectedVars(cogEvalMainClk,pllVars+progctrVars)
        checkNoAffectedVars(cogEvalMainNoClkNegEnaRst,pllVars)
        checkNoAffectedVars(cogEvalMainNoClkNegEnaNoRst,pllVars+progctrVars)
        checkNoAffectedVars(cogEvalMainNoClkNoNegEnaNoRst,pllVars+progctrVars)
        checkNoAffectedVars(cogPreEval,progctrVars+progctrVars)
        //checkNoAffectedVars(hubPreEvalOrig,pllVars)
        //checkNoAffectedVars(hubEvalMain,pllVars)


        cogRegs.removeVariable(pllASaved)
        cogRegs.removeVariable(pllBSaved)
        cogRegs.removeVariable(origFlags)

    }

    fun checkForbiddenVars(f:Function, varList:Set<ILangVariable>) {
        val inters = dependendVars(f.body).vars.union(affectedVars(f.body).vars).intersect(varList)
        if (inters.isEmpty())
            return
        println(serializeILang(f))
        throw Exception("funktion ${f.name} constains forbidden vars: $inters")
    }
    fun checkNoAffectedVars(f:Function, varList:List<String>) {
        val affVar = affectedVars(f.body).vars.intersect(varList.map{ f.getVariableByName(it) }.toSet())
        if (affVar.isEmpty())
            return
        println(serializeILang(f))
        throw Exception("funktion ${f.name} affects forbidden vars: $affVar")
    }

    fun createInOutRegReplaceMap(regSet:ILangVariableSet, f:Function, origLocalNames:List<String>):HashMap<ILangVariable, ILangVariable> {
        val result = HashMap<ILangVariable, ILangVariable>()
        for (b in origLocalNames) {
            val origVar = f.getVariableByName(b)
            val newRegisterVar = ILangVariable("${b}_wire", origVar.dataType, ILangVariableDirection.Register)
            regSet.addVariable(newRegisterVar)
            result[origVar] = newRegisterVar
        }
        return result
    }
    fun copyReplaceMapForFunction(newF:Function, origReplMap:HashMap<ILangVariable,ILangVariable>):HashMap<ILangVariable,ILangVariable> {
        val result = HashMap<ILangVariable,ILangVariable>()
        for ((oldVar,newVar) in origReplMap)
            result[newF.getVariableByName(oldVar.name)] = newVar
        return result
    }
}

fun main() {
    val baseDir = "/home/thilo/Projekte/Spielwiese/hive/propellersim/other-files"
    val module = BaseModuleImport(deserializeILang(File("$baseDir/ilang.txt").readText()))
    //input & outputs von funktionen werden durch register variablen ersetzt, hier wird die map dazu erstellt



    module.hubRegs.addVariable(ILangVariable("pinout",ILangDataType.Int32,ILangVariableDirection.Register))
    module.hubRegs.addVariable(ILangVariable("pindir",ILangDataType.Int32,ILangVariableDirection.Register))
    val origBusCogOutVars = listOf("bus_r", "bus_e", "bus_w", "bus_s", "bus_a", "bus_d")
    val pinOutVars = listOf("pin_out", "pin_dir")
    val pllInReplaceMap = module.createInOutRegReplaceMap(module.hubRegs, module.cogEvalMainNoClkNoNegEnaNoRst, listOf("pll_in"))
    val digReplaceMapA = module.createInOutRegReplaceMap(module.hubRegs, module.cogEvalMainNoClkNoNegEnaNoRst, listOf("pin_in","cnt"))
    val cogEnaReplaceMap = module.createInOutRegReplaceMap(module.cogRegs, module.cogEvalMainNoClkNoNegEnaNoRst, listOf("bus_ack"))

    val transformedHub = TransformHubFunctions(module.hubPreEval,module.hubEvalMain)
    //return


    fun createCogEvalBusSel():Function {
        //wird von dem cog aufgerufen bei dem bus_sel==1 ist, ist das bei keinem cog, dann müssen die variablen manuell gelöscht werden
        return removeUnsuedLocals(
            simplifyFunction(
                /*replaceVariables(*/removeOutputs(module.cogEvalMainNoClkNoNegEnaNoRst, pinOutVars),/*busCogOutReplaceMap),*/
                PartialValueVarState(
                    module.cogEvalMainNoClkNoNegEnaNoRst, listOf(
                        Pair("posedge_clk_pll", false),
                        Pair("posedge_vclk", false),
                        Pair("posedge_snc1", false),
                        Pair("bus_sel", true)
                    )
                )
            ).rename("cogEvalMain_bus")
        )
    }

    fun createCogEvalAsyncVid(vclk:Boolean?,snc1:Boolean?):Function {
        var name = "cogEvalMain_async"
        val extraPins = ArrayList<Pair<String,Boolean>>()
        if (vclk != null) {
            if (vclk)
                name += "_vclk"
            extraPins.add(Pair("posedge_vclk",vclk))
        }
        if (snc1 != null) {
            if (snc1)
                name += "_snc1"
            extraPins.add(Pair("posedge_snc1",snc1))
        }
        val result = removeUnsuedLocals(simplifyFunction(
            removeOutputs(module.cogEvalMainNoClkNoNegEnaNoRst, origBusCogOutVars+pinOutVars),
            PartialValueVarState(module.cogEvalMainNoClkNoNegEnaNoRst,listOf(
                Pair("posedge_clk_pll", false),
            )+extraPins)).rename(name))
        return if (snc1 == true) removeUnsuedLocals(simplifyFunction(result, PartialValueVarState(module.cogEvalMainNoClkNoNegEnaNoRst,listOf(Pair("reg_cog_vid_snc1", true))))) else result
    }

    //fPll muss einmal separat, und einmal direkt vor fCogClk aufgerufen werden
    fun createCogEvalPll():Function {
        val tmpF=removeCommonBranches(removeUnsuedLocals(simplifyFunction(
            removeOutputs(module.cogEvalMainNoClkNoNegEnaNoRst, origBusCogOutVars+pinOutVars),
            PartialValueVarState(module.cogEvalMainNoClkNoNegEnaNoRst,listOf(
                Pair("posedge_clk_pll", true),
                Pair("posedge_vclk", false),
                Pair("posedge_snc1", false)
            ))).rename("cogEvalMain_pll")))
        return tmpF
    }

    fun createCogEvalPinOut() = removeCommonBranches(removeUnsuedLocals(simplifyFunction(
        replaceVariables(removeOutputs(module.cogEvalMainNoClkNoNegEnaNoRst, origBusCogOutVars),pllInReplaceMap),
        PartialValueVarState(module.cogEvalMainNoClkNoNegEnaNoRst,listOf(
            Pair("posedge_clk_pll", false),
            Pair("posedge_vclk", false),
            Pair("posedge_snc1", false)
        ))).rename("cogEvalMain_pinout")))

    //wichtig:, bus_ack, in cogregs muss vorher gesetzt werden
    fun preoptimizeCogEvalCogClk() = removeCommonBranches(removeUnsuedLocals(simplifyFunction(
        replaceVariables(
            replaceVariables(
            removeOutputs(module.cogEvalMainClk, origBusCogOutVars+pinOutVars),digReplaceMapA),
            cogEnaReplaceMap),
        PartialValueVarState(module.cogEvalMainClk,listOf(
            Pair("posedge_clk_pll", false),
            Pair("negedge_ena", false),
            Pair("posedge_vclk", false),
            Pair("posedge_snc1", false)
        ))).rename("cogEvalMain_cogclk")))

    addCounterMaskAndVidVars(module.cogRegs)
    val preOptCogClk = preoptimizeCogEvalCogClk()
    val (cogClkFunction,cogPeripheralMacro) = findCommonMacro(MacroType.CogPeripherals,listOf(0,1,2,3,4,5).map {
        removeCommonBranches(removeUnsuedLocals(simplifyFunction(preOptCogClk, PartialValueVarState().setPartialValue(preOptCogClk.getVariableByName("state"), ILangDataType.Int32, it.toULong(), false)))).rename("cogEvalMain_cogclk_state$it")
    })

    module.cogRegs.addVariable(ILangVariable("reg_pll_saved_both",ILangDataType.Int32,ILangVariableDirection.Register))

    class CogPeripheralsFunctions(val mergedPreEvalAndPll:MergedCogPreEvalAndPll, val cogClkPeripheral:Function, val asyncVidVclk:Function, val pinOut:Function)
    val generateCogPeripheralsFunctions = CogPeripheralsFunctions(
        mergePreEvalAndPll(module.cogPreEval,createCogEvalPll()),
        transformCogClkPeripherals(cogPeripheralMacro.toFunction(),createCogEvalAsyncVid( false, true)),
        optimizeVidClkFunction(createCogEvalAsyncVid( true, false)),
        transformPinOut(createCogEvalPinOut())
    )
    module.cogRegs.removeVariable(module.cogRegs.getVariable("reg_cog_ctra_pll_saved")!!)
    module.cogRegs.removeVariable(module.cogRegs.getVariable("reg_cog_ctrb_pll_saved")!!)

    //SplittedCogPeripherals(generateCogPeripheralsFunctions.cogClkPeripheral)
    //SplittedCogPreEvalAndPll(generateCogPeripheralsFunctions.cogPreEval, generateCogPeripheralsFunctions.pll)
    //SplittedCogAsyncVid(generateCogPeripheralsFunctions.asyncVidVclk)
    //SplittedCogPinOut(generateCogPeripheralsFunctions.pinOut)
    //return

    val compoundStateManager = CompoundStateManager()
    val stateNena = compoundStateManager.reserveStateCode(CompoundState(6,CompoundState.Running.Stopped,CompoundState.Cancel.NoCancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown))
    val state0 = compoundStateManager.reserveStateCode(CompoundState(0,CompoundState.Running.Stopped,CompoundState.Cancel.NoCancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown))
    val st4Fun = CogClkState4Optimizer(compoundStateManager,cogPeripheralMacro,cogClkFunction[4])
    val st4Result = st4Fun.handleState4Function()
    CogClkOtherStateOptimizer(compoundStateManager,stateNena,state0,cogPeripheralMacro,cogClkFunction,st4Result)
    val finalCogFuncs = analyseCompoundStateTransitions(compoundStateManager,st4Result.zFlagMacros,module.cogRegs)
    val rstAndNenaFuncs = generateRstAndNenaFuncs(module.cogEvalMainNoClkNegEnaRst,module.cogEvalMainNoClkNegEnaNoRst,cogClkFunction[0],cogPeripheralMacro,origBusCogOutVars+pinOutVars, stateNena, state0)
    val busSel = BusSelOptimizerAndCompoundStateComplete(compoundStateManager,createCogEvalBusSel(),transformedHub)
    module.cogRegs.removeVariable(module.cogRegs.getVariable("reg_cog_ram_q_reg")!!)
    module.cogRegs.removeVariable(module.cogRegs.getVariable("sy_reg")!!)
    module.cogRegs.removeVariable(module.cogRegs.getVariable("compound_state")!!)

    val generatedFunctions = hashMapOf(
        "hubClkEnaBus" to transformedHub.specializedClkBusEnaVersions,
        "hubOther" to transformedHub.hubCommonMacros+listOf(
            transformedHub.hubRst,
            transformedHub.hubClkNEnaBus,
            transformedHub.hubClkPostEvalEna
        ),
        "cogClkState4" to finalCogFuncs.state4,
        "cogClkZFlagMacros" to finalCogFuncs.zMacros,
        "cogClkStateOther" to finalCogFuncs.otherState,
        "cogPeripherals" to listOf(
            generateCogPeripheralsFunctions.mergedPreEvalAndPll.mergedF,
            generateCogPeripheralsFunctions.cogClkPeripheral,
            generateCogPeripheralsFunctions.asyncVidVclk,
            generateCogPeripheralsFunctions.pinOut
        ),
        "cogOther" to listOf(transformCogControlMacroAndResetMacro(st4Result.controlMacro.toFunction(),true))+busSel.optimizedBusSel.map { it.first } +rstAndNenaFuncs,
        "cppDispatcher" to listOf(
            busSel.cppCompoundStateDispatcher,
            busSel.cppCompoundStateToState,
            busSel.cppCogBusSelDispatcher,
            busSel.cppCompoundStateToRunning,
            transformedHub.hubClkEnaBusDispatcher,
        )
    )
    module.cogRegs.removeVariable(module.cogRegs.getVariable("bus_ack_wire")!!)
    removeCounterPinAndVidVars(module.cogRegs)
    val ilangDestDir = "$baseDir/../propeller-asm/tmp"
    val asmDestDir = "$baseDir/../propeller-asm/"

    fun generateCpp() {
        File("$baseDir/../propeller-newcpp/generatedpropellernew.gen.h").writeText(generateCppCode(listOf(
            module.hubRegs,
            module.cogRegs
        ),
            generatedFunctions.values.fold(module.cppHelperFunctions){acc, list ->  acc + list}))
    }
    fun generateILang() {
        for (grp in generatedFunctions)
            File("$ilangDestDir/${grp.key}.txt").writeText(grp.value.map { serializeILang(it) }.joinToString("\n"))
    }


    fun createAsmCallFromCppStub(lst:List<AsmCodeForFunction>) {
        val tmpResAsm = StringBuilder()
        val tmpResCpp = StringBuilder()
        tmpResAsm.append(";GENERATED FILE - DO NOT EDIT\n")
        tmpResCpp.append("#ifndef PROP_ASM_CPP_CALL_STUB_H\n")
        tmpResCpp.append("#define PROP_ASM_CPP_CALL_STUB_H\n")
        tmpResCpp.append("#include <QtDebug>\n")
        tmpResCpp.append("//GENERATED FILE - DO NOT EDIT\n")
        for (f in lst) {
            val name = "asmprop_${f.name}"
            tmpResAsm.append("global $name\n")
            tmpResAsm.append("$name:\n")
            val affRegs = f.modifiedRegisters.toList().sortedBy { it }
            for (r in affRegs)
                tmpResAsm.append("    push ${r.toAsmCode(CpuAccessSize.QWord)}\n")
            tmpResAsm.append("    call ${f.name}\n")
            for (r in affRegs.reversed())
                tmpResAsm.append("    pop ${r.toAsmCode(CpuAccessSize.QWord)}\n")
            tmpResAsm.append("    ret\n\n")

            tmpResCpp.append("extern \"C\" ${if (f.hasReturnValue) "quint32" else "void"} $name(void*, void*")
            for (inp in f.inputs)
                tmpResCpp.append(", quint32")
            tmpResCpp.append(");\n")
        }
        tmpResCpp.append("#endif //PROP_ASM_CPP_CALL_STUB_H\n")

        File("$asmDestDir/asm-callstubs.gen.asm").writeText(tmpResAsm.toString())
        File("$asmDestDir/asm-callstubs.gen.h").writeText(tmpResCpp.toString())
    }

    fun createAssemblyCode(f:Function, inputs:List<AsmInputVariable>, lookupTables: ArrayList<AsmAbstractLookupTable>):AsmCodeForFunction {
        println("Translate ${f.name}")
        val tmpR = asmPreanalyse(f)
        return AsmTransformer(
            f.name,
            tmpR.readMulti,
            tmpR.readOnce.union(tmpR.written.subtract(tmpR.readMulti)).union(f.localVarSet().getVariableList()),
            inputs,
            tmpR.temporaryResultAndFlagsVariables,
            f.body,module.hubRegs,module.cogRegs,lookupTables).resultFun
    }

    fun createZFlagMacroAsm(f:Function, lookupTables: ArrayList<AsmAbstractLookupTable>):AsmCodeForFunction {
        val inputs=f.localVarSet().getVariableList().filter { it.direction == ILangVariableDirection.Input }
        val res = createAssemblyCode(f,AsmZFlagMacroRegisters.getInputMapping(inputs),lookupTables)
        if (res.modifiedRegisters.subtract(AsmZFlagMacroRegisters.destroyedRegisters.toSet()).isNotEmpty())
            throw Exception("zflag macro ${f.name} modifies too may cpu regs: ${res.modifiedRegisters}")
        return res
    }

    fun createFinalAsmCode(lst:List<AsmCodeForFunction>):String {
        val destAsmCode = StringBuilder()
        destAsmCode.append(";GENERATED FILE - DO NOT EDIT\n")
        for (code in lst) {
            destAsmCode.append(";Input rdi -> Hub Register Base Ptr\n")
            destAsmCode.append(";Input rsi -> Cog Register Base Ptr\n")
            for (inp in code.inputs)
                destAsmCode.append(";Input ${inp.cpuRegister.toString().lowercase()} -> ${inp.variable.name}\n")
            destAsmCode.append(";Destroyed: ${code.modifiedRegisters.map { it.toString().lowercase() }}\n")
            destAsmCode.append(code.name+":\n")
            destAsmCode.append(generateAsmCodeFromStream(code.code,"    "))
            destAsmCode.append("\n")
        }
        return destAsmCode.toString()
    }

    fun modifiedCpuRegisters(lst:List<AsmCodeForFunction>):Set<CpuRegister> {
        val result = HashSet<CpuRegister>()
        for (f in lst)
            result.addAll(f.modifiedRegisters)
        return result
    }

    fun createAsmPatchCode(tableName:String,patchesList: FunctionPatchesList, inputs:List<AsmInputVariable>, allowedModifiedRegs:Set<CpuRegister>, lookupTables: ArrayList<AsmAbstractLookupTable>, cogPeripheralsStr:StringBuilder, tablesStr:StringBuilder, removeLastRet:Boolean){
        fun isRet(entry:AMIEntity) = entry is AMIOpcode && entry.operation == AMIOpcode.Operation.Ret
        cogPeripheralsStr.append(";Modified Regs: $allowedModifiedRegs\n")
        tablesStr.append(";Modified Regs: $allowedModifiedRegs\n")
        for (p in patchesList.patches) {
            val code = createAssemblyCode(p, inputs, lookupTables)
            if (code.modifiedRegisters.subtract(allowedModifiedRegs).isNotEmpty())
                throw Exception("${p.name} modifies too many regs: ${code.modifiedRegisters}")
            val modCode = ArrayList(code.code)
            if (removeLastRet) {
                if (modCode.isEmpty() || !isRet(modCode.last()))
                    throw Exception("expected ret at end")
                modCode.removeLast()
                for (c in modCode)
                    if (isRet(c))
                        throw Exception("no ret allowed inside code")
            }
            cogPeripheralsStr.append(code.name+":\n")
            cogPeripheralsStr.append(generateAsmCodeFromStream(modCode,"    "))
            if (!removeLastRet)
                cogPeripheralsStr.append("        times 8 nop\n")
            cogPeripheralsStr.append("    .end:")
            cogPeripheralsStr.append("\n")
        }
        //tabelle erzeugen mit startpunkt und größe des patches
        tablesStr.append("${tableName}:\n")
        for (idx in patchesList.indices) {
            if (idx<0 || idx>=patchesList.patches.size)
                throw Exception("idx out of range")
            val p=patchesList.patches[idx]
            tablesStr.append("    dd ${tableName}-${p.name.padStart(40, ' ')}, ${p.name}.end-${p.name}\n")
        }
        tablesStr.append("\n")
    }

    fun generateAsm() {
        val lookupTables = ArrayList<AsmAbstractLookupTable>()
        val cogClkZFlagAsmCode = generatedFunctions["cogClkZFlagMacros"]!!.map { createZFlagMacroAsm(it,lookupTables) }
        val hubClkEnaBusAsmCode = generatedFunctions["hubClkEnaBus"]!!.map { createAssemblyCode(it, listOf(),lookupTables) }
        val cogClkStateOther = generatedFunctions["cogClkStateOther"]!!.map { createAssemblyCode(it, listOf(),lookupTables) }
        val cogClkState4 = generatedFunctions["cogClkState4"]!!.map { createAssemblyCode(it, listOf(),lookupTables) }
        val otherAsm = (busSel.optimizedBusSel.map { it.first }+rstAndNenaFuncs+listOf(transformedHub.hubRst)).map { createAssemblyCode(it, listOf(),lookupTables) }
        val modifiedRegsInCogClk = modifiedCpuRegisters(cogClkStateOther+cogClkState4)
        val modifiedRegsInOtherAsm = modifiedCpuRegisters(otherAsm)

        val tablesStr = StringBuilder()
        val cogPeripheralsStr = StringBuilder()
        cogPeripheralsStr.append(";GENERATED FILE - DO NOT EDIT\n")
        tablesStr.append(";GENERATED FILE - DO NOT EDIT\n")
        val splittedCogPeripherals = SplittedCogPeripherals(generateCogPeripheralsFunctions.cogClkPeripheral)
        val cogPeripheralsAllowedModRegs = setOf(CpuRegister.RAX,CpuRegister.RDX,CpuRegister.RCX)
        createAsmPatchCode("cogPeripheralsCounterAPatch",splittedCogPeripherals.counterA,listOf(),cogPeripheralsAllowedModRegs,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPeripheralsCounterBPatch",splittedCogPeripherals.counterB,listOf(),cogPeripheralsAllowedModRegs,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPeripheralsVideoPatch",splittedCogPeripherals.video,listOf(),cogPeripheralsAllowedModRegs,lookupTables,cogPeripheralsStr,tablesStr,false)

        val cogPreEvalAndPllModRegs = setOf(CpuRegister.RAX,CpuRegister.RDX,CpuRegister.RCX,CpuRegister.R8,CpuRegister.R9)
        /*val splittedCogPreEvalAndPll = SplittedCogPreEvalAndPll(generateCogPeripheralsFunctions.cogPreEval, generateCogPeripheralsFunctions.pll)
        createAsmPatchCode("cogPreEvalCounterAPatch",splittedCogPreEvalAndPll.counterAPreEval,splittedCogPreEvalAndPll.preEvalInputs,cogPreEvalAndPllModRegs,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPreEvalCounterBPatch",splittedCogPreEvalAndPll.counterBPreEval,splittedCogPreEvalAndPll.preEvalInputs,cogPreEvalAndPllModRegs,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPllCounterAPatch",splittedCogPreEvalAndPll.counterAPll,splittedCogPreEvalAndPll.pllInputs,cogPreEvalAndPllModRegs,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPllCounterBPatch",splittedCogPreEvalAndPll.counterBPll,splittedCogPreEvalAndPll.pllInputs,cogPreEvalAndPllModRegs,lookupTables,cogPeripheralsStr,tablesStr,true)
        */

        val clkVidAllowedRegs = setOf(CpuRegister.RAX,CpuRegister.RDX,CpuRegister.RCX,CpuRegister.R8,CpuRegister.R9)
        val splittedCogAsyncVid = SplittedCogAsyncVid(generateCogPeripheralsFunctions.asyncVidVclk)
        createAsmPatchCode("cogAsyncVidPatch",splittedCogAsyncVid.asyncVid,listOf(),clkVidAllowedRegs,lookupTables,cogPeripheralsStr,tablesStr,false)

        val splittedCogPinOut=SplittedCogPinOut(generateCogPeripheralsFunctions.pinOut)
        val pinoutModifiedRegisters = setOf(CpuRegister.RAX,CpuRegister.RDX,CpuRegister.RCX,CpuRegister.R8,CpuRegister.R9,CpuRegister.R10,CpuRegister.R11,CpuRegister.R12)
        createAsmPatchCode("cogPinOutCounterAPatch",splittedCogPinOut.counterA,splittedCogPinOut.inputs,pinoutModifiedRegisters,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPinOutCounterBPatch",splittedCogPinOut.counterB,splittedCogPinOut.inputs,pinoutModifiedRegisters,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPinOutVideoPatch",splittedCogPinOut.video,splittedCogPinOut.inputs,pinoutModifiedRegisters,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPinOutTopPatch",splittedCogPinOut.commonTop,splittedCogPinOut.inputs,pinoutModifiedRegisters,lookupTables,cogPeripheralsStr,tablesStr,true)
        createAsmPatchCode("cogPinOutBotPatch",splittedCogPinOut.commonBot,splittedCogPinOut.inputs,pinoutModifiedRegisters,lookupTables,cogPeripheralsStr,tablesStr,false)

        for (tbl in lookupTables) {
            tablesStr.append("${tbl.label.lbl}:\n")
            when(tbl) {
                is AsmNumericLookupTable -> tablesStr.append("    dd ${tbl.values.joinToString { "0x" + it.toString(16) }}\\n")
                is AsmJumpLookupTable -> {
                    for (x in tbl.values)
                        tablesStr.append("    dd ${x.label}-startOfCode\n")
                }
            }
        }
        tablesStr.append("hubClkEnaBusJumpTableRunningBusEFalse:\n")
        for (idx in transformedHub.runningBusEFalse)
            tablesStr.append("    dd ${hubClkEnaBusAsmCode[idx].name}-startOfCode\n")
        tablesStr.append("hubClkEnaBusJumpTableRunningBusETrue:\n")
        for (idx in transformedHub.runningBusETrue)
            tablesStr.append("    dd ${hubClkEnaBusAsmCode[idx].name}-startOfCode\n")
        /*tablesStr.append("cogClkJumpTable:\n")
        for (compoundState in compoundStateManager.getAllStates())
            tablesStr.append("    dd ${compoundState.f!!.name}-startOfCode\n")*/
        //Zu debug zwecken hier eine state decoder tabelle
        tablesStr.append("cogStateDecoderTable:\n")
        tablesStr.append("    dd ${compoundStateManager.getAllStates().size} ;number of states\n")
        for ((i,compoundState) in compoundStateManager.getAllStates().withIndex()) {
            val stateCodeInfo = i + (compoundState.state.baseState shl 16)+when(compoundState.state.running) {
                CompoundState.Running.Stopped -> 0
                CompoundState.Running.Running -> 1 shl (16+4)
            }
            tablesStr.append("    dd ${compoundState.f!!.name.padEnd(60,' ')}-startOfCode, 0x${stateCodeInfo.toString(16).padStart(8,'0')}\n")
        }

        tablesStr.append("hubStateDecoderTable:\n")
        tablesStr.append("    dd ${transformedHub.specializedClkBusEnaVersions.size} ;number of states\n")
        for ((i,hubState) in transformedHub.specializedClkBusEnaVersions.withIndex()) {
            val stateCodeInfo = i
            tablesStr.append("    dd ${hubState.name.padEnd(60,' ')}-startOfCode, 0x${stateCodeInfo.toString(16).padStart(8,'0')}\n")
        }

        File("$asmDestDir/asm-cogclk-zflagmacros.gen.asm").writeText(createFinalAsmCode(cogClkZFlagAsmCode))
        File("$asmDestDir/asm-hubclk-enabus.gen.asm").writeText(createFinalAsmCode(hubClkEnaBusAsmCode))
        File("$asmDestDir/asm-cogclk-stateother.gen.asm").writeText(createFinalAsmCode(cogClkStateOther))
        File("$asmDestDir/asm-cogclk-state4.gen.asm").writeText(createFinalAsmCode(cogClkState4))
        File("$asmDestDir/asm-other.gen.asm").writeText(createFinalAsmCode(otherAsm))
        File("$asmDestDir/asm-patches.gen.asm").writeText(cogPeripheralsStr.toString())
        File("$asmDestDir/asm-tables.gen.asm").writeText(tablesStr.toString())
        createAsmCallFromCppStub(//cogClkZFlagAsmCode+
                //hubClkEnaBusAsmCode+
                //cogClkStateOther+
                //cogClkState4+
                otherAsm+
                listOf(
                    AsmCodeForFunction("hubClkBusEna",listOf(),listOf(),modifiedCpuRegisters(hubClkEnaBusAsmCode).union(setOf(CpuRegister.RAX,CpuRegister.RDX))),
                    AsmCodeForFunction("cogClkEna",listOf(),listOf(),modifiedRegsInCogClk.union(setOf(CpuRegister.R14)).union(AsmZFlagMacroRegisters.destroyedRegisters)),
                    AsmCodeForFunction("cogPreEvalPll",listOf(),listOf(),cogPreEvalAndPllModRegs.union(clkVidAllowedRegs)),
                    AsmCodeForFunction("cogPreEvalClk",listOf(),listOf(),cogPreEvalAndPllModRegs),
                    AsmCodeForFunction("cogClkCheckVid",listOf(),listOf(),clkVidAllowedRegs),
                    AsmCodeForFunction("cogEvalPinOut",listOf(),listOf(),clkVidAllowedRegs),
                    AsmCodeForFunction("cogBusSel",listOf(),listOf(), modifiedRegsInOtherAsm.toSet().union(setOf(CpuRegister.RAX,CpuRegister.RDX))),
                    AsmCodeForFunction("updateCogPreEvalFunction",listOf(),listOf(),setOf(CpuRegister.RDX,CpuRegister.RAX,CpuRegister.RCX,CpuRegister.R8)),
                    AsmCodeForFunction("updateCogPeripheralsFunction",listOf(),listOf(), setOf(CpuRegister.RDX,CpuRegister.RAX,CpuRegister.RCX,CpuRegister.R8)),
                    AsmCodeForFunction("updateCogAsyncVid",listOf(),listOf(), setOf(CpuRegister.RDX,CpuRegister.RAX,CpuRegister.RCX,CpuRegister.R8)),
                    AsmCodeForFunction("updateCogEvalPinOut",listOf(),listOf(), setOf(CpuRegister.RDX,CpuRegister.RAX,CpuRegister.RCX,CpuRegister.R8)),
                    AsmCodeForFunction("cogUpdateAllDynamicFunctions",listOf(),listOf(), setOf(CpuRegister.RDX,CpuRegister.RAX,CpuRegister.RCX,CpuRegister.R8)),
                    AsmCodeForFunction("decodeCogStatePointer",listOf(),listOf(), setOf(CpuRegister.RCX,CpuRegister.RDX,CpuRegister.R8),true),
                    AsmCodeForFunction("decodeHubStatePointer",listOf(),listOf(), setOf(CpuRegister.RCX,CpuRegister.RDX,CpuRegister.R8),true)
                )
        )
    }

    generateILang()
    generateCpp()
    //generateAsm()


    //File("$ilangDestDir/ilang-state4MacroBottom.txt").writeText(serializeILang(extractMacro(state4MacroBottom.toFunction(), cogPeripheralMacro)))
    //File("$ilangDestDir/ilang-macroCogPeripherals.txt").writeText(serializeILang(removeCommonBranches(cogPeripheralMacro.toFunction())))
    //File("$ilangDestDir/ilang-cogEvalPinOut.txt").writeText(serializeILang(createCogEvalPinOut()))





    //println(serializeILang(extractMacro(cogClkRunning[4],cogPeripheralMacro)))


}
