package special

import ilang.*
import ilang.Function
import java.lang.Exception

class BusSelOptimizerAndCompoundStateComplete(finishedStateManager: CompoundStateManager, initialBusSelFunc:Function, transformedHub:TransformHubFunctions) {
    val optimizedBusSel:List<Pair<Function,CogBusSelVersion>>
    val cppCompoundStateDispatcher:Function
    val cppCompoundStateToState:Function
    val cppCompoundStateToRunning:Function
    val cppCogBusSelDispatcher:Function
    init {
        val compoundStateXReg = VariableExpression(initialBusSelFunc.getVariableByName("compound_statex"))
        val cogRegSet = initialBusSelFunc.registerVarSet().sets.first { it.name == "cog" }
        val hubRegSet = initialBusSelFunc.registerVarSet().sets.first { it.name == "hub" }

        fun getBusSelEntry(idx:Int)=JumpDestinationLookupExpression.Entry(idx.toULong(),transformedHub.specializedClkBusEnaVersions[idx].name)

        fun buildRunningFalse(f:Function, busW:Expression, busE:Expression, busS:Expression, busA:Expression):Function {
            if (busW !is ConstantExpression || busW.value != 0UL)
                throw Exception("bus_w must be false if running == false: $busW")
            if (busS !is ConstantExpression || busS.value != 2UL)
                throw Exception("bus_s must be 2 if running == false: $busS")
            if (busA !is AdditiveOrBitwiseExpression || busA.operation != AdditiveOrBitwiseExpression.Operation.Or || busA.withFlags)
                throw Exception("illegal bus_a expr: $busA")
            if (busA.right !is AdditiveOrBitwiseExpression || busA.right.operation != AdditiveOrBitwiseExpression.Operation.And || busA.right.withFlags || busA.right.right !is ConstantExpression || busA.right.right.value != 0x3UL)
                throw Exception("illegal bus_a right part of or expr: $busA")
            //bus_state: 0x80 | (busE ? 0x1 : 0x0) //hängt nicht von lower bits von bus_a ab
            val resultBody = ArrayList(f.body)
            val tmpBusA = f.getVariableByName("tmp_bus_a")
            resultBody.add(LetStatement(tmpBusA, partialEvaluateExpression(busA.left,PartialValueVarState().setPartialValueInitialConstraint(f.getVariableByName("progctr"),
                PartialValue(ILangDataType.Int32,0UL,0xFFFFUL)
            )).expr))
            resultBody.add(LetStatement(f.getVariableByName("bus_state"),
                opMux32(busE,
                    ConstantJumpDestinationExpression(getBusSelEntry(transformedHub.notRunningBusETrue)),
                    ConstantJumpDestinationExpression(getBusSelEntry(transformedHub.notRunningBusEFalse)),
                )
            ))
            resultBody.add(LetStatement(f.getVariableByName("bus_ashr2"), opShr32(VariableExpression(tmpBusA), opConst32(2UL))))
            return f.modified(resultBody)
        }

        fun buildRunningTrue(f:Function, busW:Expression, busE:Expression, busS:Expression, busA:Expression):Function {
            if (busS !is AdditiveOrBitwiseExpression || busS.operation != AdditiveOrBitwiseExpression.Operation.And || busS.withFlags || busS.right !is ConstantExpression || busS.right.value != 0x3UL)
                throw Exception("illegal bus_s expression if running==true: $busS")
            if (busA !is AdditiveOrBitwiseExpression || busA.operation != AdditiveOrBitwiseExpression.Operation.And || busA.withFlags || busA.right !is ConstantExpression || busA.right.value != 0xFFFFUL || busA.left !is VariableExpression)
                throw Exception("illegal bus_a expression if running==true: $busA")
            val busASrcVar = busA.left
            val resultBody = ArrayList(f.body)
            //3 bits: aclower, 2 bits sc
            val baseStateExpr = opOr32(opAnd32(busASrcVar, opConst32(0x7UL)),opShl32(busS, opConst32(3UL)))
            val stateBusETrueExpr = opOr32(baseStateExpr,opMux32(busW, opConst32(0x20UL), opConst32(0x0UL)))
            resultBody.add(LetStatement(f.getVariableByName("bus_state"), opMux32(busE,
                    JumpDestinationLookupExpression(stateBusETrueExpr,transformedHub.runningBusETrue.map { getBusSelEntry(it) }),
                    JumpDestinationLookupExpression(baseStateExpr,transformedHub.runningBusEFalse.map { getBusSelEntry(it) })
            )))
            resultBody.add(LetStatement(f.getVariableByName("bus_ashr2"), opAnd32(opShr32(busASrcVar, opConst32(2UL)), opConst32(0x3FFFUL))))
            return f.modified(resultBody)
        }

        fun reduceRunning(f:Function, running:Boolean):Function {
            var modF=simplifyFunction(f,
                PartialValueVarState(f, arrayListOf(Pair("running",running)))
                .setPartialValueInitialConstraint(f.getVariableByName("p_reg"), PartialValue(ILangDataType.Int32,0UL,0x1FFUL))
            )
            fun extractOutput(outputName:String):Expression {
                modF = moveMatchingStatementsDown(modF) { it is LetStatement && it.destVar.name == outputName }
                val lastStm = modF.body.last()
                if (lastStm !is LetStatement || lastStm.destVar.name != outputName)
                    throw Exception("unable to extract $outputName")
                modF=modF.modified(modF.body.dropLast(1))
                return lastStm.expr
            }
            val busR = extractOutput("bus_r")
            if (busR !is ConstantExpression || busR.value != if (running) 1UL else 0UL)
                throw Exception("invalid bus_r value $busR")
            val busD = extractOutput("bus_d")
            if (busD !is VariableExpression || busD.variable.name != "dest")
                throw Exception("invalid bus_d value $busR")
            val busW = extractOutput("bus_w")
            val busE = extractOutput("bus_e")
            val busS = extractOutput("bus_s")
            val busA = extractOutput("bus_a")

            return if (running) buildRunningTrue(modF,busW,busE,busS,busA) else buildRunningFalse(modF,busW,busE,busS,busA)
        }

        val newBusSelOutputs = listOf(ILangVariable("bus_state",ILangDataType.Int32,ILangVariableDirection.Output),ILangVariable("bus_ashr2",ILangDataType.Int32,ILangVariableDirection.Output))
        val allBusSelOutputs = newBusSelOutputs+listOf(initialBusSelFunc.getVariableByName("bus_d"))
        fun generateBusSel():List<Pair<Function,CogBusSelVersion>> {
            val newLocals = ILangVariableSet(initialBusSelFunc.localVarSet())
            for (v in newBusSelOutputs)
                newLocals.addVariable(v)
            newLocals.addVariable(ILangVariable("tmp_bus_a",ILangDataType.Int32,ILangVariableDirection.Local))
            val f1=initialBusSelFunc.modified(newLocals)

            val busDWire = ILangVariable("bus_d_wire",ILangDataType.Int32,ILangVariableDirection.Register)
            val busAShr2Wire = ILangVariable("bus_ashr2_wire",ILangDataType.Int32,ILangVariableDirection.Register)
            val busStateWire = ILangVariable("bus_state_wire",ILangDataType.Int32,ILangVariableDirection.Register)
            hubRegSet.addVariable(busDWire)
            hubRegSet.addVariable(busAShr2Wire)
            hubRegSet.addVariable(busStateWire)

            fun removeStateAndInstrOpcodeLower(tmpF:Function, version: CogBusSelVersion, busDRequired:Boolean):Function {
                val replaceMap = hashMapOf(
                    tmpF.getVariableByName("bus_d") to busDWire,
                    tmpF.getVariableByName("bus_ashr2") to busAShr2Wire,
                    tmpF.getVariableByName("bus_state") to busStateWire
                )
                val preRes = removeUnsuedLocals(simplifyFunction(tmpF.modified(mapExpressionsRecursive(tmpF.body) {
                    rootExpr -> rootExpr.map {
                        expr->
                        if (expr is EqualityExpression && expr.isEq && expr.left is VariableExpression && expr.left.variable.name == "state" && expr.right is ConstantExpression && expr.right.value == 5UL) {
                            when(version) {
                                CogBusSelVersion.StoppedOtherState->ConstantExpression(false)
                                CogBusSelVersion.StoppedState5->ConstantExpression(true)
                                CogBusSelVersion.RunningOtherState->ConstantExpression(false)
                                CogBusSelVersion.RunningState5AndHubOp->ConstantExpression(true)
                            }
                        }
                        else if (expr is AdditiveOrBitwiseExpression && !expr.withFlags && expr.operation == AdditiveOrBitwiseExpression.Operation.And && expr.left is VariableExpression && expr.left.variable.name == "instr_opcode" && expr.right is ConstantExpression && expr.right.value == 0x3CUL && version == CogBusSelVersion.RunningState5AndHubOp)
                            opConst32(0UL)
                        else
                            expr
                    }
                }+
                        (if (busDRequired) listOf(LetStatement(f1.getVariableByName("bus_d"),VariableExpression(f1.getVariableByName("dest")))) else listOf())
                ),PartialValueVarState()).rename(version.toLabelName()))
                return removeUnsuedLocals(replaceVariables(preRes, replaceMap))
            }

            val stmTrue = reduceRunning(f1,true)
            val stmFalse = reduceRunning(f1,false)



            return listOf(
                Pair(removeStateAndInstrOpcodeLower(stmFalse,CogBusSelVersion.StoppedOtherState,false),CogBusSelVersion.StoppedOtherState),
                Pair(removeStateAndInstrOpcodeLower(stmFalse,CogBusSelVersion.StoppedState5,false),CogBusSelVersion.StoppedState5),
                Pair(removeStateAndInstrOpcodeLower(stmTrue,CogBusSelVersion.RunningOtherState,false),CogBusSelVersion.RunningOtherState),
                Pair(removeStateAndInstrOpcodeLower(stmTrue,CogBusSelVersion.RunningState5AndHubOp,true),CogBusSelVersion.RunningState5AndHubOp)
            )
        }



        //optimizedBusSel = generateFinalBusSel(changeBusSelOutputs(initialBusSelFunc))
        optimizedBusSel = generateBusSel()

        for (v in listOf("state","match","running","cancel"))
            cogRegSet.removeVariable(initialBusSelFunc.getVariableByName(v))

        //cpp Dispatcher erstellen
        val cppDispatcherCases = finishedStateManager.getAllStates().map {
            val f = it.f ?: throw Exception("found for state not found")
            SwitchStatement.Case(listOf(it.code.toULong()),listOf(
                MacroStatement(f.name,true,f.registerVarSet(),listOf(), setOf(), listOf()) //dummy
            ))
        }

        fun generateFunction(name:String, locals:List<ILangVariable>, includeHub:Boolean, body:List<Statement>):Function {
            val localSet = ILangVariableSet("*")
            for (l in locals)
                localSet.addVariable(l)
            val setLst = arrayListOf(localSet)
            if (includeHub)
                setLst.add(hubRegSet)
            setLst.add(cogRegSet)
            return Function(name,ILangMultiVariableSet(setLst), body)
        }

        cppCompoundStateDispatcher = generateFunction("cogClkDispatchCompoundState",listOf(),true,
            listOf(SwitchStatement(compoundStateXReg,cppDispatcherCases))
        )


        val lookupStateExpr = LookupExpression(compoundStateXReg,finishedStateManager.getAllStates().map { it.state.baseState.toULong() })
        val lookupRunningEpxr = EqualityExpression(ILangDataType.Int32,true,LookupExpression(compoundStateXReg,finishedStateManager.getAllStates().map { when(it.state.running) {
            CompoundState.Running.Running -> 1UL
            CompoundState.Running.Stopped -> 0UL
        } }),ConstantExpression(1UL,ILangDataType.Int32))

        //compoundStateToState
        val state = ILangVariable("stateOutput",ILangDataType.Int32,ILangVariableDirection.Output)
        cppCompoundStateToState = generateFunction("compoundStateToState",listOf(state),false,listOf(
            LetStatement(state,lookupStateExpr)
        ))

        //compoundStateToRunning
        val running = ILangVariable("runningOutput",ILangDataType.Bool,ILangVariableDirection.Output)
        cppCompoundStateToRunning = generateFunction("compoundStateToRunning",listOf(running),false,listOf(
            LetStatement(running,lookupRunningEpxr)
        ))

        cppCogBusSelDispatcher = generateFunction("cogClkBusSelDispatcher",listOf(),true,
            listOf(SwitchStatement(VariableExpression(initialBusSelFunc.getVariableByName("bus_sel_state")),
                optimizedBusSel.map {
                    optF->SwitchStatement.Case(listOf(optF.second.toStateCode()),listOf(
                        MacroStatement(optF.second.toLabelName(),true,initialBusSelFunc.registerVarSet(),listOf(), setOf(), optF.first.localVarSet().getVariableList().filter { it.direction == ILangVariableDirection.Output }) //dummy
                    ))
                }
            ))
        )
    }
}