package special

import assembler.AsmInputVariable
import assembler.CpuRegister
import ilang.*
import ilang.Function

class FunctionPatchesList(count:Int, private val baseName:String) {
    val indices=Array<Int>(count){-1}
    val patches=ArrayList<Function>()
    fun addPatch(idx:Int,f:Function,stateName:String) {
        for ((i,p) in patches.withIndex()) {
            if (isSameStatementList(f.body,p.body)) {
                patches[i] = p.rename(p.name + "_$stateName")
                indices[idx] = i
                return
            }
        }
        indices[idx] = patches.size
        patches.add(f.rename(baseName+"_$stateName"))
    }
}

class SplittedCogPeripherals(inputF:Function) {
    val counterA:FunctionPatchesList
    val counterB:FunctionPatchesList
    val video:FunctionPatchesList
    init {
        val rawBody = ArrayList(inputF.body)
        //TODO mask statt pin
        fun extractCounter(ctraA:Boolean):FunctionPatchesList {
            val result = FunctionPatchesList(32,"cogClkCounter${if(ctraA) "A" else "B"}")
            if (rawBody.size<2)
                throw Exception("missing counter")
            val firstStm = rawBody.removeAt(0)
            val secondStm = rawBody.removeAt(0)
            val ctrModeVar = inputF.getVariableByName(if (ctraA) "reg_cog_ctra_ctrmode" else "reg_cog_ctrb_ctrmode")
            if (firstStm !is SwitchStatement || firstStm.expr !is VariableExpression || firstStm.expr.variable !== ctrModeVar)
                throw Exception("invalid first counter stm $firstStm")
            if (secondStm !is SwitchStatement || secondStm.expr !is CompareExpression || secondStm.expr.left !is VariableExpression || secondStm.expr.left.variable !== ctrModeVar)
                throw Exception("invalid second counter stm $secondStm")
            val tmpF = inputF.modified(listOf(firstStm,secondStm))
            for (ctrMode in 0 until 0x20) {
                val simplF=simplifyFunction(tmpF,PartialValueVarState().setPartialValue(ctrModeVar,ILangDataType.Int32,ctrMode.toULong(),false))
                result.addPatch(ctrMode,simplF,"M${ctrMode.toString(16)}")
            }
            return result
        }
        fun extractVideo():FunctionPatchesList {
            val result = FunctionPatchesList(4,"cogClkVideo")
            if (rawBody.isEmpty())
                throw Exception("missing video")
            val stm = rawBody.removeAt(0)
            val videoModeVar = inputF.getVariableByName("reg_cog_vid_vcfg_vmode")
            if (stm !is SwitchStatement || stm.expr !is EqualityExpression || stm.expr.left !is VariableExpression || stm.expr.left.variable !== videoModeVar)
                throw Exception("invalid video stm $stm")
            val tmpF = inputF.modified(listOf(stm))
            for (vidMode in 0 until 4) {
                val simplF=simplifyFunction(tmpF,PartialValueVarState().setPartialValue(videoModeVar,ILangDataType.Int32,vidMode.toULong(),false))
                result.addPatch(vidMode,simplF,"M${vidMode.toString(16)}")
            }
            return result
        }
        counterA = extractCounter(true)
        counterB = extractCounter(false)
        video = extractVideo()
        if (rawBody.isNotEmpty())
            throw Exception("statements after video found")
    }
}

fun transformPllSavedInCogPreEval(savedPll:ILangVariable, f:Function):Pair<LetStatement,LetStatement> {
    //val savedPll = ILangVariable("pll_saved_both",ILangDataType.Int32,ILangVariableDirection.InOut)
    //val pllCounterAIn = ILangVariable("plla_counter_in",ILangDataType.Int64,ILangVariableDirection.Input)
    //val pllCounterBIn = ILangVariable("pllb_counter_in",ILangDataType.Int64,ILangVariableDirection.Input)
    fun transformPllStm(expr:Expression, isPllA:Boolean):Expression {
        //erwartet: expr == testBit32(y,0)
        if (expr !is TestBitExpression || expr.bitNo !is ConstantExpression || expr.bitNo.value != 0UL)
            throw Exception("invalid expr: $expr")
        //erwartet: y==lower32(shr64(regPll,b))
        val y = expr.expr
        if (y !is Lower32Expression || y.expr !is ShiftExpression || y.expr.shiftLeft)
            throw Exception("invalid expr: $expr")
        val shrLeftVar = y.expr.left
        if (shrLeftVar !is VariableExpression || shrLeftVar.variable.name != (if(isPllA) "reg_cog_ctra_pllcounter" else "reg_cog_ctrb_pllcounter"))
            throw Exception("invalid expr: $shrLeftVar")
        val shrRight = y.expr.right
        if (shrRight !is AdditiveOrBitwiseExpression || shrRight.withFlags || shrRight.operation != AdditiveOrBitwiseExpression.Operation.Add || shrRight.right !is ConstantExpression || shrRight.right.value != 0x1cUL)
            throw Exception("invalid expr: $shrRight")
        val newShrLeftVar = shrLeftVar//VariableExpression(if (isPllA) pllCounterAIn else pllCounterBIn)
        return if (isPllA)
            opAnd32(opLower32(opShr64(newShrLeftVar, opAdd32(shrRight.left, opConst32(0x1cUL)))), opConst32(1UL))
        else
            opAnd32(opLower32(opShr64(newShrLeftVar, opAdd32(shrRight.left, opConst32(0x1bUL)))), opConst32(2UL))
    }
    var pllAStm:LetStatement?=null
    var pllBStm:LetStatement?=null
    for (stm in f.body) {
        if (stm is LetStatement && stm.destVar.name == "reg_cog_ctra_pll_saved")
            pllAStm=LetStatement(savedPll, opOr32(VariableExpression(savedPll),transformPllStm(stm.expr, true)))
        else if (stm is LetStatement && stm.destVar.name == "reg_cog_ctrb_pll_saved")
            pllBStm=LetStatement(savedPll, opOr32(VariableExpression(savedPll),transformPllStm(stm.expr, false)))
        else
            throw Exception("invalid stm: $stm")
    }
    return Pair(pllAStm ?: throw Exception("missing plla stm"),pllBStm?: throw Exception("missing pllb stm"))
}

fun extractCounterFromPllFunc(f:Function):Pair<SwitchStatement,SwitchStatement> {
    var stmA:SwitchStatement?=null
    var stmB:SwitchStatement?=null
    for (stm in f.body) {
        if (stm !is SwitchStatement || stm.expr !is VariableExpression)
            throw Exception("illegal stm in pll func $stm")
        if (stm.expr.variable.name == "reg_cog_ctra_ctrmode")
            stmA = stm
        else if (stm.expr.variable.name == "reg_cog_ctrb_ctrmode")
            stmB = stm
        else
            throw Exception("illegal stm in pll func $stm")
    }
    return Pair(stmA ?: throw Exception("missing ctra stm"),stmB?: throw Exception("missing ctrb stm"))
}

class MergedCogPreEvalAndPll(val savedPll:ILangVariable, val counterA:Counter, val counterB:Counter, val mergedF:Function) {
    class Counter(val preEval:LetStatement, val pll:SwitchStatement, val isCtrA:Boolean)
}

fun mergePreEvalAndPll(preEval:Function, pllFunc:Function):MergedCogPreEvalAndPll {
    val savedPll = ILangVariable("pll_saved_both",ILangDataType.Int32,ILangVariableDirection.InOut)
    val preEvalStms = transformPllSavedInCogPreEval(savedPll,preEval)
    val pllStms = extractCounterFromPllFunc(pllFunc)
    return MergedCogPreEvalAndPll(savedPll,
        MergedCogPreEvalAndPll.Counter(preEvalStms.first,pllStms.first, true),
        MergedCogPreEvalAndPll.Counter(preEvalStms.second,pllStms.second, false),
        pllFunc.localsAdded(listOf(savedPll)).modified(listOf(
            preEvalStms.first,pllStms.first,preEvalStms.second,pllStms.second
        )).rename("cogEvalPreAndPll")
    )
}

class SplittedCogPreEvalAndPll(cogPreEval:Function, cogPll:Function) {
    val preEvalInputs = listOf(
        AsmInputVariable(cogPreEval.getVariableByName("plla_counter_in"),CpuRegister.RDX,true),
        AsmInputVariable(cogPreEval.getVariableByName("pllb_counter_in"),CpuRegister.RCX,true),
        AsmInputVariable(cogPreEval.getVariableByName("pll_saved_both"),CpuRegister.RAX,true)
    )
    val pllInputs = listOf(
        AsmInputVariable(cogPll.getVariableByName("plla_counter_in"),CpuRegister.RDX,true),
        AsmInputVariable(cogPll.getVariableByName("pllb_counter_in"),CpuRegister.RCX,true),
        AsmInputVariable(cogPll.getVariableByName("pll_saved_both"),CpuRegister.RAX,true)
    )
    val counterAPreEval:FunctionPatchesList
    val counterBPreEval:FunctionPatchesList
    val counterAPll:FunctionPatchesList
    val counterBPll:FunctionPatchesList
    init {
        val rawBodyPreEval = ArrayList(cogPreEval.body)
        fun extractCounterPreEval(ctraA:Boolean):FunctionPatchesList {
            val result = FunctionPatchesList(8,"cogPreEval${if(ctraA) "A" else "B"}")
            if (rawBodyPreEval.isEmpty())
                throw Exception("illegal pre eval body")
            val first = rawBodyPreEval.removeFirst()
            if (first !is LetStatement || first.destVar.name != "pll_saved_both")
                throw Exception("illegal pre eval body $first")

            val pllDivVar = cogPreEval.getVariableByName(if (ctraA) "reg_cog_ctra_ctrplldiv" else "reg_cog_ctrb_ctrplldiv")
            val depVars = dependendVars(listOf(first)).vars.toList()
            if (!depVars.contains(pllDivVar))
                throw Exception("illegal pre eval body $first $depVars")
            for (pllDiv in 0 until 8) {
                result.addPatch(pllDiv,
                    simplifyFunction(cogPreEval.modified(listOf(first)),PartialValueVarState().setPartialValue(pllDivVar,ILangDataType.Int32,pllDiv.toULong(),false)),
                    "P${pllDiv.toString(16)}"
                )
            }
            return result
        }
        counterAPreEval=extractCounterPreEval(true)
        counterBPreEval=extractCounterPreEval(false)
        if (rawBodyPreEval.isNotEmpty())
            throw Exception("unexpected stms found")

        val rawBodyPll = ArrayList(cogPll.body)
        fun extractCounterPll(ctraA:Boolean):FunctionPatchesList {
            val result = FunctionPatchesList(32,"cogPll${if(ctraA) "A" else "B"}")
            if (rawBodyPll.isEmpty())
                throw Exception("illegal pre eval body")
            val first = rawBodyPll.removeFirst()
            val ctrModeVar = cogPll.getVariableByName(if (ctraA) "reg_cog_ctra_ctrmode" else "reg_cog_ctrb_ctrmode")
            if (first !is SwitchStatement || first.expr !is VariableExpression || first.expr.variable !== ctrModeVar)
                throw Exception("illegal pll body $first")
            for (ctrmode in 0 until 32) {
                result.addPatch(ctrmode,
                    simplifyFunction(cogPll.modified(listOf(first)),PartialValueVarState().setPartialValue(ctrModeVar,ILangDataType.Int32,ctrmode.toULong(),false)),
                    "M${ctrmode.toString(16)}"
                )
            }
            return result
        }
        counterAPll=extractCounterPll(true)
        counterBPll=extractCounterPll(false)
        if (rawBodyPll.isNotEmpty())
            throw Exception("unexpected stms found")
    }
}
