package special

import ilang.*
import ilang.Function
import java.lang.Exception

//hiermit wird ein register zu begin in eine lokale variable kopiert, und alle aufrufe bis zur Änderung des registers
//mit dieser variable geändert, damit lassen sich einige tauschoperationen für makros erledigen

fun duplicateRegisterToLocal(f:Function, regVar:ILangVariable):Function {
    val newLocals = ILangVariableSet(f.localVarSet())
    val newVar = ILangVariable(regVar.name+"_copy",regVar.dataType,ILangVariableDirection.Local)
    newLocals.addVariable(newVar)
    val replaceMap = hashMapOf(regVar to newVar)

    fun replaceInStatements(stmList:List<Statement>):Pair<Boolean, List<Statement>> { //1. argument gibt true zurück, wenn stopReplace
        val result = ArrayList<Statement>()
        var stopReplace = false
        for (stm in stmList) {
            when(stm) {
                is LetStatement -> {
                    result.add(if (stopReplace) stm else stm.modified(replaceVariablesInExpression(stm.expr,replaceMap)))
                    if (stm.destVar === regVar)
                        stopReplace = true
                }
                is LetArrayStatement -> result.add(if (stopReplace) stm else stm.modified(
                    replaceVariablesInExpression(stm.idx,replaceMap),
                    replaceVariablesInExpression(stm.expr,replaceMap)))
                is SwitchStatement -> {
                    result.add(if (stopReplace) stm else stm.modified(replaceVariablesInExpression(stm.expr,replaceMap),
                        stm.cases.map {
                            val tmpRes = replaceInStatements(it.body)
                            if (tmpRes.first)
                                stopReplace = true
                            it.modified(tmpRes.second)
                        }
                    ))
                }
                is MacroStatement -> throw Exception("no macro allowed here")
            }
        }
        return Pair(stopReplace,useNewStatements(stmList, result))
    }

    return f.modified(newLocals).modified(listOf(LetStatement(newVar,VariableExpression(regVar)))+replaceInStatements(f.body).second)
}
