package ilang

import special.*


class HubCogStartTransformer(origHubClkEna:Function, val varCogELoc:ILangVariable, val varBusSelIdxLoc:ILangVariable, val hubClkBusMacroAckCustom:MacroStatement, val hubClkBusMacroAckStartCog:MacroStatement) {
    val sysLocCVar = ILangVariable("sys_c_loc", ILangDataType.Bool, ILangVariableDirection.Local)
    val hubClkEna = origHubClkEna.localsAdded(listOf(sysLocCVar,varCogELoc,varBusSelIdxLoc))
    val numLockVar = hubClkEna.getVariableByName("num_loc")

    val alternativeLockEExpression = (parseILangStatements(
            tokenizeILang("let num_loc" +
                        " (mux32 (testBit32 lock_e 0I0)" +
                        " (mux32 (testBit32 lock_e 0I1)" +
                        " (mux32 (testBit32 lock_e 0I2)" +
                        " (mux32 (testBit32 lock_e 0I3)" +
                        " (mux32 (testBit32 lock_e 0I4)" +
                        " (mux32 (testBit32 lock_e 0I5)" +
                        " (mux32 (testBit32 lock_e 0I6) 0I7 0I6" +
                        ") 0I5" +
                        ") 0I4" +
                        ") 0I3" +
                        ") 0I2" +
                        ") 0I1" +
                        ") 0I0" +
                        ")\n"
            ), hubClkEna.allVarSets
        ).first() as? LetStatement)?.expr ?: throw Exception("parse error")
    val alternativeDC3BitSetFunc = hubClkEna.modified(
        parseILangStatements(
            tokenizeILang(
                        "let sys_c_loc (eq32 cog_e 0Iff)\n" +
                        "let sys_c sys_c_loc\n" +
                        "let num_loc" +
                        " (mux32 (testBit32 cog_e 0I0)" +
                        " (mux32 (testBit32 cog_e 0I1)" +
                        " (mux32 (testBit32 cog_e 0I2)" +
                        " (mux32 (testBit32 cog_e 0I3)" +
                        " (mux32 (testBit32 cog_e 0I4)" +
                        " (mux32 (testBit32 cog_e 0I5)" +
                        " (mux32 (testBit32 cog_e 0I6) 0I7 0I6" +
                        ") 0I5" +
                        ") 0I4" +
                        ") 0I3" +
                        ") 0I2" +
                        ") 0I1" +
                        ") 0I0" +
                        ")\n" +
                        "let num_dcd_loc (shl32 0I1 num_loc)\n" +
                        "let cog_start (mux32 sys_c_loc 0I0 (shl32 0I1 num_loc))\n" +
                        "let cog_e_loc cog_e\n" +
                        "let cog_ena_reg cog_e_loc\n" +
                        "let cog_e (or32 cog_e_loc num_dcd_loc)\n" +
                        "let sys_q num_loc\n" +
                        "let bus_q_reg_pre num_loc\n" +
                        "let bus_sel_idx_loc bus_sel_idx\n"+
                        "let bus_ack_idx (and32 (add32 bus_sel_idx_loc 0I7) 0I7)"
            ), hubClkEna.allVarSets
        ))
    val alternativeDC3BitClrFunc = hubClkEna.modified(
        parseILangStatements(
            tokenizeILang(
                "let sys_c_loc (eq32 cog_e 0Iff)\n" +
                        "let sys_c sys_c_loc\n" +
                        "let num_loc (and32 dc 0I7)\n" +
                        "let num_dcd_loc (shl32 0I1 num_loc)\n" +
                        "let cog_start (shl32 0I1 num_loc)\n" +
                        "let cog_e_loc (and32 cog_e (invert32 num_dcd_loc))\n" +
                        "let cog_ena_reg cog_e_loc\n" +
                        "let cog_e (or32 cog_e num_dcd_loc)\n" +
                        "let sys_q num_loc\n" +
                        "let bus_q_reg_pre num_loc\n" +
                        "let bus_sel_idx_loc bus_sel_idx\n"+
                        "let bus_ack_idx (and32 (add32 bus_sel_idx_loc 0I7) 0I7)"
            ), hubClkEna.allVarSets
        ))
    //In der hubClkEnaBus Funktion wird cog_start gesetzt
    //Es ist eine globale Hub variable, diese soll zu einer lokalen Variable werden
    fun transformHubCogStart(f: Function): Pair<Function,Boolean> {
        //erkennen, ob die variable statisch 0 ist
        var cogStartIsAlwaysZero = true
        f.iterate {
            if (it is LetStatement && it.destVar.name == "cog_start" && (it.expr !is ConstantExpression || it.expr.value != 0UL))
                cogStartIsAlwaysZero = false
        }
        if (cogStartIsAlwaysZero) {
            val newF = f.modified(f.body.filter { it !is LetStatement || it.destVar.name != "cog_start" })
            val affectVars = affectedVars(newF.body).vars
            if (dependendVars(newF.body).vars.union(affectVars).contains(f.getVariableByName("cog_start")))
                throw Exception("illegal stm")
            return Pair(moveMatchingStatementsUp(if (affectVars.contains(numLockVar)) transformNumLockMuxer(newF) else newF) {it is LetStatement && it.destVar.name == "sys_c"},false)
        }
        //Gibt zwei Versionen, je nachdem wie Bit 3 von ac gesetzt ist
        fun simplifyDCBit3(dcBit3: Boolean) = simplifyFunction(
            moveMatchingStatementsUp(f) { it is LetStatement && (it.destVar.name == "sys_c") },
            PartialValueVarState().setPartialValueInitialConstraint(
                f.getVariableByName("dc"),
                PartialValue(ILangDataType.Int32, if (dcBit3) 8UL else 0UL, 8UL.inv())
            )
        )

        val dc3Set = ArrayList(extractCogStartMacro(transformHubCogStartDCBit3Set(simplifyDCBit3(true))).body)
        val dc3Clr = ArrayList(extractCogStartMacro(transformHubCogStartDCBit3Clr(simplifyDCBit3(false))).body)
        /*val commonTopStm = ArrayList<Statement>()
        val commonBotStm = ArrayList<Statement>()
        while (dc3Set.isNotEmpty() && dc3Clr.isNotEmpty() && dc3Set.first().isSameStatement(dc3Clr.first())) {
            commonTopStm.add(dc3Set.first())
            dc3Set.removeFirst()
            dc3Clr.removeFirst()
        }
        while (dc3Set.isNotEmpty() && dc3Clr.isNotEmpty() && dc3Set.last().isSameStatement(dc3Clr.last())) {
            commonBotStm.add(0, dc3Set.last())
            dc3Set.removeLast()
            dc3Clr.removeLast()
        }
        fun prepareDistinctStm(
            lst: List<Statement>,
            expectedCogStartExpr: String
        ): List<Statement> { //distinct statements ohne cog_start
            val tmpF =
                moveMatchingStatementsUp(moveMatchingStatementsDown(f.modified(lst)) { it is LetStatement && it.destVar.name == "cog_start" }) { it is LetStatement && it.destVar.name == "num_loc" }
            if (tmpF.body.isEmpty())
                throw Exception("no distinct stms")
            val firstStm = tmpF.body.first()
            if (firstStm !is LetStatement || firstStm.destVar.name != "num_loc")
                throw Exception("first stm must be num_loc")
            val lastStm = tmpF.body.last()
            if (lastStm !is LetStatement || lastStm.destVar.name != "cog_start")
                throw Exception("last stm must be cog_start")
            if (lastStm.expr.toILang() != expectedCogStartExpr)
                throw Exception("expected let stm $expectedCogStartExpr, got ${lastStm.expr.toILang()}")
            return tmpF.body.dropLast(1)
        }

        val distinctDc3SetStm = prepareDistinctStm(dc3Set, "(mux32 sys_c 0I0 num_dcd_loc)")
        val distinctDc3ClrStm = prepareDistinctStm(dc3Clr, "num_dcd_loc")
        //Ein Platzhalter Makro für das starten des cogs
        //Es wird nicht wie im original eine one hot codierung übergeben, sondern der cog index*/
        val resF = removeUnsuedLocals(extractCommonStmInIf(hubClkEna).modified(listOf(
            SwitchStatement(
                TestBitExpression(
                    ILangDataType.Int32,
                    VariableExpression(f.getVariableByName("dc")),
                    ConstantExpression(3UL, ILangDataType.Int32)
                ),
                listOf(
                    SwitchStatement.Case(listOf(1UL),dc3Set),
                    SwitchStatement.Case(listOf(0UL),dc3Clr)
                )
            )
        )))
        return Pair(resF,true)
    }

    private fun extractCogStartMacro(f:Function):Function {
        fun resolveCogStart(expr:Expression):Statement {
            if (expr is ConstantExpression && expr.value == 0UL) //es wird kein cog gestartet
                return hubClkBusMacroAckCustom
            if (expr is ShiftExpression && expr.toILang() == "(shl32 0I1 num_loc)")
                return hubClkBusMacroAckStartCog
            if (expr !is MuxExpression)
                throw Exception("invalid cog start expr: $expr")
            return SwitchStatement(expr.condition,listOf(
                SwitchStatement.Case(listOf(1UL),listOf(resolveCogStart(expr.trueExpr))),
                SwitchStatement.Case(listOf(0UL),listOf(resolveCogStart(expr.falseExpr)))
            ))
        }
        val f2 = extractMacro(moveMatchingStatementsDown(moveMatchingStatementsDown(moveMatchingStatementsDown(f) { it is LetStatement && it.destVar.name == "cog_start"}) { it is LetStatement && it.destVar.name == "cog_ena_reg"}) {it is LetStatement && it.destVar.name == "bus_ack_idx"},hubClkBusMacroAckCustom)
        val remainingBody = ArrayList(f2.body)
        if (remainingBody.removeLastOrNull() !== hubClkBusMacroAckCustom)
            throw Exception("expected macro stm")
        val cogStartStm = remainingBody.removeLastOrNull()
        if (cogStartStm !is LetStatement || cogStartStm.destVar.name != "cog_start")
            throw Exception("expected cog_start stm: $cogStartStm")
        remainingBody.add(resolveCogStart(cogStartStm.expr))
        return f2.modified(remainingBody)
    }

    private fun transformHubCogStartDCBit3Set(f: Function): Function {
        val cogEVar = f.getVariableByName("cog_e")

        //Verifizieren, dass die alternative Implementierung korrekt ist
        for (cogE in 0 until 256) {
            val orig = simplifyFunction(
                f.rename(f.name + "_dc3set_orig_${cogE.toString(16)}"),
                PartialValueVarState().setPartialValue(cogEVar, ILangDataType.Int32, cogE.toULong(), false)
            )
            val alt = simplifyFunction(
                alternativeDC3BitSetFunc.rename(f.name + "_dc3set_altn_${cogE.toString(16)}"),
                PartialValueVarState().setPartialValue(cogEVar, ILangDataType.Int32, cogE.toULong(), false)
            )
            if (!isSameStatementList(orig.body, alt.body)) {
                println(serializeILang(f))
                println(serializeILang(orig))
                println(serializeILang(alt))
                throw Exception("alternative and orig implementation differ")
            }
        }
        return alternativeDC3BitSetFunc
    }

    private fun transformHubCogStartDCBit3Clr(f: Function): Function {
        val cogEVar = f.getVariableByName("cog_e")
        val dcVar = f.getVariableByName("dc")

        //verifizieren, dass die alternativimplementierung korrekt ist
        for (cogE in 0 until 256) {
            for (dc in 0 until 8) {
                val orig = simplifyFunction(
                    f.rename(f.name + "_dc3clr_orig_${dc.toString(16)}_${cogE.toString(16)}"), PartialValueVarState()
                        .setPartialValueInitialConstraint(
                            dcVar,
                            PartialValue(ILangDataType.Int32, dc.toULong(), 7UL.inv())
                        )
                        .setPartialValue(cogEVar, ILangDataType.Int32, cogE.toULong(), false)
                )
                val alt = simplifyFunction(
                    alternativeDC3BitClrFunc.rename(f.name + "_dc3clr_altn_${dc.toString(16)}_${cogE.toString(16)}"), PartialValueVarState()
                        .setPartialValueInitialConstraint(
                            dcVar,
                            PartialValue(ILangDataType.Int32, dc.toULong(), 7UL.inv())
                        )
                        .setPartialValue(cogEVar, ILangDataType.Int32, cogE.toULong(), false)
                )
                if (!isSameStatementList(orig.body, alt.body)) {
                    println(serializeILang(orig))
                    println(serializeILang(alt))
                    throw Exception("alternative and orig implementation differ")
                }
            }
        }
        return alternativeDC3BitClrFunc
    }

    //num_lock kann eine große ineffiziente mux kette haben, die wird hier umgebaut
    private fun transformNumLockMuxer(f:Function):Function {
        fun modifyNumLockStm(expr:Expression):Expression {
            if (expr.stackComplexity()<10)
                return expr
            val lockEVar = f.getVariableByName("lock_e")
            for (lockE in 0 until 256) {
                val origExpr = partialEvaluateExpression(expr,PartialValueVarState().setPartialValue(lockEVar,ILangDataType.Int32,lockE.toULong(),false))
                val altExpr = partialEvaluateExpression(alternativeLockEExpression,PartialValueVarState().setPartialValue(lockEVar,ILangDataType.Int32,lockE.toULong(),false))
                if (!origExpr.expr.isSameExpression(altExpr.expr))
                    throw Exception("mismatch expression at $lockE ${origExpr.expr}, ${altExpr.expr}")
            }
            return alternativeLockEExpression
        }
        val result = f.body.map {
            if (it is LetStatement && it.destVar === numLockVar)
                it.modified(modifyNumLockStm(it.expr))
            else
                it
        }
        val res = f.modified(result)
        //println(serializeILang(res))
        //throw Exception("TODO")
        return res
    }
}