package special
import ilang.*
import ilang.Function
import java.lang.Exception

fun transformHubMemoryDecrypt(f:Function):Function {
    val oldHubMem = f.registerVarSet().getArrayVariableByName("reg_hub_mem_memory")
    val newHubMem = f.registerVarSet().getArrayVariableByName("hub_mem_ext")
    val ac = f.getVariableByName("ac")
    val acshr2 = f.getVariableByName("acshr2")

    fun removeHubMemLoc(f:Function):Function {
        val f2=moveMatchingStatementsDown(f) {it is LetStatement && (it.destVar.name == "reg_hub_mem_q_reg_loc" || it.destVar.name == "bus_q_reg_pre") }
        if (f2.body.size<2)
            return f2
        val last=f2.body.last()
        if (last !is LetStatement || last.destVar.name != "bus_q_reg_pre" || last.expr !is VariableExpression || last.expr.variable.name != "reg_hub_mem_q_reg_loc")
            return f2
        val preLast=f2.body[f2.body.size-2]
        if (preLast !is LetStatement || last.destVar.name != "reg_hub_mem_q_reg_loc")
            return f2
        return f2.modified(f2.body.dropLast(1)+listOf(last.modified(preLast.expr)))
    }

    fun replaceHubAcAndMem(fin:Function):Function {
        fun doReplace(rootExpr:Expression):Expression = rootExpr.map {
            if (it is ShiftExpression && it.left is VariableExpression && it.left.variable === ac) {
                if (it.right !is ConstantExpression || it.right.value != 2UL)
                    throw Exception("unable to replace $it")
                VariableExpression(acshr2)
            }
            else if (it is ArrayAccessExpression && it.arrayVariable === oldHubMem)
                it.modified(newHubMem)
            else
                it

        }
        fun doReplace(lst:List<Statement>):List<Statement> {
            val result = ArrayList<Statement>()
            for (stm in lst) {
                when(stm) {
                    is LetStatement -> result.add(stm.modified(doReplace(stm.expr)))
                    is LetArrayStatement -> result.add(stm.modified(doReplace(stm.idx),doReplace(stm.expr)).modified(if (stm.destArrayVar === oldHubMem) newHubMem else stm.destArrayVar))
                    is MacroStatement -> result.add(stm)
                    is SwitchStatement -> result.add(stm.modified(doReplace(stm.expr),stm.cases.map {
                        it.modified(doReplace(it.body))
                    }))
                }
            }
            return useNewStatements(lst,result)
        }
        return fin.modified(doReplace(fin.body))
    }
    val memloc = f.localVarSet().getVariable("reg_hub_mem_q_reg_loc") ?: return replaceHubAcAndMem(f)

    val f2= moveMatchingStatementsDown(moveMatchingStatementsDown(f) { dependendVars(it).vars.contains(memloc) }) {it is LetStatement && it.destVar === memloc }


    if (f2.body.size<2)
        throw Exception("hub decrypt transform failed")
    val memAccessStm = f2.body[f2.body.size-2] as? LetStatement ?: throw Exception("hub decrypt transform failed")
    val memDecryptStm = f2.body.last() as? LetStatement ?: throw Exception("hub decrypt transform failed")

    if (memAccessStm.destVar !== memloc || memAccessStm.expr !is ArrayAccessExpression || memAccessStm.expr.arrayVariable !== oldHubMem)
        throw Exception("hub decrypt transform failed $memAccessStm")



    val f3=replaceHubAcAndMem(simplifyFunction(f2.modified(f2.body.dropLast(1)+listOf(memDecryptStm.modified(memDecryptStm.expr.map {
        expr -> if (expr is MuxExpression && expr.trueExpr is DecryptExpression && expr.falseExpr is VariableExpression) {
            if (expr.trueExpr.expr !is VariableExpression || expr.trueExpr.expr.variable !== memloc || expr.falseExpr.variable !== memloc || expr.condition.toILang() != "(testBit32 ac 0If)")
                throw Exception("illegal hub mem decrypt expression: $expr")
            ArrayAccessExpression(newHubMem, opAdd32(memAccessStm.expr.indexExpr, opAnd32(VariableExpression(acshr2),ConstantExpression(0x2000UL,ILangDataType.Int32))))
        }
        else
        expr
    }))),PartialValueVarState()))
    val f4=removeHubMemLoc(f3)
    return simplifyFunction(f4,PartialValueVarState())
}