package special
import assembler.CpuAccessSize
import ilang.*
import ilang.Function

class TransformHubFunctions(hubPreEval0:Function, hubEvalMain:Function) {

    private val hubRegs = hubEvalMain.registerVarSet().sets.first { it.name == "hub" }
    private val internalHubRegsStep1 = listOf("enc","num","num_dcd","sys").map { hubPreEval0.getVariableByName(it) }
    private val internalHubRegsStep3 = listOf("sd","ad","rd","ed").map { hubPreEval0.getVariableByName(it) }
    val hubClkEnaBusDispatcher:Function
    val hubClkNEnaBus:Function
    val hubClkPostEvalEna:Function
    val specializedClkBusEnaVersions = ArrayList<Function>()
    val notRunningBusEFalse:Int
    val notRunningBusETrue:Int
    val runningBusEFalse=Array<Int>(32) {-1} //3 bits: aclower, 2 bits sc
    val runningBusETrue=Array<Int>(64) {-1} //3 bits: aclower, 2 bits sc, 1 bit wc
    val busSelRegVar = hubEvalMain.getVariableByName("bus_sel_reg")
    val busSelIdxVar = ILangVariable("bus_sel_idx",ILangDataType.Int32,ILangVariableDirection.Register)
    val hubStateVar = ILangVariable("hub_state",ILangDataType.Int32,ILangVariableDirection.Register)
    val hubBusAckIdxVar = ILangVariable("bus_ack_idx",ILangDataType.Int32,ILangVariableDirection.Register)
    val acShr2 = ILangVariable("acshr2",ILangDataType.Int32,ILangVariableDirection.Register)
    val sysQLoc = ILangVariable("sys_q_loc",ILangDataType.Int32,ILangVariableDirection.Local)
    val cogEVar = hubEvalMain.getVariableByName("cog_e")
    val cogEnaRegVar = hubEvalMain.getVariableByName("cog_ena_reg")
    val hubCommonMacros=ArrayList<Function>()
    val hubRst:Function
    private val locDcMaskedVar = ILangVariable("dc_masked",ILangDataType.Int32,ILangVariableDirection.Local)


    private fun transformRegistersToLocals(f:Function, lst:List<ILangVariable>):Function {
        val replaceMap = HashMap<ILangVariable,ILangVariable>()
        val newLocals = ILangVariableSet(f.localVarSet())
        for (v in lst) {
            val newV = ILangVariable(v.name + "_loc", v.dataType, ILangVariableDirection.Local)
            replaceMap[v] = newV
            newLocals.addVariable(newV)
        }
        return replaceVariables(f.modified(newLocals),replaceMap)
    }

    init {
        val hubRstPre = removeUnsuedLocals(
            simplifyFunction(
                removeOutputs(transformRegistersToLocals(hubEvalMain,internalHubRegsStep1), listOf("cfg")),
                PartialValueVarState(hubEvalMain,listOf(
                    Pair("negedge_nres", true),
                    Pair("posedge_clk_cog", false)
                )
                )
            ).rename("hubRst"))
        hubRegs.addVariable(busSelIdxVar)
        hubRegs.addVariable(hubStateVar)
        hubRegs.addVariable(hubBusAckIdxVar)
        hubRegs.addVariable(acShr2)
        hubRegs.addArrayVariable(ILangArrayVariable("hub_mem_ext",ILangDataType.Int32,1024*8*3))
        //println(serializeILang(hubRst))

        val hubPreEval1 = moveMatchingStatementsDown(hubPreEval0) { it is LetStatement && it.destVar.name == "cog_start" }
        val cogStartStm = hubPreEval1.body.last() as? LetStatement ?: throw Exception("last statement of hub pre eval must be let cog_start")
        if (cogStartStm.destVar.name != "cog_start")
            throw Exception("last statement of hub pre eval must be let cog_start")
        val copyStmsFromPreEval = ArrayList<Statement>()
        for (stm in hubPreEval1.body)
            if (stm is LetStatement && internalHubRegsStep1.contains(stm.destVar))
                copyStmsFromPreEval.add(stm)
        val hubPreEval2 = removeUnsuedLocals(simplifyFunction(transformRegistersToLocals(hubPreEval1.modified(hubPreEval1.body.dropLast(1)),internalHubRegsStep1),PartialValueVarState())) //cog_start entfernen
        fun extractBusAckStmFromPreEval(f:Function):Statement {
            if (f.body.size != 1)
                throw Exception("expected exact 1 statement in hubPreEval, got ${f.body.size}")
            val stm = f.body.first()
            if (stm !is LetStatement || stm.destVar.name != "bus_ack_result")
                throw Exception("illegal first stm in hubPreEval: $stm")
            //bus_sel_reg ist ein one hot counter, in einen normalen zähler umwandeln
            val simplEdFalse = partialEvaluateExpression(stm.expr, PartialValueVarState(f, listOf(Pair("ed",false)))).expr
            if (simplEdFalse !is ConstantExpression || simplEdFalse.value != 0UL)
                throw Exception("invalid expression")
            for (cog in 0 until 8) {
                val simplEdTrue = partialEvaluateExpression(stm.expr, PartialValueVarState(f, listOf(Pair("ed",true))).setPartialValue(busSelRegVar,ILangDataType.Int32,1UL shl cog,false)).expr
                if (simplEdTrue !is ConstantExpression || simplEdTrue.value != 1UL shl ((cog+6) and 7))
                    throw Exception("expected constant expr, got $simplEdTrue")
            }
            //neue funktion bauen
            //hier wird zur cog version 7 statt 6 addiert, da das statement künftig vor der erhöhung des cog counters ausgeführt wird
            return parseILangStatements(tokenizeILang("let ${hubBusAckIdxVar.name} (mux32 ed (and32 (add32 ${busSelIdxVar.name} 0I7) 0I7) 0IF)"),f.allVarSets).first()
        }


        val busAckStm = extractBusAckStmFromPreEval(hubPreEval2)


        fun generateHubClk(enaBus:Boolean) :Function {
            val extendedBody = copyStmsFromPreEval+(if (enaBus) listOf(cogStartStm) else listOf())+hubEvalMain.body
            return removeUnsuedLocals(
                simplifyFunction(
                    removeOutputs(transformRegistersToLocals(hubEvalMain.modified(extendedBody),internalHubRegsStep1), listOf("cfg")),
                    PartialValueVarState(hubEvalMain,listOf(
                        Pair("negedge_nres", false),
                        Pair("posedge_clk_cog", true),
                        Pair("nres", true),
                        Pair("ena_bus",enaBus)
                    ))
                ))
        }
        val preHubClkNEnaBus = generateHubClk(false).rename("hubClkBusNena")
        fun isBusAssignemtnStm(stm:Statement) = stm is LetStatement && stm.expr is VariableExpression && stm.expr.variable.direction == ILangVariableDirection.Input
        val hubClkEna0 = moveMatchingStatementsDown(generateHubClk(true)) { isBusAssignemtnStm(it) }
        val hubClkEnaBusBody = ArrayList(hubClkEna0.body)
        val expectedHubClkPostEvalBody = hashMapOf("ec" to "bus_e", "rc" to "bus_r", "wc" to "bus_w", "sc" to "bus_s", "dc" to "bus_d", "ac" to "bus_a")
        while (hubClkEnaBusBody.isNotEmpty() && isBusAssignemtnStm(hubClkEnaBusBody.last())) {
            val tmpStm = hubClkEnaBusBody.last() as? LetStatement ?: throw Exception("let stm expected (internal error)")
            val expectedExpr = expectedHubClkPostEvalBody.remove(tmpStm.destVar.name)
            if (expectedExpr == null || tmpStm.expr !is VariableExpression || tmpStm.expr.variable.name != expectedExpr)
                throw Exception("invalid post hub statement $tmpStm $expectedExpr")
            hubClkEnaBusBody.removeLast()
        }
        //macro extraktion vorbereiten (2x hochbewegen, da lokale variablen im weg sind)
        fun isMemoryMacroStm(stm:Statement) = (stm is LetStatement && stm.destVar.name == "ramq") || (stm is SwitchStatement && stm.expr.toILang() == "(and32 sd 0I3)")
        val preHubClkEnaBus = moveMatchingStatementsUp(removeUnsuedLocals(
            simplifyFunction(moveMatchingStatementsUp(hubClkEna0.modified(hubClkEnaBusBody)) { isMemoryMacroStm(it) },PartialValueVarState())
        )){ isMemoryMacroStm(it) }.rename("hubClkBusEna")


        val busQRegVar = preHubClkNEnaBus.getVariableByName("bus_q_reg")
        val busQRegPreVar = ILangVariable("bus_q_reg_pre",ILangDataType.Int32,ILangVariableDirection.Register)
        hubRegs.addVariable(busQRegPreVar)

        val memQReg = preHubClkNEnaBus.getVariableByName("reg_hub_mem_q_reg")
        val memQRegLoc = ILangVariable("reg_hub_mem_q_reg_loc",ILangDataType.Int32,ILangVariableDirection.Local)

        val sysQReg = preHubClkNEnaBus.getVariableByName("sys_q")
        //val sysQRegLoc = ILangVariable("sys_q_loc",ILangDataType.Int32,ILangVariableDirection.Local)

        val (newFun,memoryMacro) = findCommonMacro(MacroType.HubMemory,listOf(preHubClkNEnaBus,preHubClkEnaBus),false,false)
        val memoryMacroFunc = replaceVariables(memoryMacro.toFunction(), hashMapOf(busQRegVar to busQRegPreVar))

        //wenn der Bus nicht aktiv ist, darf nur das memoryMakro vorhanden sein
        val hubClkNEnaBusPre = extractMacro(newFun[0],memoryMacro)
        if (hubClkNEnaBusPre.body.size != 1 || hubClkNEnaBusPre.body.first() !is MacroStatement)
            throw Exception("invalid hubClkNenaBus function")
        //das memory makro wird ersetzt durch eine zuweisung aus einem neuen pre register, da der zugriff schon auf letzten takt verschoben wurde
        hubClkNEnaBus = removeUnsuedLocals(hubClkNEnaBusPre.modified(listOf(LetStatement(busQRegVar,VariableExpression(busQRegPreVar)))))
        val hubClkEnaBus = newFun[1]

        //die variable bus_sel_reg ist ein one hot counter und soll hier ersetzt werden
        fun handleSpecialBusEnaBusSel(f:Function):Function {
            val tmpBusAckIdx = ILangVariable("tmp_bus_ack_idx",ILangDataType.Int32,ILangVariableDirection.Local)
            val tmpF = ArrayList<Function>()
            for (cog in 0 until 8)
                tmpF.add(simplifyFunction(f,PartialValueVarState().setPartialValue(busSelRegVar,ILangDataType.Int32,1UL shl cog, false)))
            val bodySize = tmpF[0].body.size
            for (cog in 1 until 8)
                if (tmpF[cog].body.size != bodySize)
                    throw Exception("mismatch body size")
            val resultBody = ArrayList<Statement>()
            resultBody.add(LetStatement(
                tmpBusAckIdx, opAnd32(
                    opSub32(VariableExpression(busSelIdxVar), ConstantExpression(1UL, ILangDataType.Int32)),
                    ConstantExpression(7UL, ILangDataType.Int32)
                )
            ))
            for ((i,cog0Stm) in tmpF[0].body.withIndex()) {
                var isSame = true
                for (cog in 1 until 8) {
                    if (!tmpF[cog].body[i].isSameStatement(cog0Stm)) {
                        isSame = false
                        break
                    }
                }
                if (isSame) {
                    resultBody.add(cog0Stm)
                    continue
                }
                //überprüfen, dass das stm passt
                if (cog0Stm is LetStatement && (cog0Stm.destVar === busQRegPreVar || cog0Stm.destVar === sysQReg)) {
                    for (cog in 0 until 8) {
                        val stm = tmpF[cog].body[i]
                        if (stm !is LetStatement || stm.destVar !== cog0Stm.destVar || stm.expr !is ConstantExpression || stm.expr.value != ((cog - 1) and 7).toULong())
                            throw Exception("illegal stm A $cog $stm")
                    }
                    resultBody.add(LetStatement(busQRegPreVar, VariableExpression(tmpBusAckIdx)))
                }
                else
                    throw Exception("illegal stm C $cog0Stm")
            }
            return removeUnsuedLocals(simplifyFunction(f.localsAdded(listOf(tmpBusAckIdx)).modified(resultBody),PartialValueVarState()))
        }

        val numLocVar = hubClkEnaBus.getVariableByName("num_loc")
        val varCogELoc = ILangVariable("cog_e_loc",ILangDataType.Int32,ILangVariableDirection.Local)
        val varBusSelIdxLoc = ILangVariable("bus_sel_idx_loc",ILangDataType.Int32,ILangVariableDirection.Local)
        val hubClkBusMacroNack = MacroStatement("hubClkBusMacroNack",true,hubEvalMain.registerVarSet(),listOf(
            LetStatement(cogEnaRegVar,VariableExpression(cogEVar)),
            LetStatement(hubBusAckIdxVar, opConst32(0xFUL))
        ), setOf(),listOf())
        val hubClkBusMacroAck = MacroStatement("hubClkBusMacroAck",true,hubEvalMain.registerVarSet(),listOf(
            LetStatement(cogEnaRegVar,VariableExpression(cogEVar)),
            LetStatement(hubBusAckIdxVar, opAnd32(opAdd32(VariableExpression(busSelIdxVar), opConst32(7UL)), opConst32(7UL)))
        ), setOf(),listOf())
        val hubClkBusMacroAckCustom = MacroStatement("hubClkBusMacroAckCustom",true,hubEvalMain.registerVarSet(),listOf(
            LetStatement(cogEnaRegVar,VariableExpression(varCogELoc)),
            LetStatement(hubBusAckIdxVar, opAnd32(opAdd32(VariableExpression(varBusSelIdxLoc), opConst32(7UL)), opConst32(7UL)))
        ), setOf(),listOf(varCogELoc,varBusSelIdxLoc))
        val hubClkBusMacroAckStartCog = MacroStatement("hubClkBusMacroAckStartCog",true,hubEvalMain.registerVarSet(),listOf(
            LetStatement(cogEnaRegVar,VariableExpression(varCogELoc)),
            LetStatement(hubBusAckIdxVar, opAnd32(opAdd32(VariableExpression(varBusSelIdxLoc), opConst32(7UL)), opConst32(7UL))),
            MacroStatement("macroCogStartUp", true, hubEvalMain.registerVarSet(), listOf(), setOf(), listOf(numLocVar)
            )
        ), setOf(),listOf(varCogELoc,varBusSelIdxLoc,numLocVar))
        hubCommonMacros.add(hubClkBusMacroNack.toFunction())
        hubCommonMacros.add(hubClkBusMacroAck.toFunction())
        hubCommonMacros.add(hubClkBusMacroAckCustom.toFunction())
        hubCommonMacros.add(hubClkBusMacroAckStartCog.toFunction())


        val hubCogStartTransformer = HubCogStartTransformer(hubClkEnaBus,varCogELoc,varBusSelIdxLoc,hubClkBusMacroAckCustom,hubClkBusMacroAckStartCog)

        fun handleSysQ(f:Function):Function {
            if (!dependendVars(f.body).vars.contains(sysQReg) || !affectedVars(f.body).vars.contains(sysQReg))
                return f
            return simplifyFunction(
                    moveMatchingStatementsDown(f.localsAdded(listOf(sysQLoc)).modified(
                listOf(LetStatement(sysQLoc,VariableExpression(sysQReg)))+
                replaceVariables(f, hashMapOf(sysQReg to sysQLoc)).body+
                listOf(LetStatement(sysQReg,VariableExpression(sysQLoc))))) { it is LetStatement && it.destVar.name == "bus_ack_idx"}
            ,PartialValueVarState())
        }

        fun extractCommonMacrosFromNonCogStart(f: Function): Function {
            if (dependendVars(f.body).vars.contains(cogEnaRegVar))
                throw Exception("hub functions must not depend on cog_ena_reg")
            val tmpF=moveMatchingStatementsDown(moveMatchingStatementsDown(moveMatchingStatementsDown(f) { it is LetStatement && it.destVar.name == "cog_e"}) { it is LetStatement && it.destVar.name == "cog_ena_reg" }) { it is LetStatement && it.destVar.name == "bus_ack_idx" }
            val tmpF2=replaceVarUntilChange(replaceVarUntilChange(tmpF.localsAdded(listOf(varCogELoc,varBusSelIdxLoc)),cogEVar,varCogELoc),busSelIdxVar,varBusSelIdxLoc)
            var tmpF3=moveMatchingStatementsDown(tmpF2) {it is LetStatement && (it.destVar.name == "bus_ack_idx" || it.destVar.name == "cog_ena_reg")}
            for (i in 0 until 3)
                tmpF3=forwardLetStatements(tmpF3,setOf(),false)
            val tmpF4= tmpF3
            val remainingBody = ArrayList(tmpF4.body)
            if (remainingBody.size<2)
                throw Exception("missing stms in hub function")
            val lastBusAck = remainingBody.removeLast()
            if (lastBusAck !is LetStatement || lastBusAck.destVar.name != "bus_ack_idx")
                throw Exception("expected bus_ack_stm as last, got $lastBusAck")
            val lastCogEna = remainingBody.removeLast()
            if (lastCogEna !is LetStatement || lastCogEna.destVar.name != "cog_ena_reg")
                throw Exception("expected cog_ena_reg as second last, got $lastCogEna")

            //1. Fall: bus_ack_idx = 0xF, cog_ena_reg muss dann cog_e werden -> hubClkBusMacroNack
            if (lastBusAck.expr is ConstantExpression && lastBusAck.expr.value == 0xFUL) {
                if (lastCogEna.expr !is VariableExpression || lastCogEna.expr.variable.name != "cog_e")
                    throw Exception("invalid hub bus nack stm: $lastCogEna")
                return extractMacro(simplifyFunction(removeUnsuedLocals(tmpF4), PartialValueVarState()),hubClkBusMacroNack)
            }
            //2. Fall: bus_ack_idx = (and32 (add32 bus_sel_idx 0I7) 0I7) -> hubClkBusMacroAck
            if (lastBusAck.expr.toILang() == "(and32 (add32 bus_sel_idx 0I7) 0I7)" && lastCogEna.expr.toILang() == "cog_e") {
                return extractMacro(simplifyFunction(removeUnsuedLocals(tmpF4), PartialValueVarState()),hubClkBusMacroAck)
            }
            //3. Fall: cog_ena_reg = cog_e_loc -> hubClkBusMacroAckCustom
            if (lastCogEna.expr.toILang() != "cog_e_loc")
                throw Exception("illegal hub macro stm")
            if (lastBusAck.expr.toILang() == "(and32 (add32 bus_sel_idx_loc 0I7) 0I7)")
                return extractMacro(removeUnsuedLocals(tmpF4),hubClkBusMacroAckCustom)
            if (lastBusAck.expr.toILang() != "(and32 (add32 bus_sel_idx 0I7) 0I7)")
                throw Exception("illegal hub macro stm")
            remainingBody.add(LetStatement(varBusSelIdxLoc,VariableExpression(busSelIdxVar)))
            remainingBody.add(lastCogEna)
            remainingBody.add(lastBusAck.modified(replaceVariablesInExpression(lastBusAck.expr, hashMapOf(busSelIdxVar to varBusSelIdxLoc))))
            return extractMacro(tmpF4.modified(remainingBody),hubClkBusMacroAckCustom)
        }

        fun replacePartialMemAccess(rootExpression: Expression):Expression {
            fun searchArrayAccessExpr(expr:Expression, andBy:ULong, shiftRightBy:ULong, baseExpression: Expression):Expression {
                if (expr !is ArrayAccessExpression)
                    return baseExpression
                if (shiftRightBy != 0UL && shiftRightBy != 8UL && shiftRightBy != 16UL && shiftRightBy != 24UL)
                    return baseExpression
                val byteOffset = (shiftRightBy shr 3).toInt()
                val shiftedMask = andBy shr shiftRightBy.toInt()
                if (shiftedMask == 0xFFFFFFFFUL && shiftRightBy == 0UL) //voller word zugriff
                    return baseExpression
                if (shiftedMask == 0xFFUL) //bytezugriff
                    return PartialArrayAccessExpression(expr.arrayVariable,expr.indexExpr,true,byteOffset)
                if (shiftedMask == 0xFFFFUL) //wordzugriff
                    return PartialArrayAccessExpression(expr.arrayVariable,expr.indexExpr,false,byteOffset)
                //anderer zugriff, kann nicht optimiert werden
                //println("Handle: ($expr >> $shiftRightBy)&${shiftedMask.toString(16)} // $byteOffset /// $baseExpression")
                return baseExpression
            }
            fun searchAndExpr(expr:Expression, shiftRightBy:ULong, baseExpression: Expression):Expression {
                return if (expr is AdditiveOrBitwiseExpression && !expr.withFlags && expr.operation == AdditiveOrBitwiseExpression.Operation.And && expr.right is ConstantExpression)
                    searchArrayAccessExpr(expr.left, expr.right.value, shiftRightBy, baseExpression)
                else
                    searchArrayAccessExpr(expr, 0xFFFFFFFFUL, shiftRightBy, baseExpression)
            }
            return rootExpression.map {
                if (it.dataType != ILangDataType.Int32)
                    it
                else if (it is ShiftExpression && !it.shiftLeft && !it.withFlags && it.right is ConstantExpression)
                    searchAndExpr(it.left, it.right.value, it)
                else
                    searchAndExpr(it,0UL,it)
            }
        }

        fun simplifyMemoryReadAccess(f:Function):Function {
            return f.modified(mapExpressionsRecursive(f.body) { replacePartialMemAccess(it) } )
        }

        fun simplifyPartialMemWrite(wrStm:LetArrayStatement):LetArrayStatement {
            fun doReplace(valueExpr:Expression, shiftLeftBy:ULong, maskPrevBy:ULong):LetArrayStatement {
                if (valueExpr !is AdditiveOrBitwiseExpression || valueExpr.withFlags || valueExpr.operation != AdditiveOrBitwiseExpression.Operation.And || valueExpr.right !is ConstantExpression)
                    throw Exception("expected write mask, got $valueExpr")
                if (shiftLeftBy != 0UL && shiftLeftBy != 8UL && shiftLeftBy != 16UL && shiftLeftBy != 24UL)
                    throw Exception("invalid shift left value $shiftLeftBy")
                //write und prev mask müssen komplementär sein
                val writeMaskOrig = valueExpr.right.value
                val writeMaskShifted = writeMaskOrig shl shiftLeftBy.toInt()
                val writeValue = valueExpr.left
                if (writeMaskShifted != (maskPrevBy.inv() and 0xFFFFFFFFUL))
                    throw Exception("write and read mask must be complementary: ${writeMaskShifted.toString(16)} / ${maskPrevBy.toString(16)}")
                val byteOffset = (shiftLeftBy shr 3).toInt()
                if (writeMaskOrig == 0xFFFFUL && (byteOffset == 0 || byteOffset == 2))
                    return LetArrayStatement(wrStm.destArrayVar,wrStm.idx,valueExpr,CpuAccessSize.Word,byteOffset)
                if (writeMaskOrig == 0xFFUL)
                    return LetArrayStatement(wrStm.destArrayVar,wrStm.idx,valueExpr,CpuAccessSize.Byte,byteOffset)
                throw Exception("invalid access $writeMaskOrig $byteOffset")
            }
            if (wrStm.accessSize != CpuAccessSize.DWord) //schon vereinfacht
                return wrStm
            val orExpr = wrStm.expr
            if (orExpr !is AdditiveOrBitwiseExpression || orExpr.withFlags || orExpr.operation != AdditiveOrBitwiseExpression.Operation.Or || orExpr.dataType != ILangDataType.Int32)
                return wrStm
            val andExpr = orExpr.right
            if (andExpr !is AdditiveOrBitwiseExpression || andExpr.withFlags || andExpr.operation != AdditiveOrBitwiseExpression.Operation.And)
                return wrStm
            val arrayRdExpr = andExpr.left
            if (andExpr.right !is ConstantExpression || arrayRdExpr !is ArrayAccessExpression || arrayRdExpr.arrayVariable !== wrStm.destArrayVar || !arrayRdExpr.indexExpr.isSameExpression(wrStm.idx))
                return wrStm
            val mask = andExpr.right.value
            val valueExpr = orExpr.left
            return if (valueExpr is ShiftExpression && !valueExpr.withFlags && valueExpr.shiftLeft && valueExpr.right is ConstantExpression)
                doReplace(valueExpr.left,valueExpr.right.value,mask)
            else
                doReplace(valueExpr,0UL,mask)
        }

        fun simplifyMemoryWriteAccess(f:Function):Function {
            val hbDinVar=f.localVarSet().getVariable("hb_din")
            val f2=if (hbDinVar == null) f else {
                fun fwdHbDin(stmList: List<Statement>, initialHbDinExpr: Expression?): List<Statement> {
                    var hbDinExpr = initialHbDinExpr
                    fun fwdHbDinInExpr(rootExpression: Expression): Expression {
                        val lastHbDinExpr = hbDinExpr ?: return rootExpression
                        return replaceVariablesByExpressionInExpression(
                            rootExpression,
                            hashMapOf(hbDinVar to lastHbDinExpr)
                        )
                    }

                    val res = ArrayList(stmList)
                    for (i in 0 until res.size) {
                        when (val stm = res[i]) {
                            is LetStatement -> {
                                val newStm = stm.modified(fwdHbDinInExpr(stm.expr))
                                if (stm.destVar.name == "hb_din")
                                    hbDinExpr = newStm.expr
                                res[i] = newStm
                            }
                            is LetArrayStatement -> res[i] =
                                stm.modified(fwdHbDinInExpr(stm.idx), fwdHbDinInExpr(stm.expr))
                            is MacroStatement -> throw Exception("macro stm not supported")
                            is SwitchStatement -> {
                                res[i] = stm.modified(fwdHbDinInExpr(stm.expr), stm.cases.map {
                                    it.modified(fwdHbDin(it.body, hbDinExpr))
                                })
                                break
                            }
                        }
                    }
                    return res
                }
                simplifyFunction(f.modified(fwdHbDin(f.body,null)), PartialValueVarState())
            }
            //tauscht and32(or32(a,b),const) mit or32(and32(a,const),and32(b,const)) aus
            fun swapAndWithOr(rootExpr:Expression):Expression = rootExpr.map {
                expr -> if (expr is AdditiveOrBitwiseExpression && expr.dataType == ILangDataType.Int32 && !expr.withFlags && expr.operation == AdditiveOrBitwiseExpression.Operation.And && expr.right is ConstantExpression && expr.left is AdditiveOrBitwiseExpression && !expr.left.withFlags && expr.left.operation == AdditiveOrBitwiseExpression.Operation.Or)
                    opOr32(opAnd32(expr.left.left,expr.right), opAnd32(expr.left.right,expr.right))
                else
                    expr
            }
            //tauscht shr32(or32(a,b),const) mit or32(shr32(a,const),shr32(b,const)) aus
            fun swapShrWithOr(rootExpr:Expression):Expression = rootExpr.map {
                expr -> if (expr is ShiftExpression && expr.dataType == ILangDataType.Int32 && !expr.withFlags && !expr.shiftLeft && expr.right is ConstantExpression && expr.left is AdditiveOrBitwiseExpression && !expr.left.withFlags && expr.left.operation == AdditiveOrBitwiseExpression.Operation.Or)
                    opOr32(opShr32(expr.left.left,expr.right), opShr32(expr.left.right,expr.right))
                else
                    expr
            }
            fun swapAndWithOr(func:Function):Function = func.modified(mapExpressionsRecursive(func.body) { swapAndWithOr(it) })
            fun joinMemWriteRead(memWrBody:List<Statement>, memRdStm:LetStatement):List<Statement> {
                if (memWrBody.isEmpty())
                    return listOf(memRdStm.modified(replacePartialMemAccess(memRdStm.expr)))
                if (memWrBody.size != 1)
                    throw Exception("invalid mem wr body in hub mem wr")
                val memWrStm = memWrBody.first()
                if (memWrStm !is LetArrayStatement)
                    throw Exception("invalid mem wr body in hub mem wr")
                val simplRdExpr = simplifyExpression(swapShrWithOr(swapAndWithOr(memRdStm.expr.map {
                    if (it is ArrayAccessExpression && it.arrayVariable === memWrStm.destArrayVar && it.indexExpr.isSameExpression(memWrStm.idx))
                        memWrStm.expr
                    else
                        it
                })))
                //memWrStm darf nicht von bus_q_reg_pre abhängen
                if (dependendVars(memWrStm).vars.contains(f.getVariableByName("bus_q_reg_pre")))
                    throw Exception("memwrstm must not depend on bus_q_reg_pre")
                //dc masked kommt doppelt vor
                if (simplRdExpr !is AdditiveOrBitwiseExpression || simplRdExpr.withFlags || simplRdExpr.operation != AdditiveOrBitwiseExpression.Operation.And || simplRdExpr.right !is ConstantExpression || simplRdExpr.left !is VariableExpression || simplRdExpr.left.variable.name != "dc")
                    return listOf(memRdStm.modified(simplRdExpr),simplifyPartialMemWrite(memWrStm))
                val wrMemMod = simplifyPartialMemWrite(memWrStm)
                return listOf(
                    LetStatement(locDcMaskedVar,simplRdExpr),
                    memRdStm.modified(VariableExpression(locDcMaskedVar)),
                    wrMemMod.modified(wrMemMod.idx,wrMemMod.expr.map {
                        if (it.isSameExpression(simplRdExpr))
                            VariableExpression(locDcMaskedVar)
                        else
                            it
                    })
                )
            }
            val f3= moveMatchingStatementsUp(simplifyFunction(swapAndWithOr(simplifyFunction(swapAndWithOr(f2),PartialValueVarState())),PartialValueVarState())) {it is LetStatement && it.destVar.name == "bus_q_reg_pre"}
            //println(serializeILang(f3))
            val newBody = ArrayList<Statement>()
            var iterator=0
            while(iterator<f3.body.size) {
                val stm = f3.body[iterator]
                val nextStm = if (iterator+1>=f3.body.size) null else f3.body[iterator+1]
                if (stm is SwitchStatement && nextStm is LetStatement && nextStm.destVar.name == "bus_q_reg_pre") {
                    newBody.add(stm.modified(stm.expr, stm.cases.map { it.modified(joinMemWriteRead(it.body,nextStm)) }))
                    iterator += 2
                    continue
                }
                newBody.add(stm)
                iterator++
            }
            return f3.localsAdded(listOf(locDcMaskedVar)).modified(newBody)
        }

        fun simplifyMemoryAccess(f:Function):Function {
            var hasWriteAccess = false
            iterateStmList(f.body) {
                if (it is LetArrayStatement)
                    hasWriteAccess = true
            }
            return if (hasWriteAccess) simplifyMemoryWriteAccess(f) else simplifyMemoryReadAccess(f)
        }

        fun removeQRegLoc(f:Function):Function {
            val resultBody = ArrayList(f.body)
            var modified = false
            for (i in 1 until resultBody.size) {
                val thisStm = resultBody[i]
                val lastStm = resultBody[i-1]
                if (thisStm is LetStatement && lastStm is LetStatement && lastStm.destVar.name == "reg_hub_mem_q_reg_loc") {
                    resultBody[i] = thisStm.modified(
                        replaceVariablesByExpressionInExpression(
                            thisStm.expr,
                            hashMapOf(lastStm.destVar to lastStm.expr)
                        )
                    )
                    modified = true
                }
            }
            if (!modified)
                return f
            return removeUnsuedLocals(simplifyFunction(f.modified(resultBody), PartialValueVarState()))
        }

        fun testMemoryMacroAndReplaceWithLocals(simplifyVarState:PartialValueVarState):Function {
            val specialFPre = extractMacro(removeUnsuedLocals(simplifyFunction(
                hubClkEnaBus, simplifyVarState)),memoryMacro)
            if (specialFPre.body.isEmpty() || specialFPre.body.first() !is MacroStatement)
                throw Exception("first stm of hubClkBus must be memory macro")
            val newLocals = ILangVariableSet(specialFPre.localVarSet())
            newLocals.addVariable(memQRegLoc)
            //newLocals.addVariable(sysQRegLoc)
            val specialFPre2=removeUnsuedLocals(simplifyFunction(
                replaceVariables(
                    transformRegistersToLocals(specialFPre.modified(/*listOf(LetStatement(sysQRegLoc, opConst32(0UL)))+*/specialFPre.body.drop(1)+memoryMacroFunc.body+listOf(busAckStm)),internalHubRegsStep3),
                    hashMapOf(memQReg to memQRegLoc/*, sysQReg to sysQRegLoc*/)
                ).modified(newLocals),PartialValueVarState().setPartialValueInitialConstraint(specialFPre.getVariableByName("ac"),
                    PartialValue(ILangDataType.Int32,0UL,0xFFFFUL)
                )
            )
            )
            val usedVars = affectedVars(specialFPre2.body).union(dependendVars(specialFPre2.body)).vars
            val (fpre3, wasCogStart) = hubCogStartTransformer.transformHubCogStart(if (usedVars.contains(busSelRegVar)) handleSpecialBusEnaBusSel(specialFPre2) else specialFPre2)
            if (wasCogStart)
                return fpre3
            return removeUnsuedLocals(extractCommonMacrosFromNonCogStart(simplifyMemoryAccess(removeQRegLoc(transformHubMemoryDecrypt(handleSysQ(fpre3))))))
        }

        fun addHubFunctionIfNew(f:Function, extraStateName:String, dcDependencyAllowed:Boolean):Int {
            val usedVars = affectedVars(f.body).union(dependendVars(f.body)).vars
            //suchen, ob es diese variante schon gibt
            for ((i, otherF) in specializedClkBusEnaVersions.withIndex()) {
                if (isSameStatementList(otherF.body,f.body)) {
                    specializedClkBusEnaVersions[i] = otherF.rename(otherF.name+"_$extraStateName")
                    return i
                }
            }
            //ac darf verwendet werden, da es noch die adresse enthält
            val illegalVars = listOf("ec","wc","sc","rc").map { hubRegs.getVariable(it) ?: throw Exception("unknown var $it") }
            if (usedVars.intersect(illegalVars).isNotEmpty())
                throw Exception("used hub state vars that are not allowed $extraStateName $usedVars $illegalVars")
            val newIdx = specializedClkBusEnaVersions.size
            val newF = f.rename("hubClkBusEna_$extraStateName")
            if (!dcDependencyAllowed && dependendVars(newF.body).vars.contains(newF.getVariableByName("dc")))
                throw Exception("this hub function must not depend on dc: ${newF.name}")
            specializedClkBusEnaVersions.add(newF)
            return newIdx
        }

        //sc ist immer 2, wc ist immer false
        fun generateSpecializedClkEnaBusVersionStopped(ec:Boolean):Int {
            val f = testMemoryMacroAndReplaceWithLocals(PartialValueVarState(
                    hubClkEnaBus,
                    listOf(Pair("ec", ec),Pair("wc", false),Pair("rc",false)),
                )
                .setPartialValueInitialConstraint(hubClkEnaBus.getVariableByName("sc"), PartialValue(ILangDataType.Int32, 0x2UL, 0UL))
            )
            val extraStateName = "S${if (ec) "E" else "x"}"
            return addHubFunctionIfNew(f,extraStateName,false)
        }


        //wertebereiche: ec,wc:Boolean, sc:0..3, ac_lower3bits:0..7
        fun generateSpecializedClkEnaBusVersionRunning(ec:Boolean, wc:Boolean?, sc:Int, acLower3Bits:Int):Int {
            val f = testMemoryMacroAndReplaceWithLocals(PartialValueVarState(
                hubClkEnaBus,
                listOf(Pair("ec", ec),Pair("rc",true))+(if(wc != null) listOf(Pair("wc", wc)) else listOf()),
            )
                .setPartialValueInitialConstraint(hubClkEnaBus.getVariableByName("sc"), PartialValue(ILangDataType.Int32, sc.toULong(), 0UL))
                .setPartialValueInitialConstraint(hubClkEnaBus.getVariableByName("ac"), PartialValue(ILangDataType.Int32, acLower3Bits.toULong(), 7UL.inv()))
            )
            val extraStateName = "R${if (ec) "E" else "x"}${if (wc == null) "" else if (wc == true) "W" else "R"}${sc}A$acLower3Bits"
            return addHubFunctionIfNew(f,extraStateName,ec)
        }

        notRunningBusEFalse = generateSpecializedClkEnaBusVersionStopped(false)
        notRunningBusETrue = generateSpecializedClkEnaBusVersionStopped(true)

        //spezialisierte hubClkBusEna Funktionen erstellen
        for (scValue in 0 until 4) {
            for (acLower3BitsValue in 0 until 8) {
                val baseCode = acLower3BitsValue + (scValue shl 3)
                runningBusEFalse[baseCode] = generateSpecializedClkEnaBusVersionRunning(false, null, scValue, acLower3BitsValue)
                runningBusETrue[baseCode] = generateSpecializedClkEnaBusVersionRunning(true, false, scValue, acLower3BitsValue)
            }
        }
        for (scValue in 0 until 4)
            for (acLower3BitsValue in 0 until 8)
                runningBusETrue[acLower3BitsValue + (scValue shl 3)+0x20] = generateSpecializedClkEnaBusVersionRunning(true, true, scValue, acLower3BitsValue)


        val newhubClkPostEvalEnaLocals = ILangVariableSet("*")
        val busDInput = ILangVariable("bus_d",ILangDataType.Int32,ILangVariableDirection.Input)
        val busAshr2Input = ILangVariable("bus_ashr2",ILangDataType.Int32,ILangVariableDirection.Input)
        val busStateInput = ILangVariable("bus_state",ILangDataType.Int32,ILangVariableDirection.Input)
        newhubClkPostEvalEnaLocals.addVariable(busDInput)
        newhubClkPostEvalEnaLocals.addVariable(busAshr2Input)
        newhubClkPostEvalEnaLocals.addVariable(busStateInput)
        hubClkPostEvalEna = removeUnsuedLocals(hubClkEna0.modified(newhubClkPostEvalEnaLocals).modified(listOf(
            LetStatement(hubClkEna0.getVariableByName("dc"),VariableExpression(busDInput)),
            LetStatement(hubClkEna0.getVariableByName("acshr2"),VariableExpression(busAshr2Input)),
            LetStatement(hubClkEna0.getVariableByName("hub_state"),VariableExpression(busStateInput))
        )
            //parseILangStatements(tokenizeILang("let hub_state (or32 (or32 (or32 (or32 (and32 bus_a 0I7) (shl32 bus_s 0I3)) (mux32 bus_w 0I20 0I0)) (mux32 bus_e 0I40 0I0)) (mux32 bus_r 0I80 0I0))"),hubClkEna0.allVarSets)
        )).rename("hubClkPostEvalEna")

        hubClkEnaBusDispatcher = removeUnsuedLocals(hubClkEnaBus.modified(listOf())).modified(listOf(SwitchStatement(VariableExpression(hubStateVar),specializedClkBusEnaVersions.mapIndexed {
            hubStateCode, destF ->
            SwitchStatement.Case(listOf(hubStateCode.toULong()),listOf(
                //dummy makro ruft die funktion auf
                MacroStatement(destF.name,true,destF.registerVarSet(), listOf(), setOf(), listOf())
            ))

        }))).rename("hubClkEnaBusDispatcher")

        hubRst = hubRstPre.modified(hubRstPre.body.filter { it !is LetStatement || (it.destVar.name != "ec" && it.destVar.name != "ed")  }+listOf(
            LetStatement(hubStateVar,ConstantJumpDestinationExpression(notRunningBusEFalse.toULong(),specializedClkBusEnaVersions[notRunningBusEFalse].name)),
            LetStatement(busSelIdxVar,ConstantExpression(7UL,ILangDataType.Int32)),
            LetStatement(hubBusAckIdxVar,ConstantExpression(0xFUL,ILangDataType.Int32))
        ))

        //println(serializeILang(extractMacro(hubClkEnaBus,memoryMacro)))
        //println(serializeILang(hubClkNEnaBus))
        //println(serializeILang(hubClkPostEvalEna))
        //println(serializeILang(hubRst))
        /*for (f in specializedClkBusEnaVersions)
            println(serializeILang(f))*/



        //interne state vars entfernen
        for (v in internalHubRegsStep1+internalHubRegsStep3+listOf("ec","wc","sc","rc","ac","reg_hub_mem_q_reg",/*"sys_q",*/"bus_sel_reg","cog_start","bus_ack_result","pll_in_wire").map { hubRegs.getVariable(it)!! })
            hubRegs.removeVariable(v)
        hubRegs.removeArrayVariable(hubRegs.getArrayVariable("reg_hub_mem_memory")!!)
    }

}