package special

import ilang.*
import ilang.Function

fun findFirstUsageOfStatement(stm: Statement, f:Function):Pair<Int,Statement>? {
    if (stm is MacroStatement)
        return null
    for ((i,stm2) in f.body.withIndex())
        if (stm.isSameStatement(stm2))
            return Pair(i, stm2)
    return null
}

fun affectedVars(stm: Statement):ILangVarAndArraySet = when(stm) {
    is LetStatement -> ILangVarAndArraySet(hashSetOf(stm.destVar),HashSet())
    is LetArrayStatement -> ILangVarAndArraySet(HashSet(),hashSetOf(stm.destArrayVar))
    is SwitchStatement -> {
        var result = ILangVarAndArraySet(HashSet(),HashSet())
        for (c in stm.cases)
            result = result.union(affectedVars(c.body))
        result
    }
    is MacroStatement -> affectedVars(stm.body)
}

fun affectedVars(stmList:List<Statement>):ILangVarAndArraySet {
    var result = ILangVarAndArraySet(HashSet(),HashSet())
    for (stm in stmList)
        result = result.union(affectedVars(stm))
    return result
}

fun dependendVars(stm:Statement):ILangVarAndArraySet = when(stm) {
    is LetStatement -> stm.expr.getUsedVariablesAndArrays()
    is LetArrayStatement -> stm.expr.getUsedVariablesAndArrays().union(stm.idx.getUsedVariablesAndArrays())
    is SwitchStatement -> {
        var result = stm.expr.getUsedVariablesAndArrays()
        for (c in stm.cases)
            result = result.union(dependendVars(c.body))
        result
    }
    is MacroStatement -> dependendVars(stm.body)
}

fun dependendVars(stmList:List<Statement>):ILangVarAndArraySet {
    var result = ILangVarAndArraySet(HashSet(),HashSet())
    for (s in stmList)
        result = result.union(dependendVars(s))
    return result
}

fun canExchangeStms(stm1:Statement, stm2:Statement):Boolean {
    //println("Try Exchange $stm1 $stm2")
    val affectedVars1 = affectedVars(stm1)
    val affectedVars2 = affectedVars(stm2)
    val depVars1 = dependendVars(stm1)
    val depVars2 = dependendVars(stm2)
    //println("Aff1/Dep2: $affectedVars1 $depVars2")
    //println("Aff2/Dep1: $affectedVars2 $depVars1")
    if (affectedVars1.intersects(affectedVars2))
        return false
    if (depVars1.intersects(affectedVars2))
        return false
    if (depVars2.intersects(affectedVars1))
        return false
    return true
}

fun trySwapStatementDown(idx:Int, maxIdx:Int, inplaceBody:ArrayList<Statement>, debug:Boolean):Boolean {
    val stm = inplaceBody[idx]
    for (k in idx+1 .. maxIdx) {
        val canSwap = canExchangeStms(stm,inplaceBody[k])
        if (debug)
            println("   > test with ($k) ${inplaceBody[k]} $canSwap")
        if (!canSwap)
            return false
    }
    inplaceBody.add(maxIdx+1, stm) //danach einfügen
    inplaceBody.removeAt(idx) //am ursprungsort löschen
    return true
}

fun trySwapStatementDownAsFarAsPossible(idx:Int, maximumIdx:Int, inplaceBody:ArrayList<Statement>, debug:Boolean) {
    if (debug)
        println("Test Move down ($idx) ${inplaceBody[idx]}")
    val stm = inplaceBody[idx]
    var maxIdx = idx
    for (k in idx+1 .. maximumIdx) {
        val canSwap = canExchangeStms(stm,inplaceBody[k])
        if (debug)
            println("   > test with ($k) ${inplaceBody[k]} $canSwap")
        if (!canSwap)
            break
        maxIdx = k
    }
    if (maxIdx == idx) //nichts zu tun
        return
    inplaceBody.add(maxIdx+1, stm) //danach einfügen
    inplaceBody.removeAt(idx) //am ursprungsort löschen
}

fun trySwapStatementUp(idx:Int, minIdx:Int, inplaceBody:ArrayList<Statement>, debug:Boolean):Boolean {
    val stm = inplaceBody[idx]
    for (k in idx-1 downTo minIdx) {
        val otherStm = inplaceBody[k]
        val canSwap = canExchangeStms(stm,otherStm)
        if (debug)
            println("   > test with ($k) $otherStm $canSwap")
        if (!canSwap)
            return false
    }
    inplaceBody.removeAt(idx) //am ursprungsort löschen
    inplaceBody.add(minIdx, stm) //davor einfügen
    return true
}

fun trySwapStatementUpAsFarAsPossible(idx:Int, mainiumIdx:Int, inplaceBody:ArrayList<Statement>, debug:Boolean) {
    if (debug)
        println("Test Move up ($idx) ${inplaceBody[idx]}")
    val stm = inplaceBody[idx]
    var minIdx = idx
    for (k in idx-1 downTo mainiumIdx) {
        val canSwap = canExchangeStms(stm,inplaceBody[k])
        if (debug)
            println("   > test with ($k) ${inplaceBody[k]} $canSwap")
        if (!canSwap)
            break
        minIdx = k
    }
    if (minIdx == idx) //nichts zu tun
        return
    inplaceBody.removeAt(idx) //am ursprungsort löschen
    inplaceBody.add(minIdx, stm) //davor einfügen
}


fun moveMatchingStatementsUp(f:Function, isOfStatementClass:(stm:Statement)->Boolean):Function {
    val tmpBody = ArrayList<Statement>(f.body)
    val doneStm = HashSet<Statement>()
    fun findNextStm():Int? {
        for ((i,stm) in tmpBody.withIndex())
            if (!doneStm.contains(stm) && isOfStatementClass(stm))
                return i
        return null
    }
    while (true) {
        val moveUpStmIdx = findNextStm() ?: return f.modified(tmpBody)
        doneStm.add(tmpBody[moveUpStmIdx])
        var minimumUpIdx = 0
        while (minimumUpIdx<tmpBody.size && isOfStatementClass(tmpBody[minimumUpIdx]))
            minimumUpIdx++
        trySwapStatementUpAsFarAsPossible(moveUpStmIdx,minimumUpIdx,tmpBody, false)
    }
}

fun moveMatchingStatementsDown(f:Function, isOfStatementClass:(stm:Statement)->Boolean):Function {
    val tmpBody = ArrayList<Statement>(f.body)
    val doneStm = HashSet<Statement>()
    fun findNextStm():Int? {
        for ((i,stm) in tmpBody.withIndex().reversed())
            if (!doneStm.contains(stm) && isOfStatementClass(stm))
                return i
        return null
    }
    while (true) {
        val moveDownStmIdx = findNextStm() ?: return f.modified(tmpBody)
        doneStm.add(tmpBody[moveDownStmIdx])
        var maximumDownIdx = tmpBody.size-1
        while (maximumDownIdx>moveDownStmIdx && isOfStatementClass(tmpBody[maximumDownIdx]))
            maximumDownIdx--
        trySwapStatementDownAsFarAsPossible(moveDownStmIdx,maximumDownIdx,tmpBody, false)
    }
}

//versucht origGroupStm maximal zu gruppieren, also statements die dazwischenliegen zu entfernen
fun adjustStatementOrder(f:Function, origGroupStm:List<Statement>, debug:Boolean, debugCoarse:Boolean):Function {
    //origGroupStm kann von anderen Funktionen kommen, daher die equivalenten aus dieser funktion suchen
    val groupStm = origGroupStm.map { findFirstUsageOfStatement(it,f)?.second ?: throw Exception("statement not found") }
    val newBody = ArrayList(f.body)
    val cannotSwapStm = HashSet<Statement>()



    fun trySwapAStatementDown():Boolean {
        val groupStmIndices = groupStm.map { newBody.indexOf(it) }
        if (debug) {
            println()
            println(groupStmIndices)
        }
        val destIdx = groupStmIndices.last()
        for (i in groupStmIndices.size-1 downTo 1) {
            val prevIdx = groupStmIndices[i-1]
            val thisIdx = groupStmIndices[i]
            if (thisIdx<=prevIdx || thisIdx<0 || prevIdx<0)
                throw Exception("statement in invalid different orders $prevIdx $thisIdx")
            if (prevIdx+1 == thisIdx)
                continue
            if (debug) {
                println("1 common stm ($prevIdx): ${newBody[prevIdx]}")
                println("2 common stm ($thisIdx): ${newBody[thisIdx]}")
            }
            for (idx in thisIdx-1 downTo prevIdx+1) {
                if (cannotSwapStm.contains(newBody[idx]))
                    continue
                if (debug)
                    println("($idx) analyse down ${newBody[idx]}")
                if (trySwapStatementDown(idx, destIdx, newBody, debug))
                    return true
                cannotSwapStm.add(newBody[idx])
            }
        }
        return false
    }



    fun trySwapAStatementUp():Boolean {
        val groupStmIndices = groupStm.map { newBody.indexOf(it) }
        if (debug) {
            println()
            println(groupStmIndices)
        }
        val destIdx = groupStmIndices.first()
        for (i in 1 until groupStmIndices.size) {
            val prevIdx = groupStmIndices[i-1]
            val thisIdx = groupStmIndices[i]
            if (thisIdx<=prevIdx || thisIdx<0 || prevIdx<0)
                throw Exception("statement in invalid different orders $prevIdx $thisIdx")
            if (prevIdx+1 == thisIdx)
                continue
            if (debug) {
                println("1 common stm ($prevIdx): ${newBody[prevIdx]}")
                println("2 common stm ($thisIdx): ${newBody[thisIdx]}")
            }
            for (idx in prevIdx+1 .. thisIdx-1) {
                if (cannotSwapStm.contains(newBody[idx]))
                    continue
                if (debug)
                    println("($idx) analyse up ${newBody[idx]}")
                if (trySwapStatementUp(idx, destIdx, newBody, debug))
                    return true
                cannotSwapStm.add(newBody[idx])
            }
        }
        return false
    }

    if (debugCoarse)
        println("adjStm down ${f.name} ${groupStm.map { newBody.indexOf(it) }}")
    while(trySwapAStatementDown()) {}
    cannotSwapStm.clear()
    if (debugCoarse)
        println("adjStm up ${f.name} ${groupStm.map { newBody.indexOf(it) }}")
    while(trySwapAStatementUp()) {}
    val finalOrder = groupStm.map { newBody.indexOf(it) }
    for (t in 1 until finalOrder.size) {
        if (finalOrder[t] != finalOrder[t - 1] + 1) {
            println("Final Order Error")
            for ((j,s) in newBody.withIndex())
                println("> ($j ${finalOrder.contains(j)}) $s")
            throw java.lang.Exception("no ascending final order of macro statements found $finalOrder")
        }
    }
    if (debugCoarse)
        println("After: $finalOrder")
    return f.modified(newBody)
}

//soll gemeinsame makros in der clk routine finden (ctr, vid)
enum class MacroType { CogPeripherals, CogReset, HubMemory}
fun findCommonMacro(macroType:MacroType, origFunctionList:List<Function>, debug:Boolean=false, debugCoarse: Boolean=false):Pair<List<Function>, MacroStatement> {
    //Findet eine Sequenz von Instruktionen, die in allen Funktionen vorkommtn

    fun statementUsedInAllFunctions(stm:Statement,fList:List<Function>):Boolean {
        for (f in fList)
            findFirstUsageOfStatement(stm, f) ?: return false
        return true
    }

    fun extractCommonStatements(fList:List<Function>, debug: Boolean):List<Statement> {
        val commonStm = ArrayList<Statement>()
        for (stm in fList.first().body) {
            var used = statementUsedInAllFunctions(stm,fList)
            //manuelle korrektur
            if (debug) {
                if (used)
                    println("> C $stm")
                else {
                    println("> . $stm")
                    println(fList.map { if (findFirstUsageOfStatement(stm, it)!=null) "X" else "_"  }.joinToString(","))
                }
            }
            if (used)
                commonStm.add(stm)
        }
        return commonStm
    }

    val origCommonStm = extractCommonStatements(origFunctionList, debug)

    val newFunctions = origFunctionList.map {
        val newF = adjustStatementOrder(it, origCommonStm, debug, debugCoarse)
        if (macroType == MacroType.CogPeripherals) removeUnsuedLocals(simplifyFunction(newF,PartialValueVarState())) else newF
    }
    val newCommonStatement = extractCommonStatements(newFunctions, false)

    val localsWrittenInMacro = HashSet<ILangVariable>()
    val localReadInMacro = HashSet<ILangVariable>()
    for (stm in newCommonStatement) {
        for (v in affectedVars(stm).vars)
            if (v.direction == ILangVariableDirection.Local)
                localsWrittenInMacro.add(v)
            else if (v.direction != ILangVariableDirection.Register)
                throw Exception("only registers or pure locals allowed in macro: ${v.name}")
        for (v in dependendVars(stm).vars)
            if (v.direction == ILangVariableDirection.Local)
                localReadInMacro.add(v)
            else if (v.direction != ILangVariableDirection.Register)
                throw Exception("only registers or pure locals allowed in macro: ${v.name}")
    }
    //überprüfen, dass alle lokalen variablen im makro auch nur dort und nicht in der funktion verwendet werden
    val localsReadInMacroAndUsedInParent = HashSet<ILangVariable>()
    for (f in newFunctions) {
        for (stm in f.body) {
            if (newCommonStatement.find { it.isSameStatement(stm) } != null) //makro statement
                continue
            val usedVars = dependendVars(stm).vars + affectedVars(stm).vars
            //variablen, die im makro geändert werden, dürfen im parent nicht vorkommen
            if (localsWrittenInMacro.intersect(usedVars).isNotEmpty()) {
                for (n in newCommonStatement)
                    println("COMMON > $n")
                for (xx in f.body)
                    println("> $xx")
                throw Exception("some locals written in macro and used in parent function: $usedVars $localsWrittenInMacro $stm")
            }
            //schauen, ob variablen im makro gelesen und im parent verwendet werden
            localsReadInMacroAndUsedInParent.addAll(localReadInMacro.intersect(usedVars))
        }
    }
    if (debugCoarse)
        println("locals used in macro and parent: $localsReadInMacroAndUsedInParent")
    return Pair(newFunctions, MacroStatement("macro$macroType",true,newFunctions.first().registerVarSet(), newCommonStatement, localsWrittenInMacro+localReadInMacro.subtract(localsReadInMacroAndUsedInParent),localsReadInMacroAndUsedInParent.toList().sortedBy { it.name }))
}

fun extractMacro(f:Function, macro: MacroStatement):Function {
    val indices = macro.body.map {
        val res = findFirstUsageOfStatement(it, f)?.first
        if (res == null) {
            println(serializeILang(f))
            println(serializeILang(macro.toFunction()))
            throw java.lang.Exception("macro sequence not found in ${f.name}, stm: $it")
        }
        res
    }
    for (t in 1 until indices.size) {
        if (indices[t] != indices[t - 1] + 1) {
            println(serializeILang(f))
            println(serializeILang(macro.toFunction()))
            throw java.lang.Exception("no ascending final order of macro statements found $indices in function ${f.name} / ${macro.macroName}")
        }
    }
    val newBody = ArrayList<Statement>()
    for ((i,stm) in f.body.withIndex()) {
        if (i == indices.first())
            newBody.add(macro)
        if (indices.contains(i))
            continue
        newBody.add(stm)
    }
    return f.modified(newBody)
}