package special

import assembler.AsmInputVariable
import assembler.CpuRegister
import ilang.*
import ilang.Function

class SplittedCogPinOut(fIn:Function) {
    val inputs:List<AsmInputVariable>
    val counterA:FunctionPatchesList
    val counterB:FunctionPatchesList
    val video:FunctionPatchesList
    val commonTop = FunctionPatchesList(1,"cogPinOutTop")
    val commonBot = FunctionPatchesList(1,"cogPinOutBot")
    init {
        val tmpBody = ArrayList(fIn.body)
        val newLocalPinOutVar = ILangVariable("cog_pin_out",ILangDataType.Int32,ILangVariableDirection.InOut)
        val globalPinOut = ILangVariable("global_pin_out",ILangDataType.Int32,ILangVariableDirection.InOut)
        val globalPinDir = ILangVariable("global_pin_dir",ILangDataType.Int32,ILangVariableDirection.InOut)
        inputs = listOf(
            AsmInputVariable(globalPinOut, CpuRegister.R8,true),
            AsmInputVariable(globalPinDir, CpuRegister.R9,true),
            AsmInputVariable(newLocalPinOutVar, CpuRegister.RDX,true)
        )
        val fout = removeUnsuedLocals(fIn.modified(listOf())).localsAdded(listOf(globalPinOut,globalPinDir,newLocalPinOutVar))
        fun extractPartAndReplacePinOut(localOutputVarName:String):Function {
            val localOutVar = fIn.getVariableByName(localOutputVarName)
            for (i in 0 until tmpBody.size) {
                val stm = tmpBody[i]
                if (stm is LetStatement && stm.destVar === localOutVar && stm.expr is ConstantExpression && stm.expr.value == 0UL) {
                    tmpBody.removeAt(i)
                    break
                }
            }
            val switchStm = tmpBody.firstOrNull { it is SwitchStatement && affectedVars(it).vars.contains(localOutVar) } as? SwitchStatement ?: throw Exception("switch stm not found")
            tmpBody.remove(switchStm)
            return fout.modified(listOf(switchStm.map {
                if (it is LetStatement && it.destVar === localOutVar)
                    LetStatement(newLocalPinOutVar, opOr32(VariableExpression(newLocalPinOutVar), it.expr))
                else
                    it
            }))
        }

        fun extractCounter(ctraA:Boolean):FunctionPatchesList {
            val result = FunctionPatchesList(0x20,"cogPinOutCounter${if(ctraA) "A" else "B"}")
            val tmpF = extractPartAndReplacePinOut(if (ctraA) "outp_cog_ctra_pin_out" else "outp_cog_ctrb_pin_out")
            val ctrModeVar = tmpF.getVariableByName(if (ctraA) "reg_cog_ctra_ctrmode" else "reg_cog_ctrb_ctrmode")
            for (ctrMode in 0 until 0x20) {
                val simplF= fout.modified(removeUnsuedLocals(simplifyFunction(tmpF,PartialValueVarState().setPartialValue(ctrModeVar,ILangDataType.Int32,ctrMode.toULong(),false))).body)
                result.addPatch(ctrMode,simplF,"M${ctrMode.toString(16)}")
            }
            return result
        }

        fun extractVideo():FunctionPatchesList {
            val result = FunctionPatchesList(4,"cogPinOutVideo")
            val tmpF = extractPartAndReplacePinOut("outp_cog_vid_pin_out")
            val ctrModeVar = tmpF.getVariableByName("reg_cog_vid_vcfg_vmode")
            for (ctrMode in 0 until 4) {
                val simplF= fout.modified(removeUnsuedLocals(simplifyFunction(tmpF,PartialValueVarState().setPartialValue(ctrModeVar,ILangDataType.Int32,ctrMode.toULong(),false))).body)
                result.addPatch(ctrMode,simplF,"V${ctrMode.toString(16)}")
            }
            return result
        }

        fun verifyLetStm(stm:Statement, expectedVar:String, expectedExpr:String) {
            if (stm !is LetStatement || stm.destVar.name != expectedVar || stm.expr.toILang() != expectedExpr)
                throw Exception("invalid bottom stm: $stm")
        }

        counterA = extractCounter(true)
        counterB = extractCounter(false)
        video = extractVideo()
        if (tmpBody.size != 2)
            throw Exception("invalid bottom stms")
        verifyLetStm(tmpBody[0], "pin_out", "(and32 (or32 (or32 (or32 outa outp_cog_ctra_pin_out) outp_cog_ctrb_pin_out) outp_cog_vid_pin_out) dira)")
        verifyLetStm(tmpBody[1],"pin_dir","dira")


        commonTop.addPatch(0,fout.modified(listOf(
            LetStatement(newLocalPinOutVar,VariableExpression(fIn.getVariableByName("outa")))
        )),"A")

        val dirAVar = VariableExpression(fIn.getVariableByName("dira"))
        commonBot.addPatch(0,fout.modified(listOf(
            LetStatement(globalPinOut, opOr32(VariableExpression(globalPinOut),opAnd32(VariableExpression(newLocalPinOutVar),dirAVar))),
            LetStatement(globalPinDir,opOr32(VariableExpression(globalPinDir),dirAVar))
        )),"A")
    }
}