package special

import ilang.*
import ilang.Function

fun isSpecialOpcodeForReadMemBuffer(opCode6Bit:Int?)=(opCode6Bit == 0x17 || opCode6Bit == 0x39 || opCode6Bit == 0x3A || opCode6Bit == 0x3B)
val expectedState3SourceExpression = "(mux32 (testBit32 instruction_reg 0I16) instr_src (mux32 (le32 instr_src 0I1f0) sy_reg (mux32 (eq32 instr_src 0I1f0) (and32 (shr32 progctr 0Ic) 0Ifffc) (mux32 (eq32 instr_src 0I1f1) cnt_wire (mux32 (eq32 instr_src 0I1f2) pin_in_wire (mux32 (eq32 instr_src 0I1fc) (lower32 reg_cog_ctra_phase) (mux32 (eq32 instr_src 0I1fd) (lower32 reg_cog_ctrb_phase) sy_reg)))))))"

fun getExpectedState3MemReadBufferExpression(f:Function, opCode6Bit:Int?, readFromSourceVar:()->String):Expression {
    return ArrayAccessExpression(f.allVarSets.getArrayVariableByName("reg_cog_ram_r"),
        if (isSpecialOpcodeForReadMemBuffer(opCode6Bit))
            opAnd32(
                VariableExpression(f.getVariableByName(readFromSourceVar())),
                opConst32(0x1FFUL)
            )
        else
            VariableExpression(f.getVariableByName("p_reg")),
        )
}

class CogClkOtherStateOptimizer(val stateManager: CompoundStateManager, stateNena:CompoundStateManager.StateCode, state0: CompoundStateManager.StateCode, val cogClkPeripheralsMacro: MacroStatement, val baseCogClkFunctions: List<Function>, val state4Result: CogClkState4Result) {
    private val state4ResultRunningCondTrueExpression = LookupExpression(
        opShr32(
            VariableExpression(state4Result.runningCondFalseFunc.f!!.getVariableByName("instruction_reg")),
            ConstantExpression(23UL,ILangDataType.Int32)
        ),state4Result.runningCondTrueFunc.map { it.code.toULong() }
    )

    class PredefinedState5Function(val running: CompoundState.Running, val cond:Boolean, val dependsOnMatch:Boolean, val preF:Function, val reservedState:CompoundStateManager.StateCode, var extraStateName: String, val state5OpCodeType: CompoundState.State5OpCodeType)

    private val state5PredfinedState = ArrayList<PredefinedState5Function>()

    init {
        val predefStates = state4Result.reservedStates

        handleStateNena(stateNena)
        handleState0(state0)
        handleState1(predefStates.state1StoppedNoCancel,CompoundState.Running.Stopped, CompoundState.Cancel.NoCancel)
        handleState1(predefStates.state1StoppedCancel,CompoundState.Running.Stopped, CompoundState.Cancel.Cancel)
        handleState1(predefStates.state1RunningNoCancel,CompoundState.Running.Running, CompoundState.Cancel.NoCancel)
        handleState1(predefStates.state1RunningCancel,CompoundState.Running.Running, CompoundState.Cancel.Cancel)
        //state5 auflösen
        for (pre in state5PredfinedState)
            resolveState5(pre)
        //state 3 soure prüfen
        for (st3 in stateManager.getAllStates())
            if (st3.state.baseState == 3)
                st3.f=validateSourceRegInState3AndRemoveSyReg(st3.f!!,st3.state.running)
    }

    private fun addOrLookupFunction(f:Function, state:CompoundState, extraStateName:String):CompoundStateManager.StateCode = stateManager.addFunction(extractMacroAndOptimize(f), state, extraStateName)

    private fun addFunction(f:Function, state:CompoundState, extraStateName:String):ConstantExpression {
        return ConstantExpression(addOrLookupFunction(f, state, extraStateName).code.toULong(),ILangDataType.Int32)
    }

    private fun getVersionOfState(initialState:CompoundState):Function {
        val initialPins = arrayListOf(Pair("ena",true),Pair("running",initialState.running.toBool()))
        when (initialState.match) {
            CompoundState.Match.Match -> initialPins.add(Pair("match", true))
            CompoundState.Match.NoMatch -> initialPins.add(Pair("match", false))
            CompoundState.Match.Undefined -> {}
        }
        when (initialState.cancel) {
            CompoundState.Cancel.Cancel -> initialPins.add(Pair("cancel", true))
            CompoundState.Cancel.NoCancel -> initialPins.add(Pair("cancel", false))
            CompoundState.Cancel.Undefined -> {}
        }
        val f=baseCogClkFunctions[initialState.baseState]
        val varState = PartialValueVarState(f,initialPins)
        varState.setPartialValueInitialConstraint(f.getVariableByName("p_reg"), PartialValue(ILangDataType.Int32,0UL,0x1FFUL))
        if (initialState.baseState == 3 || initialState.baseState == 5)
            varState.setNeverForwardVariable(f.getVariableByName("cond"))
        return removeUnsuedLocals(simplifyFunction(f,varState))
    }

    private fun replaceBusAckWire(f:Function):Function {
        return replaceVariablesByExpression(f, hashMapOf(
            f.getVariableByName("bus_ack_wire") to opEq32(VariableExpression(f.getVariableByName("bus_ack_idx")), GetCogNumberExpression())
        ))
    }

    private fun handleStateNena(destCompoundState: CompoundStateManager.StateCode) {
        //dummy funktion wenn ena==false
        //nichts machen, da alle cog peripherie schon bei negedge ena deaktiviert wurden
        destCompoundState.setFunction(removeUnsuedLocals(baseCogClkFunctions[0].modified(listOf())),"")
    }

    private fun extractCogMemAccess(f:Function, destVar:String):Pair<Function,Expression> {
        val f1=moveMatchingStatementsDown(f) { it is LetStatement && it.destVar.name == destVar }
        if (f1.body.isEmpty())
            throw Exception("expected stm")
        val lastStm = f1.body.last()
        if (lastStm !is LetStatement || lastStm.destVar.name != destVar)
            throw Exception("invalid hub mem stm: $lastStm")
        val memExpr = lastStm.expr
        val f2 = f1.modified(f1.body.dropLast(1))
        return Pair(f2,memExpr)
    }

    private fun extractCogMemToBuffer(f:Function):Pair<Function,Expression> {
        //extrahiert den ausdruck, in dem der speicher nach reg_cog_ram_q_reg gelesen wird
        return extractCogMemAccess(f,"reg_cog_ram_q_reg")
    }

    private fun modifyCogMemBufferFromLastCycle(f:Function, memExpressionLastCycle:Expression):Function {
        val f1= moveMatchingStatementsUp(f) { it is LetStatement && it.expr is VariableExpression && it.expr.variable.name == "reg_cog_ram_q_reg"}
        if (f1.body.isEmpty())
            throw Exception("expected stm")
        val firstStm = f1.body.first()
        if (firstStm !is LetStatement || firstStm.expr !is VariableExpression || firstStm.expr.variable.name != "reg_cog_ram_q_reg")
            throw Exception("invalid first mem stm: $firstStm")
        return  f1.modified(listOf(firstStm.modified(memExpressionLastCycle))+f1.body.drop(1))
    }

    private fun handleState0(destCompoundState: CompoundStateManager.StateCode) {
        val initialState = CompoundState(0,CompoundState.Running.Stopped,CompoundState.Cancel.NoCancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown)
        val f = extractCompoundState(getVersionOfState(initialState), initialState) {
                newState ->
            if (newState.running != initialState.running || newState.baseState != 1 || newState.cancel != initialState.cancel || newState.match != initialState.match)
                throw Exception("invalid state transition $initialState -> $newState")
            ConstantExpression(state4Result.reservedStates.state1StoppedNoCancel.code.toULong(),ILangDataType.Int32)
        }
        destCompoundState.setFunction(extractMacroAndOptimize(f),"")
    }

    private fun handleState1(destCompoundState: CompoundStateManager.StateCode, running: CompoundState.Running, cancel: CompoundState.Cancel) { //match ist immer false
        val initialState = CompoundState(1,running,cancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown)
        //ursprünglich wird hier der cog speicher eingelesen um dann im nächsten takt verwendet zu werden
        //das einlesen wird in den nächsten Takt verschoben
        val (f1,memExpression) = extractCogMemToBuffer(getVersionOfState(initialState))
        val f2 = extractCompoundState(f1, initialState) {
                newState ->
            if (newState.running != running || newState.baseState != 2 || newState.cancel != cancel || newState.match != CompoundState.Match.NoMatch)
                throw Exception("invalid state transition $initialState -> $newState")
            handleState2(running, cancel, memExpression)
        }
        destCompoundState.setFunction(extractMacroAndOptimize(f2),"")
    }

    private fun handleState2(running: CompoundState.Running, cancel: CompoundState.Cancel, cogMemExpressionFromState1:Expression):Expression { //match ist immer false
        val initialState = CompoundState(2,running,cancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown)
        val (f0,memExprForState3)=extractCogMemToBuffer(moveMatchingStatementsUp(modifyCogMemBufferFromLastCycle(getVersionOfState(initialState),cogMemExpressionFromState1)) { it is LetStatement && it.destVar.name == "sy_reg" })
        val (f1,syRegExpr) = extractCogMemAccess(simplifyFunction(f0,PartialValueVarState()),"sy_reg")
        val f2 = extractCompoundState(f1, initialState) {
            newState ->
            if (newState.running != running || newState.baseState != 3 || newState.cancel != cancel || newState.match != CompoundState.Match.NoMatch)
                throw Exception("invalid state transition $initialState -> $newState")
            handleState3General(running, cancel, memExprForState3, syRegExpr)
        }

        //println(serializeILang(extractMacroAndOptimize(f2)))
        return addFunction(f2,initialState,"")
    }

    private fun validateSourceRegInState3AndRemoveSyReg(f:Function,running:CompoundState.Running):Function {
        val f1=moveMatchingStatementsUp(f) { it is LetStatement && (it.destVar.name == "sy_reg" || it.destVar.name == "source" || it.destVar.name == "instr_src")}
        val remainingBody = ArrayList(f1.body)
        val newBody = ArrayList<Statement>()
        val syReg = f.getVariableByName("sy_reg")
        val sourceVar = f.getVariableByName("source")
        val syRegStm = remainingBody.removeFirstOrNull()
        if (syRegStm !is LetStatement || syRegStm.destVar !== syReg)
            throw Exception("expected let sy_reg stm, $syRegStm")
        val instrSrc = remainingBody.firstOrNull()
        if (instrSrc is LetStatement && instrSrc.destVar.name == "instr_src")
            newBody.add(remainingBody.removeAt(0))
        val sourceStm = remainingBody.removeFirstOrNull()
        if (sourceStm !is LetStatement || sourceStm.destVar !== sourceVar)
            throw Exception("expected let source stm, $sourceStm")
        //überprüfen, dass source die richtige expr besitzt
        val expectedExpr = when(running) {
            CompoundState.Running.Stopped -> "sy_reg"
            CompoundState.Running.Running -> expectedState3SourceExpression
        }
        if (sourceStm.expr.toILang() != expectedExpr)
            throw Exception("invalid let source stm: ${sourceStm.expr} / $expectedExpr")
        //sy_reg austauschen
        newBody.add(sourceStm.modified(replaceVariablesByExpressionInExpression(sourceStm.expr, hashMapOf(syReg to syRegStm.expr))))
        //der restliche teil darf nicht mehr von sy_reg abhängen, es setzen oder source setzen
        val affVars = affectedVars(remainingBody).vars
        if (affVars.contains(syReg) || affVars.contains(sourceVar))
            throw Exception("illegal affected vars in remaining state3 body: $affVars")
        val depVars = dependendVars(remainingBody).vars
        if (depVars.contains(syReg))
            throw Exception("illegal affected vars in remaining state3 body: $depVars")
        newBody.addAll(remainingBody)
        val f2=f1.modified(newBody)
        return f2
    }

    private fun handleState3General(running:CompoundState.Running,cancel:CompoundState.Cancel, cogMemExpressionFromState2:Expression, syRegExpression: Expression):Expression { //match ist immer false
        //cancel wird erst später gefiltert
        val f0 = moveMatchingStatementsDown(modifyCogMemBufferFromLastCycle(getVersionOfState(CompoundState(3,running,CompoundState.Cancel.Undefined,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown)),cogMemExpressionFromState2)){ it is LetStatement && it.destVar.name == "source"}
        val f1 = f0.modified(listOf(LetStatement(f0.getVariableByName("sy_reg"),syRegExpression))+f0.body)

        val extracted = ExtractedConditionCode(f1,running.toBool(),listOf())
        val condition = partialEvaluateExpression(extracted.condition,PartialValueVarState(f1,listOf(Pair("cancel",when(cancel) {
            CompoundState.Cancel.Cancel -> true
            CompoundState.Cancel.NoCancel -> false
            CompoundState.Cancel.Undefined -> throw Exception("handleState3 cancel must not be undefined")
        })))).expr

        if (condition is ConstantExpression)
            return if (condition.value != 0UL)
                handleState3Specific(running, cancel,true, extracted.condTrueFunction)
            else
                handleState3Specific(running, cancel,false, extracted.condFalseFunction)
        //condition vereinfachen
        return MuxExpression(ILangDataType.Int32,condition,
            handleState3Specific(running, cancel,true, extracted.condTrueFunction),
            handleState3Specific(running, cancel,false, extracted.condFalseFunction)
        )
    }

    private fun handleState3Specific(running: CompoundState.Running, cancel: CompoundState.Cancel, cond: Boolean, baseF:Function):Expression { //match ist immer false
        fun processState3Fun(f0:Function,opCode: Int?):ConstantExpression {
            val condName = if(cond) "T" else "F"
            val extraStateName = if (opCode!=null) "${condName}OP$opCode" else condName
            val initialState = CompoundState(3,running,cancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown)
            val (f1,memExpr) = extractCogMemToBuffer(simplifyFunction(f0,PartialValueVarState().setPartialValueInitialConstraint(f0.getVariableByName("p_reg"), PartialValue(ILangDataType.Int32,0UL,0x1FFUL))))
            val f2 = moveMatchingStatementsDown(f1) {it is LetStatement && it.destVar.name == "source" }
            if (opCode == null) {
                if (running == CompoundState.Running.Running && cond)
                    throw Exception("invalid state3 without opcode: $running $cond")
            }
            else {
                if (running != CompoundState.Running.Running || !cond)
                    throw Exception("invalid state3 with opcode: $running $cond")
            }
            //für opcodes 0x17, 0x39, 0x3A, 0x3B findet der array access über source statt
            val expectedMemExpr = getExpectedState3MemReadBufferExpression(f2,opCode) { "sx"}
            if (!expectedMemExpr.isSameExpression(memExpr))
                throw Exception("invalid memExpr ($opCode): $memExpr, expected: $expectedMemExpr")
            //validateSourceRegInState3(f2,running)

            return addFunction(replaceBusAckWire(extractCompoundState(f2,initialState) {
                newState ->
                if (newState.running != running || newState.match != CompoundState.Match.NoMatch || newState.cancel != cancel)
                    throw Exception("invalid state transition $initialState$extraStateName -> $newState")
                if (newState.baseState == 4)
                    handleState4(running,cond)
                else if (newState.baseState == 5)
                    preHandleState5(running,cond,opCode)
                else
                    throw Exception("invalid state transition $initialState$extraStateName -> $newState")
            }),initialState,extraStateName)
        }

        if (baseF.localVarSet().getVariableList().find { it.name == "instr_opcode" } == null)
            return processState3Fun(baseF,null)
        //alle opcodes durchgehen
        val instructionReg = baseF.getVariableByName("instruction_reg")
        val instrOpCode = opAnd32(
            opShr32(VariableExpression(instructionReg),ConstantExpression(0x1AUL,ILangDataType.Int32)),
            ConstantExpression(0x3FUL,ILangDataType.Int32)
        )
        val opcodeSelectTable = ArrayList<ULong>()
        for (opcode in 0 until 64) {
            val f= removeUnsuedLocals(simplifyFunction(baseF,
                PartialValueVarState().setPartialValueInitialConstraint(instructionReg,
                    PartialValue(ILangDataType.Int32,(opcode.toULong() shl 26) ,(0xFC000000UL).inv())
                )))
            opcodeSelectTable.add(processState3Fun(f,opcode).value)
        }
        return LookupExpression(instrOpCode,opcodeSelectTable)
    }

    fun handleState4(running: CompoundState.Running, cond:Boolean):Expression {
        return when(running) {
            CompoundState.Running.Stopped -> ConstantExpression((if (!cond) state4Result.stoppedCondFalseFunc.code else state4Result.stoppedCondTrueFunc.code).toULong(),ILangDataType.Int32)
            CompoundState.Running.Running -> {
                if (!cond) ConstantExpression(state4Result.runningCondFalseFunc.code.toULong(),ILangDataType.Int32) else state4ResultRunningCondTrueExpression
            }
        }
    }


    //state5 kann sich selbst aufrufen, deshalb wird hier erstmal nur die states reserviert
    //zu diesem zeitpunkt ist match immer NoMatch
    fun preHandleState5(running: CompoundState.Running, cond:Boolean, opCode:Int?):Expression {
        //cancel wird erst später gefiltert
        val extracted = ExtractedConditionCode(
            getVersionOfState(
                CompoundState(
                    5,
                    running,
                    CompoundState.Cancel.Undefined,
                    CompoundState.Match.Undefined,
                    CompoundState.State5OpCodeType.Unknown
                )
            ), running.toBool(),listOf()
        )

        val f0 = if (cond) extracted.condTrueFunction else extracted.condFalseFunction
        val f = replaceBusAckWire(if (opCode == null) f0 else removeUnsuedLocals(
            simplifyFunction(
                f0,
                PartialValueVarState().setPartialValueInitialConstraint(
                    f0.getVariableByName("instruction_reg"),
                    PartialValue(ILangDataType.Int32, (opCode.toULong() shl 26), (0xFC000000UL).inv())
                )
            )
        ))
        val condName = if(cond) "T" else "F"
        val extraStateName = if (opCode!=null) "${condName}OP$opCode" else condName


        //schauen, ob es schon eine passende funktion gibt
        for (pre in state5PredfinedState) {
            if (pre.running == running && pre.cond == cond && isSameStatementList(pre.preF.body, f.body)) {
                pre.extraStateName+="_$extraStateName"
                return ConstantExpression(pre.reservedState.code.toULong(), ILangDataType.Int32)
            }
        }
        //neuen state reservieren und in zu tun liste hinzufügen
        val dependsOnMatch = dependendVars(f.body).vars.contains(f.getVariableByName("match"))
        val resState = stateManager.reserveStateCode(
            CompoundState(5, running, CompoundState.Cancel.Undefined,
                if (dependsOnMatch) CompoundState.Match.NoMatch else CompoundState.Match.Undefined,
                if (opCode == null)
                    CompoundState.State5OpCodeType.Unknown
                else if ((opCode and 0x3c) == 0)
                    CompoundState.State5OpCodeType.HubOperation
                else
                    CompoundState.State5OpCodeType.NoHubOperation
            )
        )
        state5PredfinedState.add(PredefinedState5Function(running, cond, dependsOnMatch, f, resState, extraStateName, resState.state.state5OpCodeType))
        return ConstantExpression(resState.code.toULong(), ILangDataType.Int32)
    }
    fun resolveState5(pre:PredefinedState5Function) {
        fun resolve(f:Function, initialState: CompoundState, newState5Cb:(newMatchState:CompoundState.Match)->CompoundStateManager.StateCode):Function {
            val varState = PartialValueVarState()
            when (initialState.match) {
                CompoundState.Match.Match -> varState.setPartialValue(f.getVariableByName("match"),true,false)
                CompoundState.Match.NoMatch -> varState.setPartialValue(f.getVariableByName("match"),false,false)
                CompoundState.Match.Undefined -> {}
            }
            varState.setPartialValueInitialConstraint(f.getVariableByName("p_reg"), PartialValue(ILangDataType.Int32,0UL,0x1FFUL))
            return extractCompoundState(simplifyFunction(f,varState),initialState) { newState ->
                if (newState.cancel != CompoundState.Cancel.Undefined || newState.running != pre.running || newState.match == CompoundState.Match.Undefined)
                    throw Exception("invalid state transition $initialState${pre.extraStateName} -> $newState")
                if (newState.baseState == 4)
                    handleState4(newState.running, pre.cond)
                else if (newState.baseState == 5)
                    //
                    ConstantExpression(newState5Cb(newState.match).code.toULong(),ILangDataType.Int32)
                else
                    throw Exception("invalid state transition $initialState${pre.extraStateName} -> $newState")
            }
        }

        val matchCompoundState = CompoundState(5, pre.running, CompoundState.Cancel.Undefined, CompoundState.Match.Match,pre.state5OpCodeType)

        val noMatchState = pre.reservedState
        val matchState = if (pre.dependsOnMatch) {
            addOrLookupFunction(
                resolve(pre.preF, matchCompoundState) { throw Exception("illegal transition from state5 match to state5 nomatch") },
                matchCompoundState,
                pre.extraStateName
            )
        }
        else
            noMatchState

        noMatchState.setFunction(extractMacroAndOptimize(resolve(pre.preF,noMatchState.state) {
            newMatchState -> if (newMatchState == CompoundState.Match.Match) matchState else noMatchState
        }),pre.extraStateName)

    }

    fun extractMacroAndOptimize(f:Function):Function {
        return removeUnsuedLocals(reduceFunction(extractMacro(simplifyFunction(
            moveMatchingStatementsUp(f) { it is LetStatement && (it.destVar.name == "source" || it.destVar.name == "sx" || it.destVar.name == "dest" || it.destVar.name == "compound_state")},
            PartialValueVarState())
            ,cogClkPeripheralsMacro),true))
    }
}