package special
import ilang.*
import ilang.Function
import java.lang.Exception

//ersetzt flags durch flagz und flagc
fun replaceFlagRegister(f:Function, flagsVar:ILangVariable, zflagVar:ILangVariable, cflagVar:ILangVariable):Function {
    fun replaceInExpr(rootExpr:Expression):Expression = rootExpr.map {
        if (it is TestBitExpression && it.expr is VariableExpression && it.expr.variable === flagsVar && it.bitNo is ConstantExpression) {
            //testbit32 (flags, no)
            VariableExpression(if (it.bitNo.value == 0UL) zflagVar else if (it.bitNo.value == 1UL) cflagVar else throw Exception("unknown flag no ${it.bitNo.value}"))
        }
        else if (it is ShiftExpression && !it.shiftLeft && it.left is AdditiveOrBitwiseExpression && it.right is VariableExpression && it.right.variable === flagsVar) {
            val newFlags = opOr32(
                MuxExpression(ILangDataType.Int32, VariableExpression(zflagVar), ConstantExpression(1UL,ILangDataType.Int32),ConstantExpression(0UL,ILangDataType.Int32)),
                MuxExpression(ILangDataType.Int32, VariableExpression(cflagVar), ConstantExpression(2UL,ILangDataType.Int32),ConstantExpression(0UL,ILangDataType.Int32))
            )
            val newExpr = it.modified(it.left,newFlags)
            newExpr
        }
        else
            it
    }.map {
        if (it is VariableExpression && it.variable === flagsVar)
            throw Exception("unsupported flags expression $rootExpr")
        it
    }
    fun replaceInStm(stmList:List<Statement>):List<Statement> {
        val result = ArrayList<Statement>()
        for (stm in stmList) {
            when (stm) {
                is LetStatement ->
                    result.addAll(if (stm.destVar === flagsVar) {
                        //Let flags statement
                        if (stm.expr is ConstantExpression && stm.expr.value == 0UL) { //let flags = 0
                            listOf(LetStatement(zflagVar,ConstantExpression(false)),LetStatement(cflagVar,ConstantExpression(false)))
                        }
                        else if (stm.expr is MuxExpression
                            && stm.expr.trueExpr is AdditiveOrBitwiseExpression && stm.expr.trueExpr.operation == AdditiveOrBitwiseExpression.Operation.Or && stm.expr.trueExpr.left is VariableExpression && stm.expr.trueExpr.left.variable === flagsVar && stm.expr.trueExpr.right is ConstantExpression
                            && stm.expr.falseExpr is AdditiveOrBitwiseExpression && stm.expr.falseExpr.operation == AdditiveOrBitwiseExpression.Operation.And && stm.expr.falseExpr.left is VariableExpression && stm.expr.falseExpr.left.variable === flagsVar && stm.expr.falseExpr.right is ConstantExpression
                        ) { //let let flags (mux32 * (or32 flags C1) (and32 flags C2))
                            //zeroflag: Bit 0
                            //caryflag: Bit 1
                            val newStm = if (stm.expr.trueExpr.right.value == 1UL && stm.expr.falseExpr.right.value == 2UL)
                                LetStatement(zflagVar,stm.expr.condition)
                            else if (stm.expr.trueExpr.right.value == 2UL && stm.expr.falseExpr.right.value == 1UL)
                                LetStatement(cflagVar,stm.expr.condition)
                            else
                                throw Exception("invalid let flags stm $stm")
                            listOf(newStm)
                        }
                        else
                            throw Exception("unknown let flags statement $stm")
                    }
                    else
                        listOf(stm.modified(replaceInExpr(stm.expr)))
                )
                is LetArrayStatement -> result.add(stm.modified(
                    replaceInExpr(stm.idx),
                    replaceInExpr(stm.expr)
                ))
                is SwitchStatement -> result.add(stm.modified(
                    replaceInExpr(stm.expr),
                    stm.cases.map { it.modified(replaceInStm(it.body)) })
                )
                is MacroStatement -> throw Exception("Macro Statement not allowed here")
            }
        }
        return useNewStatements(stmList, result)
    }
    val newBody = replaceInStm(f.body)
    if (dependendVars(newBody).vars.contains(flagsVar) || affectedVars(newBody).vars.contains(flagsVar))
        throw Exception("body still contains flags var")
    return f.modified(newBody)
}