package special

import ilang.*
import ilang.Function

fun findFirstLetStatement(destVar:ILangVariable, stmList:List<Statement>):Int? {
    for ((i,stm) in stmList.withIndex())
        if (stm is LetStatement && stm.destVar === destVar)
            return i
    return null
}

class ReservedCompoundStates(manager:CompoundStateManager) {
    val state1StoppedNoCancel:CompoundStateManager.StateCode
    val state1StoppedCancel:CompoundStateManager.StateCode
    val state1RunningNoCancel:CompoundStateManager.StateCode
    val state1RunningCancel:CompoundStateManager.StateCode
    init {
        state1StoppedNoCancel = manager.reserveStateCode(CompoundState(1,CompoundState.Running.Stopped,CompoundState.Cancel.NoCancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown))
        state1StoppedCancel = manager.reserveStateCode(CompoundState(1,CompoundState.Running.Stopped,CompoundState.Cancel.Cancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown))
        state1RunningNoCancel = manager.reserveStateCode(CompoundState(1,CompoundState.Running.Running,CompoundState.Cancel.NoCancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown))
        state1RunningCancel = manager.reserveStateCode(CompoundState(1,CompoundState.Running.Running,CompoundState.Cancel.Cancel,CompoundState.Match.NoMatch,CompoundState.State5OpCodeType.Unknown))
    }
}
class CogClkState4Result(val reservedStates:ReservedCompoundStates, val stoppedCondFalseFunc:CompoundStateManager.StateCode, val stoppedCondTrueFunc:CompoundStateManager.StateCode, val runningCondFalseFunc:CompoundStateManager.StateCode, val runningCondTrueFunc:List<CompoundStateManager.StateCode>, val zFlagMacros:List<MacroStatement>, val controlMacro:MacroStatement)

class CogClkState4Optimizer(private val stateManager:CompoundStateManager, val cogClkPeripheralsMacro:MacroStatement, val rawState4Function: Function) {
    //hier wird die cogclk_state4 Funktion optimiert und in viele Varianten zerteilt

    enum class StatementClass {
        NonPeripherals,ZFlagAndSimpleState,SetControlRegs,PReg,ZFlagAndSimpleStateAndControlRegsAndPeripherals,PeripheralsMacro,NotZFlagAndSimpleStateAndControlRegsAndPeripherals
    }

    val reservedStates = ReservedCompoundStates(stateManager)

    //Ein Makro für das Setzen von Counter und Vid Registern
    var controlRegisterMacro:MacroStatement? = null

    //hauptfunktion von running==false, ruft je nach cond stoppedCondFalseResult oder stoppedCondTrueResult auf
    var cogClkState4StoppedDispatch:Function? = null
    //running==false, cond==false
    var stoppedCondFalseResult:CompoundStateManager.StateCode? = null
    //running==false, cond==true
    var stoppedCondTrueResult:CompoundStateManager.StateCode? = null


    //Hauptfunktion von running == true, ruft je nach cond runningCondFalseFunction oder runningCondTrueFunctionsResult[runningCondTrueMap[(instr_reg)>>23]] auf
    var cogClkState4RunningDispatch:Function? = null

    //running == true, cond==false
    var runningCondFalseFunction:CompoundStateManager.StateCode? = null

    //running == true, cond == true, sehr viele funktionen, je nach opcode variationen
    val runningCondTrueFunctionsResult = HashSet<CompoundStateManager.StateCode>()

    //das Ende der jeweiligen Cog State 4 Funktionen ist immer sehr ähnlich und eines der folgenden Makros
    val zFlagMacros = ArrayList<MacroStatement>()




    val runningCondTrueMap = Array<CompoundStateManager.StateCode?>(512) {null}//mapping von opCode*2+inst(bit17) -> runningCondTrueFunctionsResult[idx]

    private val outpCogAluRVar = rawState4Function.getVariableByName("outp_cog_alu_r")
    private val ramQRegVar = rawState4Function.getVariableByName("reg_cog_ram_q_reg")

    private val instrNextVar = ILangVariable("instr_next",ILangDataType.Int32,ILangVariableDirection.Local)

    private val controlRegisters = listOf("reg_cog_ctra_phase","reg_cog_ctra_frequency","reg_cog_ctra_ctrmode","reg_cog_ctra_ctrplldiv",
        "reg_cog_ctra_ctrbpin","reg_cog_ctra_ctrapin","reg_cog_ctrb_phase","reg_cog_ctrb_frequency","reg_cog_ctrb_ctrmode",
        "reg_cog_ctrb_ctrplldiv","reg_cog_ctrb_ctrbpin","reg_cog_ctrb_ctrapin",
        "reg_cog_vid_vcfg_vmode","reg_cog_vid_vcfg_cmode","reg_cog_vid_vcfg_chroma1","reg_cog_vid_vcfg_chroma0",
        "reg_cog_vid_vcfg_auralsub","reg_cog_vid_vcfg_vgroup","reg_cog_vid_vcfg_vpins","reg_cog_vid_vscl_pixelclocks","reg_cog_vid_vscl_frameclocks",
        "dira","outa"
    )

    private fun addFunction(f:Function, running:CompoundState.Running, cond:Boolean,opCode6Bit:Int?,subOpcode3Bit:Int?):CompoundStateManager.StateCode {
        val condName = if (cond) "T" else "F"
        val extraStateName = if (opCode6Bit==null || subOpcode3Bit==null) condName else "${condName}M${opCode6Bit}F$subOpcode3Bit"
        return stateManager.addFunction(removeUnsuedLocals(f),CompoundState(4,running,CompoundState.Cancel.Undefined,CompoundState.Match.Undefined,CompoundState.State5OpCodeType.Unknown),extraStateName)
    }

    fun cppDispatcherFunctions():List<Function> {
        return listOf(cogClkState4RunningDispatch!!,cogClkState4StoppedDispatch!!)
    }

    fun removeReadArrayAccessDependency(f:Function):Function {
        //im state 4 wird das array zwar nach reg_cog_ram_q_reg gelesen, aber dieser wert wird niemals gelesen
        return f.modified(f.body+ listOf(LetStatement(ramQRegVar,ConstantExpression(0UL,ILangDataType.Int32))))
    }

    fun handleState4Function():CogClkState4Result {
        handleState4StoppedFunction(removeUnsuedLocals(simplifyFunction(rawState4Function,
            PartialValueVarState(rawState4Function,listOf(
                Pair("ena",true),
                Pair("running",false)
            ))
                .setPartialValueInitialConstraint(rawState4Function.getVariableByName("p_reg"), PartialValue(ILangDataType.Int32,0UL,0x1FFUL))
                .setNeverForwardVariable(rawState4Function.getVariableByName("instr_dst")))
        ))

        handleState4RunningFunction(removeUnsuedLocals(simplifyFunction(rawState4Function,
            PartialValueVarState(rawState4Function,listOf(
                Pair("ena",true),
                Pair("running",true)
            )))))

        val stopCF = stoppedCondFalseResult ?: throw Exception("function not generated")
        val stopCT = stoppedCondTrueResult ?: throw Exception("function not generated")
        val runCF = runningCondFalseFunction ?: throw Exception("function not generated")
        val runCT = runningCondTrueMap.map { it ?: throw Exception("function not generated") }
        val ctrlMacro = controlRegisterMacro ?: throw Exception("function not generated")

        return CogClkState4Result(reservedStates, stopCF, stopCT, runCF, runCT, zFlagMacros, ctrlMacro)
    }

    private fun handleState4StoppedFunction(rawState4StoppedFunction: Function) {
        val instrDstVar = rawState4StoppedFunction.getVariableByName("instr_dst")
        val extractedCondition = ExtractedConditionCode(removeReadArrayAccessDependency(rawState4StoppedFunction),false, listOf(instrDstVar))

        val falseFunc = moveStatementsOfClassDown(simplifyFunction(moveStatementsOfClassDown(moveStatementsOfClassDown(moveStatementsOfClassUp(moveStatementsOfClassDown(extractedCondition.condFalseFunction,StatementClass.NonPeripherals),StatementClass.SetControlRegs),StatementClass.ZFlagAndSimpleState),StatementClass.PReg),PartialValueVarState().setNeverForwardVariable(instrDstVar)),StatementClass.SetControlRegs)
        val trueFunc = moveStatementsOfClassDown(simplifyFunction(moveStatementsOfClassDown(moveStatementsOfClassDown(moveStatementsOfClassUp(moveStatementsOfClassDown(extractedCondition.condTrueFunction,StatementClass.NonPeripherals),StatementClass.SetControlRegs),StatementClass.ZFlagAndSimpleState),StatementClass.PReg),PartialValueVarState().setNeverForwardVariable(instrDstVar)),StatementClass.SetControlRegs)
        //ControlRegisterMakro erkennen
        val macroBody = ArrayList<Statement>()
        var idx=0
        while(idx<trueFunc.body.size && !isOfStatementClass(trueFunc.body[idx],StatementClass.SetControlRegs))
            ++idx
        while(idx<trueFunc.body.size && isOfStatementClass(trueFunc.body[idx],StatementClass.SetControlRegs)) {
            macroBody.add(trueFunc.body[idx])
            ++idx
        }
        while(idx<trueFunc.body.size && !isOfStatementClass(trueFunc.body[idx],StatementClass.SetControlRegs))
            ++idx
        if (idx<trueFunc.body.size) {
            println(serializeILang(trueFunc))
            throw Exception("control macro statements not in order: ${trueFunc.body[idx]}")
        }
        if (macroBody.isEmpty())
            throw Exception("control macro body was empty")


        val depVars = dependendVars(macroBody)
        val affVars = affectedVars(macroBody)
        if (
            depVars.arrs.isNotEmpty()
            || affVars.arrs.isNotEmpty()
            || affVars.vars.subtract(controlRegisters.map{trueFunc.getVariableByName(it)}).isNotEmpty()
            || depVars.vars.subtract(listOf("instr_dst","outp_cog_alu_r").map{trueFunc.getVariableByName(it)}).isNotEmpty()
        )
            throw Exception("invalid dep/aff vars in controlRegisterMacro")
        val ctrlRegMacro = MacroStatement("macroCog4SetControlRegisters",true,rawState4StoppedFunction.registerVarSet(), macroBody, setOf(), depVars.vars.toList().sortedBy { it.name })
        controlRegisterMacro = ctrlRegMacro
        stoppedCondFalseResult = addFunction(extractZFlagMacro(optimizeLocals(falseFunc,listOf(outpCogAluRVar,instrDstVar)),false, false, null),CompoundState.Running.Stopped,false,null,null)
        if (dependsOnHubMem(stoppedCondFalseResult?.f!!))
            throw Exception("state4 stopped !cond depends on bus_q_reg")
        stoppedCondTrueResult = addFunction(extractZFlagMacro(optimizeLocals(trueFunc,listOf(outpCogAluRVar,instrDstVar)),false, true, null),CompoundState.Running.Stopped,true,null,null)

        //Dummy Dispatcher Cpp Funktion erstellen
        val regSets = rawState4StoppedFunction.registerVarSet()
        cogClkState4StoppedDispatch = Function("cogClkState4StoppedDispatch",
            ILangMultiVariableSet(listOf(ILangVariableSet("*"))+regSets.sets),
            listOf(
                SwitchStatement(extractedCondition.condition,listOf(
                    SwitchStatement.Case(listOf(0UL),listOf(generateDummyMacroFromNonParameterFunc(stoppedCondFalseResult!!.f!!))),
                    SwitchStatement.Case(listOf(1UL),listOf(generateDummyMacroFromNonParameterFunc(stoppedCondTrueResult!!.f!!)))
                ))
            )
        )
    }

    private fun removeTmpZFlag(f:Function):Function {
        //korrektur mit zflag outp_cog_alu_zo
        //entfernt nach möglichkeit outp_cog_alu_zo innerhalb eines ifs
        if (affectedVars(f.body).vars.count { it.name == "outp_cog_alu_zo" } == 0)
            return f
        val oldBody = ArrayList(moveMatchingStatementsUp(f) { it is LetStatement && it.destVar.name == "zflag"}.body)
        val newBody = ArrayList<Statement>()
        while (oldBody.isNotEmpty()) {
            val stm = oldBody.removeFirst()
            if (stm !is SwitchStatement || oldBody.isEmpty()) {
                newBody.add(stm)
                continue
            }
            val nextStm = oldBody.removeFirst()
            if (nextStm !is LetStatement || nextStm.destVar.name != "zflag" || nextStm.expr !is VariableExpression) {
                newBody.add(stm)
                newBody.add(nextStm)
                continue
            }
            //zflag in den if body schieben
            newBody.add(stm.modified(stm.expr,stm.cases.map { SwitchStatement.Case(it.values,it.body+listOf(nextStm)) }))
        }
        return f.modified(newBody)
    }
    private fun removeLastSetCogRamReg0Stm(f:Function):Function {
        //es wurde ein let reg_cog_ram_q_reg,0 künstlich eingefügt
        //dieses jetzt wieder entfernen
        if (f.body.isEmpty())
            return f
        val lastStm = f.body.last() as? LetStatement ?: return f
        if (lastStm.destVar !== ramQRegVar || lastStm.expr !is ConstantExpression || lastStm.expr.value != 0UL)
            return f
        return f.modified(f.body.dropLast(1))
    }
    private fun optimizeLocals(initF:Function, neverFwdVars:List<ILangVariable>):Function {
        val moveUpVars = listOf("instr_dst","outp_cog_alu_r","px","cflag","zflag")
        val tmpRes1 = removeTmpZFlag(initF)

        val varState = PartialValueVarState()
        for (nf in neverFwdVars)
            varState.setNeverForwardVariable(nf)

        val tmpRes2 = moveStatementsOfClassDown(removeLastSetCogRamReg0Stm(simplifyFunction(
            moveMatchingStatementsDown(
                moveMatchingStatementsUp(tmpRes1) {it is LetStatement && moveUpVars.contains(it.destVar.name) },
            ){ it is LetStatement && it.destVar === ramQRegVar},
            varState
            )),StatementClass.ZFlagAndSimpleState)
        return tmpRes2
    }

    private fun generateDummyMacroFromNonParameterFunc(f:Function):MacroStatement =
        MacroStatement(f.name,true,f.registerVarSet(),listOf(), setOf(), listOf())

    private fun simplifyPReg(f:Function):Function = simplifyFunction(f,
        PartialValueVarState().setPartialValueInitialConstraint(
            f.getVariableByName("p_reg"),
            PartialValue(ILangDataType.Int32,0UL,0x1FFUL)))

    private fun dependsOnHubMem(f:Function):Boolean = dependendVars(f.body).vars.contains(f.getVariableByName("bus_q_reg"))

    private fun handleState4RunningFunction(rawState4RunningFunction: Function) {

        val extractedCondition = ExtractedConditionCode(removeReadArrayAccessDependency(rawState4RunningFunction),true, listOf())
        val falseFunc = simplifyPReg(moveStatementsOfClassDown(moveStatementsOfClassDown(moveStatementsOfClassDown(extractedCondition.condFalseFunction,StatementClass.NonPeripherals),StatementClass.ZFlagAndSimpleState),StatementClass.PReg))
        val trueFunc = moveStatementsOfClassDown(moveStatementsOfClassDown(extractedCondition.condTrueFunction,StatementClass.NonPeripherals),StatementClass.PReg)


        runningCondFalseFunction = addFunction(extractZFlagMacro(optimizeLocals(falseFunc,listOf(outpCogAluRVar)),true,false, null),CompoundState.Running.Running,false,null,null)
        if (dependsOnHubMem(runningCondFalseFunction?.f!!))
            throw Exception("state4 running !cond depends on bus_q_reg")
        handleState4CondTrueSpecialized(false, trueFunc)
        handleState4CondTrueSpecialized(true, trueFunc)


        //Dummy Dispatcher Cpp Funktion erstellen
        val regSets = rawState4RunningFunction.registerVarSet()
        val runningCases = runningCondTrueMap.mapIndexed {
                opCode,state ->
            val callFunc = state?.f ?: throw Exception("no state4 func found for $opCode")
            SwitchStatement.Case(listOf(opCode.toULong()),listOf(MacroStatement(callFunc.name,true,regSets, listOf(),
                setOf(), listOf())))
        }
        cogClkState4RunningDispatch = Function("cogClkState4RunningDispatch",
            ILangMultiVariableSet(listOf(ILangVariableSet("*"))+regSets.sets),
            listOf(
                SwitchStatement(extractedCondition.condition,listOf(
                    SwitchStatement.Case(listOf(0UL),listOf(generateDummyMacroFromNonParameterFunc(runningCondFalseFunction!!.f!!))),
                    SwitchStatement.Case(listOf(1UL),listOf(SwitchStatement(opShr32(VariableExpression(regSets.getVariableByName("instruction_reg")),ConstantExpression(23UL,ILangDataType.Int32)),runningCases))))
                ))
        )
    }

    private fun isOfStatementClass(stm:Statement, cls:StatementClass):Boolean {
        if (cls == StatementClass.NotZFlagAndSimpleStateAndControlRegsAndPeripherals)
            return !isOfStatementClass(stm,StatementClass.ZFlagAndSimpleStateAndControlRegsAndPeripherals)
        if (cls == StatementClass.PeripheralsMacro || cls == StatementClass.ZFlagAndSimpleStateAndControlRegsAndPeripherals) {
            for (os in cogClkPeripheralsMacro.body)
                if (os.isSameStatement(stm))
                    return true
            if (cls == StatementClass.PeripheralsMacro)
                return false
        }
        if (stm is SwitchStatement && stm.expr.dataType == ILangDataType.Bool) {
            val tc = stm.matchingCase(1UL)
            for (subS in tc.body)
                if (!isOfStatementClass(subS,cls))
                    return false
            val tf = stm.matchingCase(0UL)
            for (subS in tf.body)
                if (!isOfStatementClass(subS,cls))
                    return false
            return true
        }
        if (stm !is LetStatement)
            return false
        val simpleState = listOf("state","match","p_reg","cancel","instruction_reg","jump_cancel","px","running","compound_state","instr_next")
        val allowedDestVars = when(cls) {
            StatementClass.NonPeripherals -> listOf("state","match","instruction_reg","zflag","cflag","px","p_reg","cancel","jump_cancel","jump","outp_cog_alu_r","outp_cog_alu_zo","outp_cog_alu_co","loc_cog_alu_tmpResBitOp","loc_cog_alu_tmpResSigned","loc_cog_alu_tmpResUnsigned","in_cog_alu_ci","in_cog_alu_zi","instr_dst")
            StatementClass.ZFlagAndSimpleState -> simpleState
            StatementClass.PReg -> listOf("p_reg")
            StatementClass.SetControlRegs -> controlRegisters
            StatementClass.ZFlagAndSimpleStateAndControlRegsAndPeripherals -> controlRegisters+simpleState
            StatementClass.PeripheralsMacro,StatementClass.NotZFlagAndSimpleStateAndControlRegsAndPeripherals -> listOf()
        }
        return allowedDestVars.contains(stm.destVar.name)
    }

    private fun moveStatementsOfClassUp(f:Function, cls:StatementClass):Function = moveMatchingStatementsUp(f) {
        isOfStatementClass(it, cls)
    }

    private fun moveStatementsOfClassDown(f:Function, cls:StatementClass):Function = moveMatchingStatementsDown(f) {
        isOfStatementClass(it, cls)
    }

    private fun reduceJumpCancel(f:Function):Function {
        fun dependsOnJumpCancel(stm:Statement) = dependendVars(stm).vars.count { it.name == "jump_cancel" } > 0
        val f2 = moveMatchingStatementsDown(moveMatchingStatementsDown(f) { dependsOnJumpCancel(it) } ) {
            it is SwitchStatement && it.expr is VariableExpression && it.expr.variable.name == "jump_cancel"
        }
        if (f2.body.isEmpty())
            return f2
        val last = f2.body.last()
        if (last !is SwitchStatement || last.expr !is VariableExpression || last.expr.variable.name != "jump_cancel")
            return f2
        val remainingTop = ArrayList(f2.body.dropLast(1))
        val jumpCancelStms = ArrayList<Statement>()
        while (remainingTop.isNotEmpty() && dependsOnJumpCancel(remainingTop.last()))
            jumpCancelStms.add(0,remainingTop.removeLast())
        if (jumpCancelStms.isEmpty())
            return f2
        return f2.modified(remainingTop+ listOf(last.modified(
            last.expr,last.cases.map {
                it.modified(jumpCancelStms+it.body)
            }
        )))
    }

    private fun modifyPxVar(f:Function):Function {
        return f.modified(mapStmList(f.body) {
            if (it is LetStatement && it.destVar.name == "px") {
                if (it.expr !is AdditiveOrBitwiseExpression || it.expr.left.toILang() != expectedState3SourceExpression)
                    throw Exception("invalid state 4 let px stm: $it")
                //px = and(mux...,c) durch px = and(source,c) ersetzen
                it.modified(it.expr.modified(VariableExpression(f.getVariableByName("source")),it.expr.right))
            }
            else
                it
        })
    }

    private fun handleInstructionRegLoad(f0:Function, opCodeUpper6Bit: Int?):Function {
        //das laden des opcodes wird von state3 nach state4 verschoben und hier eingefügt
        val f1=f0.localsAdded(listOf(instrNextVar)).modified(moveMatchingStatementsUp(modifyPxVar(f0)) {it is LetStatement && it.destVar.name == "px" }.body)
        val firstPxStm = f1.body.firstOrNull()
        val f2Body = if (firstPxStm is LetStatement && firstPxStm.destVar.name == "px") {
            val loadStm = LetStatement(instrNextVar, getExpectedState3MemReadBufferExpression(f1, opCodeUpper6Bit) { "px" })
            listOf(firstPxStm,loadStm)+f1.body.drop(1)
        }
        else {
            val loadStm = LetStatement(instrNextVar, getExpectedState3MemReadBufferExpression(f1, opCodeUpper6Bit) { throw Exception("missing px") })
            listOf(loadStm)+f1.body
        }
        return f1.modified(mapStmList(f2Body) {
            if (it is LetStatement && it.destVar.name == "instruction_reg") {
                if (it.expr !is VariableExpression || it.expr.variable.name != "reg_cog_ram_q_reg")
                    throw Exception("invalid stm: $it")
                it.modified(VariableExpression(instrNextVar))
            }
            else
                it
        })
    }

    private fun addOrSearchZMacro(remainingFunction:List<Statement>,macroStms:Function, extractControlRegs:Boolean):MacroStatement {
        val mainFunDepend = dependendVars(remainingFunction).vars
        val mainFunAffected = affectedVars(remainingFunction).vars
        val macroAffectedVars = affectedVars(macroStms.body)
        val macroDependVars = dependendVars(macroStms.body)
        //if (macroAffectedVars.arrs.isNotEmpty() || macroDependVars.arrs.isNotEmpty())
        //throw Exception("macro must not contain arrays")

        if ((macroDependVars.vars.union(mainFunDepend)).find {
                listOf(
                    "match",
                    "state",
                    "cancel",
                    "running"
                ).contains(it.name)
            } != null)
            throw Exception("state 4 fun depends on illegal vars")
        if (mainFunAffected.find { listOf("match", "state", "cancel", "running").contains(it.name) } != null)
            throw Exception("state 4 fun affects illegal vars $mainFunAffected")


        val macroDependLocals = macroDependVars.vars.filter { it.direction != ILangVariableDirection.Register }
        val macroAffectedLocals = macroAffectedVars.vars.filter { it.direction != ILangVariableDirection.Register }

        //variablen, die in makros geändert werden, dürfen niemals in der hauptfunktion vorkommen
        val macroAffectedVarsUsedInMain = macroAffectedLocals.intersect(mainFunAffected.union(mainFunDepend))
        if (macroAffectedVarsUsedInMain.isNotEmpty())
            throw Exception("macro affects variables, that are modified in parent: $macroAffectedVarsUsedInMain")

        val macroInputs = macroDependLocals.intersect(mainFunAffected.union(mainFunDepend))
        if (macroInputs.isNotEmpty()) {
            for (mi in macroInputs) {
                if (mi.name != "outp_cog_alu_r" && mi.name != "instr_dst" && mi.name != "instr_src")
                    throw Exception("TODO warning: invalid macro input ${mi.name}")
            }
        }
        val macroPureLocals = macroDependLocals.union(macroAffectedLocals).subtract(macroInputs)

        val macroStms1 = moveStatementsOfClassUp(macroStms,StatementClass.ZFlagAndSimpleState)

        val macroBody = extractMacro(
            (if (extractControlRegs) extractMacro(
                macroStms1,
                controlRegisterMacro ?: throw Exception("no control register macro set")
            ) else macroStms1),
            cogClkPeripheralsMacro
        ).body

        fun findOrAddMacro(): MacroStatement {
            for (m in zFlagMacros) {
                if (isSameStatementList(macroBody, m.body)) {
                    if (m.pureLocals != macroPureLocals || m.inputs.toSet() != macroInputs)
                        throw Exception("mismatch macro inputs/locals: ${m.pureLocals} vs $macroPureLocals / ${m.inputs} vs $macroInputs")
                    return m
                }
            }
            val m = MacroStatement(
                "macroCogState4ZF${zFlagMacros.size}",
                true,
                macroStms1.registerVarSet(),
                macroBody,
                macroPureLocals,
                macroInputs.toList().sortedBy { it.name })
            zFlagMacros.add(m)
            return m
        }
        return findOrAddMacro()
    }

    private fun extractZFlagMacro(func0:Function, isRunningPrev:Boolean, extractControlRegs:Boolean, opCodeUpper6Bit:Int?):Function {
        val f = simplifyFunction(moveStatementsOfClassUp(reduceJumpCancel(extractCompoundState(handleInstructionRegLoad(func0,opCodeUpper6Bit), CompoundState(4,if (isRunningPrev) CompoundState.Running.Running else CompoundState.Running.Stopped,CompoundState.Cancel.Undefined,CompoundState.Match.Undefined,CompoundState.State5OpCodeType.Unknown)) {
            newState ->
            if (newState.baseState != 1 || newState.match != CompoundState.Match.NoMatch)
                throw Exception("illegal transition vom state4 func: $newState")
            val isRunningNew = newState.running.toBool()
            val isCancel = when(newState.cancel) {
                CompoundState.Cancel.Cancel -> true
                CompoundState.Cancel.NoCancel -> false
                CompoundState.Cancel.Undefined -> throw Exception("illegal transition vom state4 func: $newState")
            }
            ConstantExpression((
                        if (isRunningNew)
                            if (isCancel) reservedStates.state1RunningCancel else reservedStates.state1RunningNoCancel
                        else
                            if (isCancel) reservedStates.state1StoppedCancel else reservedStates.state1StoppedNoCancel
                    ).code.toULong(),ILangDataType.Int32)
        }),StatementClass.NotZFlagAndSimpleStateAndControlRegsAndPeripherals),
            PartialValueVarState()
                .setNeverForwardVariable(rawState4Function.getVariableByName("instr_dst"))
                .setNeverForwardVariable(rawState4Function.getVariableByName("outp_cog_alu_r"))
        )
        val newTopBody = ArrayList(f.body)
        val newBottomBody = ArrayList<Statement>()
        //von unten an stms entfernen die zur klasse gehören
        while (newTopBody.isNotEmpty() && (newTopBody.last() is LetArrayStatement || isOfStatementClass(newTopBody.last(), StatementClass.ZFlagAndSimpleStateAndControlRegsAndPeripherals))) {
            newBottomBody.add(0, newTopBody.last())
            newTopBody.removeLast()
        }
        fun noTrailingIf():Function = f.modified(newTopBody+listOf(addOrSearchZMacro(newTopBody,f.modified(newBottomBody),extractControlRegs)))
        //schauen, ob das letzte statement ein if stm ist mit einem letarr
        if (newTopBody.size<2)
            return noTrailingIf()
        val lastStm = newTopBody.last()
        if (lastStm !is SwitchStatement || lastStm.expr.dataType != ILangDataType.Bool)
            return noTrailingIf()

        fun checkIfBranchIsLetArrOrNull(branch:SwitchStatement.Case):Boolean = branch.body.isEmpty() || (branch.body.size == 1 && branch.body[0] is LetArrayStatement)
        val trueBranch = lastStm.matchingCase(1UL)
        val falseBranch = lastStm.matchingCase(0UL)
        if (!checkIfBranchIsLetArrOrNull(trueBranch) || !checkIfBranchIsLetArrOrNull(falseBranch))
            return noTrailingIf()
        val secondLastStm = newTopBody[newTopBody.size-2]
        if (secondLastStm !is LetStatement || secondLastStm.destVar !== instrNextVar || secondLastStm.expr !is ArrayAccessExpression)
            throw Exception("invalid secondlast stm: $secondLastStm")
        val remainingTopBody = newTopBody.dropLast(2)
        return f.modified(remainingTopBody + listOf(
            lastStm.modified(lastStm.expr, listOf(
                trueBranch.modified(listOf(addOrSearchZMacro(remainingTopBody,f.modified(listOf(secondLastStm)+trueBranch.body+newBottomBody),extractControlRegs))),
                falseBranch.modified(listOf(addOrSearchZMacro(remainingTopBody,f.modified(listOf(secondLastStm)+falseBranch.body+newBottomBody),extractControlRegs)))
            ))
        ))
    }

    private fun handleState4CondTrueSpecialized(instructionRegBit17:Boolean, state4Function:Function) {
        val reducedFunctions = ArrayList<Function>()
        for (opCodeUpper in 0 until 256) {
            if (opCodeUpper % 64 == 0)
                println(opCodeUpper)
            reducedFunctions.add(removeUnsuedLocals(simplifyFunction(state4Function,PartialValueVarState()
                .setPartialValueInitialConstraint(state4Function.getVariableByName("instruction_reg"),
                    PartialValue(ILangDataType.Int32,(opCodeUpper.toULong() shl 24)+(if (instructionRegBit17) (1UL shl 23) else 0UL) ,(0xFF800000UL).inv())
                )
                .setNeverForwardVariable(outpCogAluRVar)
                .setPartialValueInitialConstraint(
                    state4Function.getVariableByName("p_reg"),
                    PartialValue(ILangDataType.Int32,0UL,0x1FFUL))
            )
            ))
        }

        for ((opCodeUpper,f) in reducedFunctions.withIndex()) {
            val zMovedDown = moveStatementsOfClassDown(moveStatementsOfClassDown(f, StatementClass.NonPeripherals), StatementClass.ZFlagAndSimpleState)
            val controlRegsMovedUp = moveStatementsOfClassUp(zMovedDown, StatementClass.SetControlRegs)
            val opCode6Bit = (opCodeUpper shr 2)
            val subOpcode3Bit = ((opCodeUpper and 3) shl 1) + if(instructionRegBit17) 1 else 0

            val st = addFunction(extractZFlagMacro(optimizeLocals(controlRegsMovedUp,listOf(outpCogAluRVar)),true,instructionRegBit17,opCode6Bit),CompoundState.Running.Running,true,opCode6Bit,subOpcode3Bit)
            val opCode9Bit = (opCode6Bit shl 3) + subOpcode3Bit

            runningCondTrueMap[opCode9Bit] = st
            runningCondTrueFunctionsResult.add(st)
        }
    }
}



